"""The module defines the compounds of example experiments that are shipped with
peewit. In order to run the experiments call one of the run-modules 
example_1.py, example_2.py, example_3.py, example_4.py. 

The descent definition of the e-nodes are rather short here. In real-world experiments, 
they can be longer, e.g. more code is needed to make the e-nodes work together. 
"""





## the native peewit classes needed for any peewit experiments
from peewit.core import ENode
## in case we read results from disk
from peewit.vcubes import VCube

## a class for looking at the results
#from peewit.presentation import Viewer 

## imports the example user provides for this experiment 
from numpy import mean

from ml.table import TableData
from ml.ml_data import sonar
from ml.sampling import SubSampling
from ml.kernel import KernelTray, Linear, Poly
from ml.preprocessing import S01, PPC, NanToMedian

from ml.libsvm3 import LIBSVM


## pathkeeper is used by e-node data_source
from project_conf import PK




class data(ENode):
    
    def init(self, build=True, store=False):
        pass
    
    def descent(self):
        ## the user uses pathkeeper to get the path to the data
        TableData.global_datadir = PK.abs_path('example_data_sets/')
        ##  some setting in users code
        TableData.global_build   = self.build
        TableData.global_store   = self.store
        TableData.global_seed    = 2012
        

        return  [
                  sonar(),
                ]
        


## The class names in lower case serve as identifiers (e-names) for the e-nodes. 
## As for any names parts, e-names must not contain "--" or "__" (double dash or double underscore).
## In order to references another e-node fpr descent-call, we use its e-name as arg-name of 
## the decent function.  

class sample(ENode):
    
    def init(self,
             n_folds=4):
        pass
    
    def descent(self, data):    
        
        self.send("\nDataset:", data.get_name(),'\n')

        sampler = SubSampling(
                              sequence  = data.get_X(), 
                              output_sequence = data.get_YS(), 
                              seed      = 2012,
                              n_folds   = self.n_folds,
                              max_train_size = 1000,
                              )
        
        
        return sampler.get_cv_indices()



class kernel(ENode):
    
    def descent(self):
        ## Named objects:
        ## The object below have a get_name() method that that is called 
        ## to get the value name which is use for later reference
        ## If there is no get_name method, a name is guessed. If the object 
        ## is a number, the string of that number is its name.
        k_1 = Linear()
        k_2 = Poly(g=3)
        
        return  [ k_1, 
                  k_2  
                  ]



class kernel_tray(ENode):
    
    def descent(self, data, sample, kernel ):
        
        kernel.prep = PPC(chain=[NanToMedian(), S01()])
        
        kt = KernelTray( kernel = kernel,
                         data_tray = data, 
                         )
        
    
        A = sample[0]
        kt.train_prep(kernel,A)                     
        self.kt=kt
                                    
        return [ kt ]
    
    def ascent(self):
        
        self.kt.flash_cache()
        return [None]

class sub_sample(ENode):

    def init(self,
             n_folds=4):
        pass
    
    def descent(self, data, sample ):    

        A,EP = sample
        
        
        sampler = SubSampling(sequence  = data.get_X(), 
                              restriction  = A, 
                              output_sequence   = data.get_YS(),
                              seed      = 2012,
                              n_folds   = self.n_folds,
                              max_train_size = 1000,
                              )
        
        out = sampler.get_cv_indices() 
        #assert set(out[2][1]).issubset(A)
        return out
      

class c(ENode):
    
    def descent(self):
        return list(range(-4,4))

    #def ascent(self,leaf):
    #    return leaf  


    ## An e-node can aggregate the sub-tree itself.
    #def ascent(self, vcube):   
    #    opt_cube = vcube.aggs([('argmaxo', 'c'),
    #                           ('mean','sub_sample'),
    #                           ('sel', 'leaf', 'accuracy')])
    #    return opt_cube.cube.ravel()[0]   
        
    ## Or a subsequent e-node has to do it. If no ascent function 
    ## is defined for c ..  
class c_opt(ENode):

    ## .. querying c from an node that is non-descendant does not 
    ## yield the ascent value (the output of the ascent function) 
    ## but an object that has c's p-node and vcube as attributes.
    def descent(self, leaf):
        
        ## aggs take list of aggregation that are computed in 
        ## inverse order.  
    
        vc_opt    = leaf.aggs([('argmax', 'c'),
                               ('mean','sub_sample'),
                               ('sel', 'leaf', 'accuracy')])

        ## since argmax is over on dim only, vc_opt is v-cube
        ## otherwise it would be a mapping dim-name -> v-cube
        ## in any case we have to peel the value
        c_opt = vc_opt.get_value()
        
        return [ c_opt ]
   
 

class algorithm(ENode):
    """values expected to have train and predict method"""

    def descent(self, data, c, kernel, kernel_tray):
        
        
        engine = LIBSVM( data_tray   = data,
                         kernel      = kernel, 
                         C           = 4.**c,
                         kt          = kernel_tray, 
                       )
                

        return [ engine ]


class leaf(ENode):

    def descent(self, algorithm, data, sample):
        
        A, EP   = sample
        
        model  = algorithm.train(A)
        
        Y      = data.get_YS(EP)
        
        Y_hat  = algorithm.predict(model, EP) 
        

        accuracy   =  1. * sum(  Y == Y_hat ) / sum( ( Y == 1 ) + (Y == -1) )
        precission =  1. * sum( (Y == 1 ) * (Y_hat == 1) ) / sum( Y_hat == 1 )
        recall     =  1. * sum( (Y == 1 ) * (Y_hat == 1) ) / sum( Y == 1 )
        
        
        return [ accuracy, precission, recall ]
        
        

