"""This module is the starting point for an example experiment for peewit. The experiment 
is run by calling this module. Please read the corresponding section in 
peewit/doc/readme.txt that give instructions how to make it run.

It is classical binary classification experiment. The experimental code can be 
divided into five parts: 

    1)  project configuration (given in project_conf.py): basic settings are made which
        are mostly independent of the concrete experiment
    
    2)  definitions of the e-nodes (given in enodes_1.py): the compounds of the experiment,   
        the experiment is given as tree of those. Different experiments typically share
        some e-nodes. 
    
    3)  definition of the e-tree (given in this module): a concrete experimental tree
    
    4)  run: specifying and calling the run of that tree and storing the raw results in v-cubes
    
    5)  presentation of results (given in show_1.py): read result cubes, select parts, and
        build output for human eyes. 
        
The specific goal of this example is to illustrate the use of e-trees with branches
and ascent production.
"""

## the native peewit classes needed for any peewit experiments
from peewit.core import Experiment

#### 1) project configuration

## import some paths meta-data that have been out-sourced 
## because they are valid not only for this experiment
## since that conf module adds paths to python-path
## we have to import it before some of the imports below
from project_conf import *


#### 2) definitions of the e-nodes are given the module enodes_1.

from enodes_1 import *

#### 3) definition of the e-tree 
## 
## This sets the order in which the e-node are visited.
## Here, some e-nodes are used more than once. This means that the identifier is
## not unique in the e-tree. But it should be unique on any path from the root to a leaf.

etree = [     data(),
              sample(ts=[0,2]), 
              kernel(),
              kernel_tray(),
              [ [ c(),
                  sub_sample(n_folds=4),
                  algorithm( ),
                  leaf(input_refs = [['sample', 'sub_sample']] ), 
                ] , 
                [ c_opt(),
                  algorithm( input_refs = [['c', 'c_opt']] ), # don't take c, take c_opt
                  leaf(),
                ] , 
              ]
            ]



#### 4) application/run of the e-tree

if __name__ == '__main__':

    ## experiment name (tree-name) will be set to the name of the module
    ## set version name by hand
    EP = Experiment(pathkeeper = PK,
                   #launcher   = LC,   ## parallelization, specify launch below
                   #archivist  = AR,   ## making experiment persistent
                   version    = 7,     ## string or stringable
                   #remove_current_files=1,  # start over again
                   #verbose_descent = ['leaf'], ## anounce each descent call
                                                ## for these enodes
                   test_run = True,   ## restrict descents to test-selections (ts)
                                       ## appends '__test' to version-name  
                   )
    
    
    ### code and file management services
    ## set archivist in experiment constructor above to permanently keep the
    ## the code and other files of the run (see also settings on AR in 
    ## project_conf.py. 
    
    ## go back to earlier version of your code and quit
    ## note that this will change this very file as well,
    ## Restore by run-number: The run_number 
    ## is hex-number (for instance a33c0b228e) that is attached to the results 
    ## of a run if the archivist is used. If there is a run-number, the result-viewer
    ## will include it in the output such that you have an ancor when look at the
    ## output at later time
    #AR.restore('a33c0b228e')
    ## Such call after a run as done here is a bit strange.
    ## You can make this call as soon as AR is sufficiently initialized, e.g. after 
    ## project_conf.py is imported.   
    ## However, you can also refer to other version of an given experiment
    ## to which AR is already attached.
    #AR.restore( version='6' )
    ## To come back to present, call AR.restore() without args
    #AR.restore()
    
    
    ### parallelization
     
    ## Depends on unison and needs a target.  
    ## Rerun module later to collect partial cubes,
    ## specify e-nodes along which to parallelize (axes).
    ## By now, number of d-values has to be given along with 
    ## the repspective e-name.        
    #LC.specify_launch(target   = 'leaf',
    #                  axes     = [ ('kernel',2), ('sample',3) ],
    #                  dismiss  = 0, ## if dismiss, dismiss run from cluster
    #                  dont     = 0, ## if dont, dont use launcher
    #                 )
    

    ### now the experiment
    
    ## read and digest specification 
    EP.define(etree)

    ## go 
    vc = EP.run()
    
    ## store v-cube
    vc.save()
    
    ## by default  rightmost leaf is the target,
    ## but you can also specify target e-name,
    ## if its not unique, the rightmost e-node with that e-name is chosen 
    #vc = EP.build_vcube('leaf')

    ## get another v-cube containing kernel object 
    ## ( whatever the motivation may be  )
    #vc_ = EP.build_vcube('kernel')
    #vc_.save()
    
    
    #### show results
    
    ## above we announced to separate the run-module from a module 
    ## that cares for producing readable outputs 
    ## but for simplicity we do some output right here
    
    #vc.aggregations([('sample','mean')])   
    ## create an object that cares for presentation of result data cube
    
    vc.show()
    vc.vcubes['c_opt'].show()
    #vc.vcubes['c_opt'].vcubes['leaf'].show()
    
 
   