"""Global setting of the project may be define here.

The settings made here concern the names and paths for experiment and result cubes.
They are used in order to identify and address experiments and results.

Directories are given as relative paths in posix-notation but before they are applied 
they translated into the paths notation of the given os. That is, when we use those strings
to set the paths of an experiment, the strings a taken as I{paths-descriptors} from which the 
actual path string is derived. 
"""

from peewit.paths import PathKeeper 
from peewit.parallelization import OPBS_Launcher
from peewit.archiving import Archivist

## cares for paths and name parts
PK = PathKeeper() 

## an identifier of your working group
## peewit assume unique host name within your group 
PK.group = 'ke_tud_de'

##  base_dir is given as absolute path (starting with /) or relative to 
##  your home directory. 
##  All paths below are relative to that base-dir or some dir below. 
##  All paths are given as _path-descriptors_ using 
##  posix notation. peewit cares for translation to windows notation. 
PK.base_dir    = 'code/peewit/pw_base' 
                            
## project directory relative to base_dir
PK.project_dir = 'example_project/'

### python-path management

##  All paths will be added on the fly all but path to peewit itself
##  which is needed in advance. Paths are given relative to base_dir. 
##  plain user space paths
PK.python_paths    = [ 'example_user_packages/',                        
                     ]
## some paths may depend on architecture, 32 versus 64 bit
PK.python_paths_32 = ['../../modules_extended_32/libsvm-3.16/python/' ]
PK.python_paths_64 = ['../../modules_extended_64/libsvm-3.16/python/' ]



## others may not belong to user space but are from outer  
## space what also demands extra treatment. Those can also 
## be given as absolute paths. Keep in mind that
## "Outer space, its not a better place, Grace."
PK.python_paths_os = [] 

## more dirs that peewit should know about
## by now, more_dirs relevant for launcher archivist and launcher
## below, they inherit this list if proper ones are not set 
PK.more_dirs = ['example_data_sets/']

### archiving settings

AR = Archivist()

## by default archivist covers the following dirs: PK.project_dir, 
## peewit framework dir, all PK.python_paths_X-dirs except the _os ones.
## Define dirs to be covered by the archivist here or inherit 
## extra dirs from PK.more_dirs
#AR.more_dirs = [] ## no extra dirs for AR


## Exclude individual paths through path-descriptors or groups by
## adding before-mentioned groups of paths
AR.exclude_from_archive += PK.python_paths_32 + PK.python_paths_64

## integrate repos that exist independently form peewit
## passive: only write revision-number to history
## active: commit on each run
#AR.repos_passive_integration = []
#AR.repos_active_integration  = []

#AR.omit_pyd_files = True

## Update the peewit framework from sourceforge.net
## This also overwrites the example project 
#AR.pull_peewit()

#AR.verbosity = 1


### settings for parallelization

LC = OPBS_Launcher()

LC.remote_head         = 'cluster'
LC.remote_user         = 'lw'
LC.notification_mail_address = 'lorenz@ke.tu-darmstadt.de'


## some more advanced cluster settings

#LC.queue               = 'all.q@@rack' 
#LC.python              = 'python2.6'
#LC.remote_head_architecture = '64bit',
#LC.exclude_from_sync       = [
#                              'modules_extended_32/libsvm-2.86/python',
#                              'modules_extended_64/libsvm-2.86/python',
#                              ]

## you may have source a script on remote haed that sets env-variables 
## needed for submitting jobs.  
#LC.path_cluster_settings ='/usr/local/sge-root/default/common/settings.sh'

## Set true for debugging synchronization 
## Unison-sync can break. In many cases, this can be fixed by removing 
## the unison-archive-files as unison tells you if you turn on this flag. 
#LC.verbose_sync = 0


## paths on remote head 

## paths that are not set are assumed to be same relative paths 
## as on local system. Further, the user space 
## python paths (not .._os) defined above will be 
## also apply on remote hosts

LC.remote_home_dir     = '/home/lw/'
#LC.remote_base_dir     = None

## this one is always needed
LC.remote_peewit_path = 'peewit/peewit_packages/'


## outer space python paths, again relative to base dir or absolute
LC.remote_python_paths_os  = []    

## extra dirs? See AR.more_dirs above.
#LC.more_dirs = None ## unset: inherit PK.more_dirs

## another launcher for parallelization 
## it plants jobs on each node from given list by hand

## variant 1: local machine is connected to same file-system as 
## computing nodes
#LC = Local_NFS_Launcher('LFS') 
## variant 2: local machine is outside the shared file-system 
#LC = Remote_NFS_Launcher()

## in any case you define a head of the remote party
#LC.remote_head         = 'kerze'

## specify computing nodes
## spec-element: (host-name, number of cores) 
#LC.remote_nodes_spec   =  [('kehrblech',  2),
#                           ('kerze',      2),
#                           ('ketzer',     2),
#                           ('key',        1),                           
#                          ]


## some trouble-shouting on synchronization for launching

## Using user space sync-executables can avoid compatibility problems.
## In general, peewit assumes paths to be identical relative to base-dir
## on local and remote side unless the user says something else
#LC.sync_executable_32         = '../../../rechner/programme/unison-2.27'
#LC.remote_sync_executable_32  = '../../../rechner/programme/unison-2.27'
#LC.remote_sync_executable_64  = '../../../rechner/programme/unison-2.27_64bit'

## unison-metadata-files can get broke (with LC.verbose_sync = 1
## you see unisons messages). In such case you have to delete some of those 
## files, currently by hand. On the local side they are in project-dir/.peewit/.unison, 
## on the remote side in home-dir/.unison


### digest all this now

PK.digest_paths(AR,LC)


### no matplotlib? try
#PathKeeper.use_pylab = 0


