from numpy import linalg, shape, mean, array, sqrt, nanmax,\
     median, nanmin,  isnan
from ml.textprocessing import TextProcessing as TP
from cut.baseclass import Named
from cut.container import indices
from cut.numpytools import nan_std, nan_median, col, nan_mean
from cut.container import sort_A_by_B
#from scipy.stats import gaussian_kde
#from cu.plotting import FunctionPlot

class Prep(Named): 
    """Pre-Processing"""
    
    def train(self,A):
        pass
        
    def apply(self, X):
        return X


class PPC(Prep):
    
    def init(self,
             chain = [ Prep() ]
             ):
        pass
    
    def get_name(self):
        return "+".join([ pp.get_name() for pp in self.chain] )
    
    def train(self, X):
        X_ = X.copy()
        for pp in self.chain:
            pp.train(X_)
            X_ = pp.apply(X_)
    
    def apply(self, X):
        X_ = X.copy()
        for pp in self.chain:
            X_ = pp.apply(X_)
        return X_
        
    

class NanToMedian(Prep):
    
    def train(self,X):
        
        _,m = X.shape
        self.medians = []
        for j in range(m):            
            self.medians += [nan_median(X[:,j]) ]
                

    def apply(self, X):
        
        n,m = X.shape
        X_  = X.copy()
        
        for j in range(m):
            for i in range(n):            
                if isnan(X[i,j]):
                    X_[i,j] = self.medians[j]
        return X_





class S01(Prep):
    
    def center(self,X):
        return nanmin(X,axis=0)
    
    def size(self,X):
        return nanmax( X - self.center(X), axis=0)
    
    def train(self,X):  
        
        n, d    = shape(X)
        mu      = self.center(X)        
        sigma   = self.size(X)
            
        for i in indices(sigma):
            if sigma[i] == 0.:
                sigma[i] = 1. 
                
        self.mu    = mu
        self.sigma = sigma
        self.n     = n
        self.d     = d
        
        
    def apply(self,X):
        
        X2  =  (X - self.mu)/self.sigma 
        return X2


class SPM(S01):
    
    def center(self,X):
        return nan_mean(X)

    def size(self,X):
        return nanmax( abs(X - self.center(X)), axis=0)


class Standardization(S01):
    
    def center(self,X,axis=0):
        return mean(X,axis=axis)
        
    def size(self,X,axis=0):
        return nan_std(X,axis=axis)


class ScalingPM1MD(S01):
    
    def center(self,X,axis=0):
        return median(X,axis=axis)
 

class Norming(Named):
    def norm(self,x):
        n = linalg.norm(x)
        if n == 0:
            return x
        else:
            return x/n 
       

class N(Prep, Norming):
    
    def train(self,X):  
        pass
            
    def apply(self,X):
        X1  =  array( [ self.norm(x) for x in X])
        return X1


class SN( Standardization,Norming):
    
            
    def apply(self,X):
        X1  =  (X - self.mu)/self.sigma 
        X2  =  array( [ self.norm(x) for x in X1])
        return X2
     

class SNN( Standardization,Norming):
            
    def apply(self,X):
        X1  =  (X - self.mu)/(sqrt(self.d)*self.sigma) 
        X2  =  array( [ self.norm(x) for x in X1] )
        return X2
     


class NormalizationPM1( SPM,Norming):
    
            
    def apply(self,X):
        X2  =  (X - self.mu)/self.sigma 
        X1  =  array( [ self.norm(x) for x in X2])
        return X1


class S01N(Norming,S01):
            
    def apply(self,X):
        X2  =  (X - self.mu)/self.sigma 
        X1  =  array( [ self.norm(x) for x in X2])
        return X1


        
    
class Quantiles(Prep):
    
    def init(self,
             min_size=3):
        pass
    
    def get_name(self):
        return  self.classname() + '.' + str(self.min_size)
    
    def train(self, X):
        
        self.X = X.copy()
        #self.X_sorted = X.copy()
        #self.X_sorted.sort(axis=0)
        self.N = len(self.X) * 1.
        
        S =  set(list(self.X.ravel()))
        
        assert all([ not isnan(v) for v in S ] )
        self.X_min = nanmin(self.X, axis=0)
        assert len(self.X_min ) == self.X.shape[1]
        
        
        
    def cdf(self, x, j):
        Xj = self.X[:,j]
        
        n_leq = sum(Xj <= x)
        F     = n_leq / self.N
        return F
    
    def cdf_(self, x, j):
        return self.cdf(x, j)

            
    def apply(self, X):
        
        n,m = X.shape
        X_  = X.copy()
        
    
        for j in range(m):
            if len(set(self.X[:,j])) >= self.min_size:
                for i in range(n):
                    X_[i,j] = self.cdf_( X[i,j], j)
        
        #print "X_", self.get_name()
        #print X_[:5,:8]
        print("DDD")                               
        return X_ 


class QuantilesMP(Quantiles):
    
    def apply(self, X):
        return super(QuantilesMP,self).apply( X) * 2 - 1



class CDF(Quantiles): 


    def ecdf(self, Xj):
        
        Xj = Xj.copy()
        
        def f(x):
            return 1. * sum( Xj.ravel() <= x ) / len(Xj.ravel())

        return f
    
    def train(self, X):
        
        self.X = X.copy()
        #self.X_sorted = X.copy()
        #self.X_sorted.sort(axis=0)
        n,m = self.X.shape
        
        gdes = []
        
        for j in range(m):
            if len( set(self.X[:,j] ) ) >= self.min_size:
                #xj = array(list(set(X[:,j].ravel().tolist())))
                xj = X[:,j].ravel()
                #print xj
                #gde = gaussian_kde(array(xj))
                gde  = self.ecdf(xj)
                gdes += [gde]
            else:
                gdes += [ None ]
    
        self.gdes = gdes
        
    def apply(self, X):
        n,m = X.shape
        X_  = X.copy()
        
        for j in range(m):
            if len( set(self.X[:,j] ) ) >= self.min_size:
                for i in range(n):
                    #X_[i,j] = self.gdes[j].integrate_box(-inf, X[i,j])
                    X_[i,j] = self.gdes[j](X[i,j])
                    #print self.gdes[j](X[i,j])
        
        return X_




class QuantilesLC(Quantiles):
    """Left Cut quantiles"""
    

    
    def cdf_(self, x, j):
        x_min = self.X_min[j]
        if x <= x_min:
            return 0
        else:
            return self.cdf(x, j)


            
class PhiForText(Named):
    
    def vanilla(self,
                token_sequences,
                n_max_features = 5000):
        
        self.token_set      =  list(TP.get_terms(self.token_sequences))
            
            
        self.send("Number of terms:", len(self.token_set))
        TF    = TP.TF(token_sequences, self.token_set)
        self.idf   = TP.IDF(TF)    
            # simple feature selection: sorted by growing idf 
            # -> sorted by declining df, take first n_max_features
        self.token_set = sort_A_by_B(self.token_set, self.idf)[:self.n_max_features]
        self.idf       = sort_A_by_B(self.idf, self.idf)[:self.n_max_features]
        self.send("Number of terms used:", len(self.token_set))

    
    
    def apply(self,token_sequence):
        tf_vector = TP.tf(token_sequence, self.token_set)
        assert shape(tf_vector) == shape(self.idf)
        out =  tf_vector * self.idf
        return out
 
 

if __name__ == "__main__":
    
    P = CDF()
    
    
    X = col([3,2,5,43,40,45,44])
    
    P.train(X)
    
    def f (x ):
        out = P.apply(array([[x]])).ravel()[0]
        print(out, type(out))
        return out
    def g (x ):
        print(x,P.gdes[0](x))
        return P.gdes[0](x)
    
    
    #fp = FunctionPlot(functions  = [f,g],
    #                  x_interval = [0,50])
    

    #fp.go()

