from cut.mixedtools import frugal_permutation
from cut.container import indices
from cut.baseclass  import Named 
from copy import deepcopy

class SubSampling(Named):
    """Class for sub-sampling from input sequence. It allows 
    stratified sampling with respect to the outputs. 
    
    Though the subset selection is randomized, the original order 
    is kept and can be restored.
    """
    
    def init(self,
                sequence           = None,
                output_sequence    = None,
                restriction        = [], ## list of indices on sequence
                output_restriction = [],
                
                randomize         = True,
                keep_order        = False,
                seed              = None,
                n_folds           = 8,
                n_delivered_folds = None,
                max_train_size    = None,
                max_test_size     = None,                
                ):

        assert list(sequence)
        assert len(sequence) == len(output_sequence)
        assert set(restriction).issubset(indices(sequence))
        assert n_folds <= len(sequence)
        
        self.set_sequence(sequence) 
        
        self.build_equivalence_classes()
        
        if n_delivered_folds is None:
            self.n_delivered_folds = n_folds
                
    def set_sequence(self, sequence):
          
        self.sequence         = sequence 
        self.ordered_indices  = indices(self.sequence)
        self.all_indices      = self.ordered_indices[:]
        self.relevant_indices = self.all_indices
        
        if self.restriction:
            self.relevant_indices = [ i for i in self.all_indices
                                      if i in self.restriction    ]
        if self.output_restriction:
            self.relevant_indices = [ i for i in self.relevant_indices
                                      if self.output_sequence[i] in 
                                      self.output_restriction   
                                    ]
        
        
        #print( [ (self.output_sequence[i], self.output_restriction) 
        #         for i in self.all_indices ])
        if self.randomize:
            n = len(self.relevant_indices)
            random_permutation = frugal_permutation(n,seed=self.seed)
            
            self.relevant_indices = [ self.relevant_indices[i] 
                                      for i in random_permutation ]

            #print(random_permutation)
    
    def get_sequence_from_subset(self, subset):
        """Return items refered by the indices in subset"""
        if self.keep_order:
            subset = self.restore_order(subset)       
        return [ self.sequence[i] for i in subset ]


    def restore_order(self, subset):
        """restore original order in subset"""
        return [ i for i in self.ordered_indices if i in subset ]


    def build_equivalence_classes(self):
        """Build index subsets one for each output of self.output_sequence"""
        
        codomain  = sorted(list( set([ self.output_sequence[i] 
                                for i in self.relevant_indices ] ) ))

        classes   = [ [] for _ in codomain ]
        
        for i in self.relevant_indices:
            classes[ codomain.index(self.output_sequence[i]) ] += [ i ]
        
        self.classes = classes
        
    
    def stratified_partition(self, k):
        """Computes weakly stratified partition with k subset.
        This is done by round robin item distribution such that the 
        number of items from given class differ by at most 1."""
        
        partition = [ [] for _ in range(k) ] 
       
        
        # distribute indices
        for c in deepcopy(self.classes):
            while len(c) > 0:
                taken = c.pop()
                subset = partition.pop(0)
                subset += [ taken ]
                partition.append(subset)
            
        return partition

 
  
    def get_cv_indices(self):
        """Computes cross stratified cross validation folds as list 
        of pairs of list of indices. If the sizes are bound the subsets 
        are not stratified anymore.""" 
         
        subsets = self.stratified_partition(self.n_folds)
        
        folds = []

        for test_part in subsets:
            
            train_part = list(set(self.relevant_indices).difference(set(test_part)))

            if self.max_train_size:
                train_part = train_part[:self.max_train_size]
            if self.max_test_size:
                test_part = test_part[:self.max_test_size]
                
            folds += [(train_part,test_part)]
        
        
        return folds[:self.n_delivered_folds]
        

    def get_cv_folds(self):
        """Computes cross validation folds. If the sizes are bound, 
        the subsets are not stratified anymore.""" 

        indices = self.get_cv_indices()                    
        
        out =  [ ( self.get_sequence_from_subset(A), 
                   self.get_sequence_from_subset(B)  ) 
                   for (A,B) in indices ]
        
        return out
        


if __name__ =='__main__':
    
    pass
    




