"""
stages of data
Data Source part
0) file: csv-like including comments
1) very raw rows: uncsved list of list of strings
2) raw rows: headers removed, missing values as Nones
3) columns: put in column with a view meta infos, numerized numeric values
4) separated output column
Data Tray part
5) columns: missing value replaced somehow, nonimal values murerized somehow,
    both only for input
6) arrays: input, output as 2d-arrays

Copyright (C) 2011 by Lorenz Weizsacker. 
This module is distributed under the GNU General Public License version 3 
(GPLv3), see U{http://www.opensource.org/licenses/gpl-3.0.html}.


Related:
http://astropython.blogspot.de/2011/08/reading-ascii-tables-with-python.html
http://cxc.harvard.edu/contrib/asciitable/
http://www.webwareforpython.org/MiscUtils/Docs/Source/Docs/MiscUtils.DataTable.html
http://penandpants.com/2012/03/09/reading-text-tables-with-python/
http://www.astrobetter.com/atpy-and-asciitable/
"""

import os

from numpy import array, nan, empty, isnan

from cut.fileaccess import load_from_shelve, save_in_shelve, get_lines, \
                          homedir, create_dir, save_file
from cut.container import indices, transpose_list, has_unique_value, distinct_entries
from cut.mixedtools import frugal_permutation
from cut.container import sort_A_by_B 
from cut.baseclass  import Named 

from cut.numpytools import row
from collections import OrderedDict


class RawTable(Named):
    
    def init(self,
             rows = [[]],
             column_header=None,
             row_header=None,):
        
        assert has_unique_value([ len(row) for row in rows ])


    
class Column(Named):
    
    def init(self,
             n_missing_values = 0,
             name = None,
             entry_set = set(),
             width = 0,
             index = None,
             table = None,
             ):
        pass

    def get_value(self,i):
        return self.table.X[i, self.index]
    
    def get_values(self,A):
        return [ self.table.X[i, self.index] for i in A ]
    
    
    
    def is_numeric(self):
        return self.type == 'scalar'
    


class TableReader(Named):
    """Function read reads table from file and puts in a RawTable-object.
    Maybe relaced by something better, see links in module doc-string."""   
    

    def init(self,
             
             path = None,    
             has_column_header = False,
             has_row_header    = False,
             comment_starts =  ['%', '@', '//'],            
             separator      = None,                
              
            ):
        
        pass
    
    def read(self, path=None):
        
        if not path:
            path = self.path
            
        self.path = path
        
        
        self.send('Read raw data from')
        self.send(self.path)

        ## get lines that are not comment nor empty        
        lines  = self.get_data_lines(self.path)

        ## seperate column header
        ## row header seperated later 
        ## sind it can happen that row_header is not represented in column-header
        column_header =None
        if self.has_column_header:
            column_header = lines.pop(0)
        
        splitted_lines = self.get_splitted_lines(lines)

        ##   
        row_header = None
        if self.has_row_header:
            row_header = []
            for line in splitted_lines:
                row_header += [ line.pop(0) ]
        
        
        rows = splitted_lines
        assert rows
        
        self.send('Number of rows and columns:', len(rows), len(rows[0]),lb=2)
        
        rows = [ [ e.strip() for e in row ] for row in rows ]
        
        
        table = RawTable( rows           = rows,
                          row_header     = row_header,
                          column_header  = column_header)
        
        
        return table

              
    def get_data_lines(self,file_path):
        
        def is_data_line(line):
            
            return len(line) > 0 and\
                   not any([line.strip().startswith(s) 
                            for s in self.comment_starts ])
        
        lines_striped =  [ line.strip() for line in get_lines(file_path)]

        lines = [ line for line in lines_striped if is_data_line(line) ]
        return lines

    def get_splitted_lines(self, lines):
        """splits data lines of input file by separator
        which is automatically detected using an heuristic"""
        
        if self.separator is None:
            self.separator = self.detect_separator(lines)
        else:
            assert type(self.separator) is type('')

        splitted_lines = [ line.split(self.separator)
                           for line in lines ]
        
        return splitted_lines
 
    def detect_separator(self,lines):
        
        
        chr2int = lambda x: [chr(i) for i in range(255)].index(x)
        
        favorate_separators = [chr2int(c) for c in [',', '\t', ';'] ]
        
        indices    = list(range(255))
        C_rows     = [ [  line.count(chr(i)) for i in indices ]  for line in lines]
        C_columns  = transpose_list(C_rows)  
        
        candidates = [ i for i in indices 
                          if has_unique_value(C_columns[i]) and C_columns[i][0] > 0  ]
        
        
        #self.send()
        l = len(candidates)
        if l == 0:
            self.send('cant find seperator.',lb=2)
        if l == 1:
            separator = candidates[0]    
        if l > 1:
            # drop alphanumerics from candidate list
            candidates = [ c for c in candidates if not chr(c).isalnum()]  
            l          = len(candidates)
            if l == 0:
                self.send('there are at least two alphanumeric'+\
                          ' seperator candidates, but no other.')
            if l == 1:
                separator = candidates[0]
            
            if l > 1:
                # more than one remaining candidate: 
                # take the first of most frequent ones. If runner-up is exactly 
                # one count less, take runnner-up
                
                # sort candidates descending according to counts
                
                #counts = BetterList([ C_columns[i][0] for i in candidates ])
                #counts_sorting    = counts.get_sorting_permutation(reverse=True)
                #counts_sorted     = counts.get_sorted_list(reverse=True)
                #candidates_sorted = [candidates[i] for i in counts_sorting]
                
                counts            = [ C_columns[i][0] for i in candidates ]
                counts_sorted     = sorted(counts,reverse=True)
                candidates_sorted = sort_A_by_B(candidates,counts)
                
                self.send("Separator candidates: ", [chr(i) for i in candidates_sorted], \
                    ", Indices: ", candidates_sorted, 
                    ", Counts: ", counts_sorted)
                    
                # take most frequent candidate or runner up
                maximizer         = candidates_sorted[0]
                maximum           = counts_sorted[0]
                runner_up         = candidates_sorted[1]
                runner_up_count   = counts_sorted[1]
                 
                if maximum == runner_up_count + 1:
                    # some chars in each column, take runner up    
                    separator = runner_up
                else: 
                    # take most frequent one
                    separator = maximizer
                
                # prefer favorate separators regardless their counts
                favorate_separators.reverse()
                for c in favorate_separators:
                    if c in candidates_sorted:
                        separator = c
                        break
                             
        self.send("detected separator: '%s' (%d)" % (chr(separator), separator))             
        return chr(separator)
      

    
    
    
class TableData(Named):
    """base set is a list of points each of which consist of
    an input and an output part. The joint distribution
    of the points is assumed to be exchangeable such that 
    random sub-sampling makes sense.
    
    Further digests data from a TableReader. It makes more disputable
    pre-processing that partially depends on training_sample. 
    * replace missing numeric values by means on training_sample
    * replace nominals from a set of size k by k disjoint binary values.
        Missing values are represented as k zeros.
        
    TODO: rewrite doc-string
    TODO: ordered type
    
        """
    
    global_datadir = None
    global_build   = True
    global_store   = True
    global_seed    = 1
    
    def get_name(self):
        if self.key:
            return self.key.lower()
        else:
            return super(TableData, self).get_name().lower() 
 

    def init( self,
              key=None,
              
              nominal_columns   = [], ## indices of explicit nominal columns 
              output_index      = -1,
              drob_indices      = [],
              binary_output     = True,
              missing_value     = ['?'], ## list of string, have to match literally (after striping) 
              positives         = None,
              shuffle           = True,
              seed              = None,  ## if None, overwritten by global_seed
              
              build    = None,
              store    = None,   
              datadir  = None,
              basename = None, ## file basename
              source_filename = None,
              source_suffix   = 'csv',
              
              
              columns = OrderedDict(),
              verbose_build  = False,
              
              has_column_header = False,
              has_row_header    = False, 
              show_n            = 10, 
              show_shuffled     = False, 
                    
            ):
        
        
        if datadir is None:
            if not TableData.global_datadir is None:
                datadir = TableData.global_datadir
            else:
                self.bugger("Please specify datadir or TableData.global_datadir")

        

        dir_name = self.get_name()
        
        if not basename:
            basename = self.get_name()
        self.basename = basename
        
        if not source_filename:
            source_filename = basename + '.' + source_suffix
        
        self.source_filename = source_filename 
        
        self.dir_abs =  os.path.join(datadir, dir_name )
        
        
        if seed is None:
            self.seed   = TableData.global_seed     
            
        if build is None:
            build   = TableData.global_build     
        
        if store is None:     
            store   = TableData.global_store 
         
        if build:
            self.build_table()
            if store:
                self.save()
        else:
            self.load()


    def load(self):
        path = os.path.join( self.dir_abs, self.basename + '.pyd')
        self.send('Get data set from\n', path)
        o = load_from_shelve(path) 
        self.set_attributes(o.__dict__) 
        return self


    def save(self):
        create_dir(self.dir_abs)
        path = os.path.join( self.dir_abs, self.basename + '.pyd') 
        self.send('\nSave data set in\n%s\n' % path )
        save_in_shelve(self,path)

    def export_table(self, split_fraction=None):
        out = ""
        
        if split_fraction:
            
            #n = len(self.output_column.entries)
            pass
        
        columns = self.columns + [ self.output_column ]
    
        for i in range(self.n_rows):
            first = True
            for column in columns:
                if not first:
                    out += " "
                entry = column.entries[i]
                out += ' ' * (column.width - len(str(entry)) + 2) + str(entry)
                              
            out += '\n'
        
        source_dir = self.source.dir_abs
        
        create_dir(source_dir)
        path =  os.path.join( dir, self.source.get_name() + '.bsv')
        
        save_file(out,path)


   

    def build_table(self):
        """
        Transforms table from file 2d-float32-array self.X and 1d output array self.Y. 
        Missing numeric values appear as nans in self.X. 
        
        TODO: non-binary outputs
        """
        
        ## read from file
        reader   = TableReader(has_column_header=self.has_column_header,
                               has_row_header=self.has_row_header)
        source_path = os.path.join(self.dir_abs, self.source_filename)
        raw_rows    = reader.read(source_path).rows
        
        
        ## fix permutation
        n = len(raw_rows)
        random_permutation = frugal_permutation(n,seed=self.seed)
        sorting     = sort_A_by_B(list(range(n)),raw_rows)    
        self.permutation = [ random_permutation[i] for i in sorting ]
        
        
        if self.shuffle:
            raw_rows = [ raw_rows[i] for i in self.permutation ]
            
            
        columns = transpose_list(raw_rows)
        output  = columns.pop(self.output_index)
        
        ### inputs 
        
        self.X_raw  = array(transpose_list(columns), dtype=object)
        
        #print self.X_raw
        ## replace missing values by nones
        X_rmv       = self.replace_missing_values( self.X_raw)
        
        ## also detects whether column is numeric or nominal
        self.build_columns(X_rmv) 
        if self.verbose_build:
            self.show_columns()
        
        X_nn    = self.numberize_numbers(X_rmv)
        if self.verbose_build:
            self.show_table_1(X_nn)
        
        X_nnn   = self.numerize_non_numerics(X_nn)
        if self.verbose_build:
            self.show_table_2(X_nnn)
        
        
        self.X    = array(X_nnn, dtype=float)
        
        #print(self.X)
        #for i,x in enumerate(self.X):
        #    if any( [ isnan(v) for v in x.ravel()] ):
        #        print(i,x)

        ### outputs
        
        self.YR  = array(output,dtype=object)
        
        self.build_output_column(self.YR)
        #Y_sign  = self.binary_outputs_as_sign(self.YR)
    
        if self.verbose_build:
            self.show_outputs(self.YR)

                      

    def build_columns(self,T):
        
        _,m = T.shape
        
        self.columns = []
        
        for j in range(m):
            if 1 or not j in self.drob_indeces:
                column =  Column()
                column.index = j
                column.width = 0        
                column.entry_set = distinct_entries(T[:,j].ravel())
                column.table = self
                
                if None in column.entry_set:
                    column.n_missing_values = list(T[:,j].ravel()).count(None)
                
                self.set_column_type(column)
                        
                self.columns += [ column ]
    
    def build_output_column(self,Y):
        
        self.output_column = column =  Column()
        #column.index = 0
        column.entry_set =  distinct_entries(Y.ravel())
    
        #self.set_column_type(column)

    def positive(self, y):
        
        if self.positives:
            #print(self.positives)
            return y in self.positives
        else:
            return y == self.output_column.entry_set[0]
     

    def set_column_type(self, column):
        
        if column.index in self.nominal_columns:
                column.type = 'nominal' 
        else:
            if self.are_numeric(column.entry_set):
                column.type = 'scalar'
            else:
                column.type = 'nominal'
        
         
    def numberize_numbers(self, T):
        
        
        T_ = empty(T.shape, dtype=T.dtype)
        
    
        for j, column in enumerate(self.columns):
            
            if column.is_numeric():
                
                for i in range(len(T)):
                    e = T[i,j]
                    T_[i,j] = self.numerize(e)
                    
            else:
                
                T_[:,j] = T[:,j]
        
        return T_

    def replace_missing_values(self, T):
        
        
        T_ = empty(T.shape, dtype=T.dtype)

        _,m = T.shape
        for j in range(m):

            T_[:,j] = T[:,j]
            for i in indices(T):
                
                if T[i,j] in self.missing_value:
                    T_[i,j] = None
            
        return T_
  

    def are_numeric(self,column_entries):
        
        out = True
        for entry in column_entries:
            entry_ = self.numerize(entry)
            if entry_ == None or isnan(entry_):
                out = False
                break
        
        return out

            
    def numerize(self,s):
        """tries to convert string s into a number"""
        if s == None:
            return nan
        
        ## detect comma numbers
        if s.count(',') == 1:
            l,r = s.split(',')
            ln = self.numerize(l)
            rn = self.numerize(r)
            if type(ln) is int and type(rn) is int:
                s = '.'.join([str(ln),str(rn)])
        
        try:
            v = int(s)    
        except:
            try:
                v = float(s)
            except:
                v = nan   
        
        return v

    def numerize_non_numerics(self, T):    
        
        n = len(T)
        columns = []
        J = 0
        
        T_columns = [] 

        
        for j, column in enumerate(self.columns):
            
            values_pre = T[:,j]
            
            if column.is_numeric():
                column.original_index = j
                T_columns += [ values_pre ]
                columns        += [ column ]
                J              += 1
                
            else:
                value_set = set(values_pre)
                if None in value_set:
                    value_set.remove(None)
                value_set  = sorted(value_set)
                #print(value_set)
                for v in value_set:
                    column_new = Column()
                    values_new = [ 0 ] * n
                    for i in range(n):
                        ## no ones for missing values
                        values_new[i] = int((v == values_pre[i] ))
                        
                    column_new.is_numeric = False
                    column_new.index = J
                    column_new.original_index = j
                    columns += [ column_new ]
                    
                    J += 1
            
                    T_columns += [ values_new ]
            
        self.columns = columns
        
        return array(transpose_list(T_columns))

    
    def n_distinct_outputs(self, sel=None):

        if sel is None:
            sel = indices(self.outputs)

        
        output_set = set([ self.output_column.entries[i] 
                           for i in sel ] )
        n_outputs  = len(list(output_set))

        return n_outputs
          
    
    ## TODO: remove these getters 
    ## instead use table.X[A] directly
    ## and table.Y[A], table.Y_PM[A], Y_EN, ..
    ## Y_PM: pos: Y_EN == 0
    

    
    def get_X(self, sel=None):
        
        if sel is None:
            sel = indices(self.X)

        return self.X[sel]   

    def get_YR(self, sel=None):
            
        if sel is None:
            sel = indices(self.YR)
        
        return self.YR[sel]   

    def get_YB(self, sel=None):
    
        return array([ self.positive(y) for y in self.get_YR(sel)])
        
    def get_YS(self, sel=None):
            
        return self.get_YB(sel) * 2.0  - 1
        
                
    def get_XYR(self, sel):
        return ( self.get_X(sel), self.get_YR(sel) )

    def get_XYB(self, sel):
        return ( self.get_X(sel), self.get_YB(sel) )
    
    def get_XYS(self, sel):
        return ( self.get_X(sel), self.get_YS(sel) )
    
    

    def output_mapping(self,index):

        return self.Y[index]   
     
     

    def remove_duplicates(self, A, contradictions=False):
        
        X,Y = self.get_XYR(A)
        
        D = []
        for i in indices(X):
            for j in range(i+1,len(A)):
                if (X[i] == X[j]).all() and ((Y[i] != Y[j]).all() or not contradictions):
                    D += [A[i]]  
        A_  = [ i for i in A if not i in D]
        
        return A_
  
  

           
    def binary_outputs_as_sign(self, Y):
      
        n = len(Y)
        
        Y_  = Y[:]
        
        if self.binary_output:
            
            value_set = sorted(set(Y))
            
            for i in range(n):
                if   Y[i] == value_set[0]:
                    Y_[i] = -1
                elif Y[i] == value_set[1]:
                    Y_[i] = +1
        return Y_


    
        
    def format_table(self, table, h_lines=[]):
        """like http://foutaise.org/code/texttable/ """  
        
        widths = []
        columns = transpose_list(table)
        
        for column in columns:
            widths += [ max([len(str(x)) for x in column] ) ]
        
        out  = '\n'
        
        for i, row in enumerate(table):
            m = len(row)
            for j in range(m):
                entry = row[j]
                out += ' ' * (widths[j] - len(str(entry)) + 2) + str(entry)             
            out += '\n'
            if i in h_lines:
                out += '  ' + '-' * (sum(widths) + 2*m -2 )  + '\n'
                #with_header = False
                
        return out
        
        
    def format_table_prepare(self,T):
        if self.show_shuffled:
            sel = self.permutation[:self.show_n]
        else:
            sel = list(range(self.show_n))
        T   = T[sel]
        m   = T.shape[1]
        if type(T) is type(empty(0)):    
            T = T.tolist()
        return T,m

    def show_columns(self):
        
        m = len(self.columns)
           
        T = [ list(range(m)) ] 

        data_type     = []
        missing  = []
        distinct = []

        for c in self.columns:
            data_type += [ 'nu' if c.is_numeric() else 'no' ]
            missing   += [ c.n_missing_values ]
            distinct  += [ len(c.entry_set) ]
        
        T += [ data_type, missing, distinct ]

        table = self.format_table(T, h_lines=[0])
        
        self.send()
        self.send("Input Columns:")
        
        self.send(table)    
  
    def show_table_1(self, T):
        
        T, m = self.format_table_prepare(T)
           
        T = [ list(range(m)) ] + T

        table = self.format_table(T, h_lines=[0])
    
        self.send()
        self.send("Considered numbers and missing values:")
        self.send(table)    
        

    def show_table_2(self, T ):
        
        T, m = self.format_table_prepare(T)
        
        header = []
        for column in self.columns:
            header += [ column.original_index ]
        assert len(header) == len(T[0])
            
        T = [ range(m), header ] + T

        table = self.format_table(T, h_lines=[1])
        
        self.send("With binarized nominals:")
        self.send(table)   
     
      
    def show_outputs(self,Y):
        Y = list(Y.ravel())
        
        V = sorted(set(Y))
        self.send("Outputs:")
        H = []
        for v in V:
            h = Y.count(v) * 1./len(Y)
            H += [ '{0:0.3f}'.format(h) ]   
        
        V = sort_A_by_B(V, H, reverse=1)
        H = sort_A_by_B(H, H, reverse=1)
        
        self.send(self.format_table([V,H],h_lines=[0]))
        
        

            
if __name__ == '__main__':
    
    from ml.ml_data import *
    TableData.global_datadir = os.path.join(homedir(), 'code/data_sets_1/')
 
    #ds = cleveland_heart_disease(verbose_build=1)
    #ds = yeast(verbose_build=1)
    
    #print(ds.X)
    ds = vowel(verbose_build=1)
    
    #ds = clean1()
    #ds = parkinsons()
    #ds = vehicle()
    #ds = ionosphere()
    #ds = anneal(binary_output=0)
    #ds = soybean(binary_output=0)
    #ds = vote()
    #ds = sonar()
    #ds = RD2D_A(data_dir=global_datadir)
    #ds = acute()
    #ds = breast_cancer()
    #ds = colic()
    #ds = diabetes()
    
    
    
    
