import sys, re, time, datetime, inspect, os, multiprocessing
from subprocess import Popen, PIPE
from collections import OrderedDict
from copy import deepcopy





def frugal_permutation( n, seed=1):
    """Compute pseudo-pseudo-random permutation using linear 
    congruence generator with multiple rounds. 
    
    @param n: length of computed permutation

    This random permutation generator is meant to be used 
    for instance to select sub-sample from data sets. In such cases 
    we only need 'weak randomness' but want to be able to 
    easily reproduce the output permutation for a given seed. 
    
    For information on LCG see 
    http://random.mat.sbg.ac.at/results/karl/server/node4.html 
    as of 07/26/2012.
    """    
    if seed == None:
        seed = 1
        
    r = 64        ##  number of mixing rounds

    m = 2**64     ## base
    a = 19030423  ## factor, Kolmogorovs date of birth, coprime to base

    
    assert 0 < n and n < m
    assert r > 1, """One round is too weak, it likely to produce 
                    'regular' sequences over incremented seeds."""
    assert type(seed) in [int,int], type(seed)
    assert not seed % m == 0, "This would yield range(n) as output."
    
    
    ## compute r*n 'random' numbers in {0,..,2**64-1} 
    ## chucked in r rows having n entries 
    
    random_rows  = [ None ] * r
    x            = seed
    
    for k in range(r):
        row = [ None ] * n
        for i in range(n): 
            row[i] = x = ( a * x ) % m
        random_rows[k] = row
         
        
    ## reduce each random row to a permutation over {0,..,n-1} 
    ## and apply the permutations iteratively      
    
    output_permutation = list(range(n)) ## initial permutation
    
    for k in range(r):
        permutation         = [ None ] * n
        remaining_elements  = list(range(n))

        for i in range(n):    
            n_ = len(remaining_elements)
            index_next     = random_rows[k][i] % n_        
            permutation[i] = remaining_elements.pop(index_next)
        
        
        output_permutation = [output_permutation[i] for i in permutation ]
        
    return output_permutation






def shell(command_fields,  **opts):
    return shello(command_fields, **opts)

def shello(command_fields, **opts):
    opts['output'] = 'dict'
    return shelloex(command_fields, **opts)['stdout']

def shelle(command_fields, **opts):
    opts['output'] = 'dict'
    return shelloex(command_fields, **opts)['stderr']

def shellx(command_fields, **opts):
    opts['output'] = 'dict'
    return shelloex(command_fields, **opts)['exitst']

def shelloex(command_fields, 
             show='', 
             shell=False, 
             env=None,
             env_extra={},
             output = 'list',
             show_only_nonempty=True,
             show_cmd_on_error=True):
    """call a single command command_fields[0] with args
    command_fields[1:]. Cares for quotation, and escapes.
    """
    
    ## setup env 
    if env or env_extra:
        if not env:
            env = os.environ
        for k in env_extra:
            env[k] = env_extra[k]
    
    ## 
    if type(command_fields) == type(""):
        command_string = command_fields
        command_fields = command_string.split()
    else:
        command_string = " ".join(command_fields)
    
    ## TODO: should we do this?
    command_fields =  [ x.strip() for x in command_fields 
                        if x and x.strip() ]
    
    ## TODO: allow more pipes
    if command_fields.count('|'):
        sep_index = command_fields.index('|')
        command_fields_1 = command_fields[:sep_index]
        command_fields_2 = command_fields[sep_index+1:]
                
        #print(command_fields_1)
        p_1 = Popen(command_fields_1, stdout=PIPE, stderr=PIPE)
        p_2 = Popen(command_fields_2, stdin=p_1.stdout,stdout=PIPE, stderr=PIPE)
        p_1.stdout.close()
        stdout, stderr = p_2.communicate()
        exitst         = p_2.returncode   
        
    else:
        ## run cmd
        p = Popen(command_fields , shell=shell, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        exitst         = p.returncode
    
    ## digest outcomes
    oes = OrderedDict()
    oes['stdout'] = stdout.decode("utf-8")
    oes['stderr'] = stderr.decode("utf-8")
    oes['exitst'] = int(exitst)
    
    ## speak
    
    show_always = not show_only_nonempty 
    if show.count('c') or ( show_cmd_on_error and oes['exitst'] ):
        print('commnd: '  + " ".join(command_fields))
    if show.count('o'):
        if show_always or oes['stdout'].strip():
            print('stdout:\n' + oes['stdout'].strip())
    if show.count('e'):
        if show_always or oes['stderr'].strip():
            print('stderr:\n' + oes['stderr'].strip())
    if show.count('x'):
        if show_always or oes['exitst']:
            print('exitst: '  + str(oes['exitst']))
    
    ## 
    if   output == 'list':
        return oes.values()
    elif output == 'dict':
        return oes
   



    
def timeout_call(method, obj, timeout=600, default_value=None, silent=False):
        
    def g(q,x):
        y = method(x)
        q.put(x)
        q.put(y)
    
    q       = multiprocessing.Queue()
    process = multiprocessing.Process(target=g, args=(q,obj))        
    process.start()
    process.join(timeout=timeout)
    
    if not q.empty():
        obj_ = q.get()
        try:
            for member in obj_.__dict__:
                obj.__dict__[member] = obj_.__dict__[member]
        except AttributeError:
            pass
        y = q.get()  
    else:
        if not silent:
            print('run out of time')
        y    = default_value
        
    process.terminate()

    return y


def timeout_call_2(method, *objects, timeout=600, default_value=None, silent=False,delta=0.1):
    
    def g(inq, outq):
        objects = inq.get()
        y = method(*objects)
        out = list(objects) + [y]
        
        outq.put(out)
        
        
        
        return None
    
    
    iq       = multiprocessing.Queue()
    oq       = multiprocessing.Queue()
    
    p = multiprocessing.Process(target=g, args=(iq,oq), 
                                name=method.__name__)        
    iq.put(objects)
    p.start()
    
    
    ## does not work: always waits timeout
    ## at least when use in peewit
    ## we tried many variants, for instance passing args directly
    ## but p.join always waited till timout
    ## even though result was ready
    #p.join(timeout=timeout)
    
    ## hacky homebrewed timeout
    start_time = time.time()
    while oq.empty():
        now  = time.time()
        time.sleep(delta)
        if now - start_time > timeout:
            break
        
    if not oq.empty():
            
        qout     = oq.get()
        y        = qout.pop()
        objects_ = qout
        finished = True
        out      = (objects_, y, finished)
    
    
    else:
        if not silent:
            print('run out of time')
        y        = default_value
        finished = False
        out      = (objects, y, finished)
    
    
    p.terminate()
    
    return out



def tok(stage=None, key=None, message=None):
    
    delta = tokclock(key=key)
    
    if message is None:
        print('elapsed time, %s: %s' % (stage, seconds2time(delta)))
    else:
        print(message + ' ' + seconds2time(delta) )
    

def tokclock(key=None,restart=False):
    
    if not 'clocks' in dir(tokclock):
        tokclock.clocks = {}
    
    if not 'last_key' in dir(tokclock): 
        tokclock.last_key = key

    if key is None:
        key = tokclock.last_key
     
    new_tick               = time.time()
        
    if key in tokclock.clocks:
        delta                    = new_tick - tokclock.clocks[key]
        tokclock.clocks[key]     = new_tick
    else:
        tokclock.last_key = key
        tokclock.clocks[key]     = new_tick 
        delta = 0.
    
    
    if restart:
        tokclock.clocks[key]     = 0.
    
    
    return delta



def timestamp():
    date_time_raw = str(datetime.datetime.now())
    date_time     = re.sub(r'[:\.\ -]','_', date_time_raw )
    return date_time

def timestemp2():
    return ':'.join( [ '%02d' % x  for x in time.localtime()[:6] ] ) 

def seconds2time(delta):
    """@param delta: float of seconds
    
    Return string representation of delta of form ..h ..m ..s ..ms
    """ 
    pass

    timetuple  = time.localtime(delta)

    ours    = int(delta // 3600)    
    seconds = timetuple[5]
    minutes = timetuple[4]
    milliseconds = int( ( delta * 1000 ) % 1000 )   
    
    out = str(milliseconds) + 'ms'
    
    if seconds:
        out = str(seconds) + 's ' + out
    if minutes:
        out = str(minutes) + 'm ' + out
    if ours:
        out = str(ours) + 'h ' + out
        
    return out



def stems_from(an_object, class_B):
    """Wether the class of an_object is class_B or inherits from class_B.""" 
    
    ## to avoid confusion we forbid type-objects here
    assert not type(an_object) == type, \
    ( "Please provide non-type object as first arg"
      " or use stem_from_2" )
    return stems_from_2(an_object, class_B)


def stems_from_2(x, class_B):
    """Wether x/x' class is class_B or inherits from class_B.
    The latter is tested if x is a type itself.""" 
    
    if type(x) == type:
        class_A = x
    else:
        class_A = type(x)
    
    bases = list(class_A.__mro__)
    bases.append(class_A)
    bases_   = [ ( c.__module__, c.__name__) for c in bases ]
    class_A_ = (class_B.__module__, class_B.__name__)
    
    return class_A_ in bases_

def module_name(full=False):
    """Return caller module name"""
    parent_frame  = inspect.stack()[1][0]
    path          = parent_frame.f_code.co_filename
    base          = os.path.basename(path)
    module_name   = os.path.splitext(base)[0]
    if full:
        return path
    else:
        return str(module_name)

def get_args(f):
    """Assumes that f is instance method and returns the names of the
    mandatory args --excluding 'self'--, those of optionals args, 
    and a dict that maps optional args to default values."""         
    
    all_varnames    = f.__code__.co_varnames
    all_arguments   = all_varnames[:f.__code__.co_argcount]
    defaults        = f.__defaults__
    n_optional      = len(defaults) if defaults else 0 
    n_mandatory     = len(all_arguments) - n_optional   # including 'self'
    mandatory_names = all_arguments[1:n_mandatory]      # exclude 'self'
    optionals_names = all_arguments[n_mandatory:]
    
    
    if defaults:
        defaults = deepcopy(defaults)
        default_dict = dict( list(zip( optionals_names, defaults)))
    else:
        default_dict = {}
    
    
    #print self, mandatory_names, defaults, default_dict
    return mandatory_names, optionals_names, default_dict




def get_cli_arg(number, default=None):
    try:
        out = sys.argv[number]
    except:
        out =  default
    return out



def value2string(value,strf=str):
    try:   
        name = value.get_name()
    except:
        if   type(value) == type(""):
            name =  value
        elif type(value) in [type(0),type(0.)]:
            name = strf(value)
        elif type(value) == type(lambda x:x):
            name = value.__name__
        elif type(value) in  [ type([]),type((None,)) ]:
            name = strf( [ value2string(v,strf=(lambda x:x)) 
                           for v in value ] )
        elif type(value) == type(object):
            name = value.__name__
        else:
            name =  strf(value)
    
    return name


def str2bool(s):
    
    s = str(s)
    
    custom_trues  = ['true', 'yes', 'y', 'si', '1', 'ja', 'on', 'aber hallo',
                     'normal']
    
    custom_falses = [ 'false','no', 'n', 'nein', 'off', 'lieber nicht', 
                      'nix da', '0', 'gehts noch?' ,'']
    
    if s.lower() in custom_trues:
        return True
    
    if s.lower() in custom_falses:
        return False
     
    try:
        return bool(float(s))
    except:
        pass
    
    try:
        return bool(s)
    except:
        pass
    
    return False


if __name__ == '__main__':
    
    
    if 1:
        
        def h(x,y):
            #time.sleep(2)
            return x.pop() 
            
        
        a = ([66,33],set([44,55,66]))
        
        tok(1,1)
        (objects_, y, finished) = timeout_call_2(h, *a, timeout=1)
        tok(2)
        
        print((objects_, y, finished))
        
        tok(3)
    #print(44)
    if 0:
        print(get_args(shell))
        pass
    #if 0:
        #for i in range(1,20):
            #print(permutation(10, seed=i)[:20])
    
        
    if 0:
        for i in range(1,20):
            print(frugal_permutation(64, seed=-i)[:])
    
