"""
Peewit provides archiving services that facilitates restoring older experiments. 
This modules takes care for that by keeping track of the project directory, the 
peewit directory and any other directory you register. 

Each core.Experiment E may have a Archivist A object that makes the
code of the experiment persistent by means of a distributed version control
systems. For each E.run() call, A commits to the registered repos before
the experiment is carried out and stores the resulting repo revision keys 
along with unique key for the run. This allows you to go back to the code
that produced for instance given a plot, if you know, the project, the 
tree-name and the version of the run (the latter two are part of the 
title of the plot). 
    
To make it work you should not lose the repos, nor the history table
stored in project_dir/.peewit/. Be aware that you neither should lose the 
history of the peewit framework itself. In particular you should not 
replace your peewit/ directory with a freshly down-loaded one, since 
then you would also lose the peewit revision keys needed by the archivist. 
Fortunately, there are simple means to  avoid this trap. First, you might 
simply take care not to delete the peewit/.hg directory when saving the new 
peewit version. Second, you can call A.pull_peewit() instead of down-loading
a new version by hand. This function pulls directly from the development 
repo, so you get the most recent version of peewit while the former 
version(s) are kept locally available.    

The proposed utilization is very simple but a bit odd: You call the 
restore method on the archivist on within the experiment-modul, for instance 
AR.restore(version=4). Since that command restores an
older revision of your code, this very module will change as well, and that
line will disappear as soon as you refresh your editor. Are you trapped 
in past now? No. By  adding AR.restore() with no argument, you get back 
to the tip.  


One might include repos (additional to the peewit) that 
exist independent from the current project. On such I{outer repos} we 
may not want to make run-commits. In this case we either make no run-commits at
all, but still read the version to write to history, or, we use two different dvs
on the very same dir (hg and git, we dont know whether hg support mulitple repo on
the smae dir). In the latter case we have to specify which dvs is used for run-commits 
and which as I{outer dvs}.

The case peewit repo itself is special. We do consider it as vanilla dir
with outer-repo. It is not packed into an project export because then one would have 
mulitple peewits when reading an exported project. 
THe peewit repo is thus exluded from run-commit. If you make (relevant) changes two
peewit that are not pushed to the public peewit repo, you cannot make a running export 
of your project anymore. 
    
In order to make the versions valuable we can either
    * ship peewit including the version history
    * get the version history online when older version is needed

If peewit is downloaded without versions, it still has the version number with 
it such that the latter can be used in the project history. 


"""


from cut.baseclass  import Named
from cut.mixedtools import shell 
from cut.fileaccess import save_file, system_path, create_dir, get_content_lines,\
    posix_path, read_file
from collections import OrderedDict

from peewit.paths import PathKeeper


import os, time, hashlib, re, csv
from re import escape as esc 
from os.path import join as opj
from os.path import sep
import shutil
import posixpath
 
 
class Archivist(Named):
    """
    This is the main class for archiving. It keeps track of all runs made,
    and the revision of the registered repositories. In order to use it, 
    register an archivist object to the core.Experiment object E
    as E.archivist. See also doc-string of this module. 
    
    Following the suggestion from U{http://mercurial.selenic.com/wiki/API_Changes} 
    hg is called through command line.
    """
    
    revision_file = '.peewit_revision'
    
    def init(self,
             pathkeeper = None,
             more_dirs  = None,          ## inherit from pathkeeper if None keep own
             exclude_from_archive = [],
             repos_active_integration = [],
             repos_passive_integration = [],

             excluded_suffices   = [], ## excluded from be added automtically
             excluded_dirs       = [], ## to archive of any repo
             omit_peewit_restore = False,
             omit_pyd_files      = True,
             
             ):
        """
        @param pathkeeper: path.PathKeeper object. When self is registered as
            archivist of an core.Experiment object E, pathkeeper will be replaced 
            by a copy of E.pathkeeper.
        @param more_dirs: list of dirs on which peewit should keep track too. 
        @param exclude_from_archive: exclude paths from tracking 
        @param omit_pyd_files: wether to add reep (regular expression exclude pattern)
            to exclude .pyd-files from being added and removed from run-repo automatically.
            This means that new output .pyd-files will never be part of the archive. This 
            way we can keep large files out that contain for instance preprocessed data.
        @param repos_active_integration: list of paths to external repos that 
            should be integrated in archiving. An external repo is an version control repo
            that extist independently of peewit. If a path is listed here,
            every run will commit that external repo. 
        @param repos_passive_integration: Covered external repos that are to be 
            integrated in archiving without making run commits. Commits are done by hand
            but peewit still restores former revisions. Note that these are in general not 
            identical with the state of files for a given run-number, since you might have made 
            changes that have not commited.
            
        Peewit covers paths that are registered in pathkeeper as python_-paths and that are 
        not mentioned in repos_passive_integration or repos_passive_integration automatically
        with in its on repo dir. By now this is .git_peewit. Peewit will not interfere with 
        extsting repo dir unless you mention the path in repos_passive_integration or 
        repos_passive_integration.         
        """   
          
        pass 
    
    ### public functions

    def set_paths(self, pathkeeper):
        """When object is created  most paths are not yet set. This should 
        be called when they are. It is called by pathkeeper.PK.digest_paths."""
        
        self.pathkeeper = self.pk = pathkeeper
        
        pk = self.pathkeeper
        np = posixpath.normpath
        
        self.repo_paths_ni  = []
        
        for group in pk.path_groups_keys_a: 
            self.repo_paths_ni += [ np(p) for p in pk.path_groups[group] ]

        if self.more_dirs is None:
            self.more_dirs  = pk.more_dirs       
        
        
        self.repo_paths_ni += [ np(p) for p in self.more_dirs ]

        self.repo_paths_pi   = [ np(p) for p in self.repos_passive_integration ]
    
        self.repo_paths_ai   = [ np(p) for p in self.repos_active_integration  ]
        
       
        
        
        for p in self.repo_paths_pi:
            assert p in self.repo_paths_ni, \
                'No such dir as %s. ' % p  +\
                'repos_passive_integration must only contain registered paths'   

        for p in self.repo_paths_ai:
            assert p in self.repo_paths_ni, \
                'No such dir as %s. ' % p  +\
                'repos_active_integration must only contain registered paths'   
        
        
        self.repo_paths_ni = [ p for p in self.repo_paths_ni 
                               if not p in (self.repo_paths_pi + self.repo_paths_ai) ]
        
                                        
        assert not set(self.repo_paths_ai).intersection(set(self.repo_paths_pi )),\
               "repos_passive_integration and repos_active_integration must not overlap"
        
        
        self.peewit_path    = np(pk.path_groups['peewit_path'][0])
        self.project_path   = np(pk.path_groups['project_path'][0])

        
        self.repo_paths_pi += [ self.peewit_path ]
        
        excluded   = [ np(p) for p in self.exclude_from_archive ]
        
        for p in excluded:
            assert p in (self.repo_paths_ni + self.repo_paths_pi + self.repo_paths_ai),\
            'Cannot exclude %s. ' % p  + 'No such dir covered by archivist.'  

        self.repo_paths_ni = [ p  for p in  self.repo_paths_ni 
                               if not p in excluded            ]

        
        self.repo_paths_pi = [ p  for p in  self.repo_paths_pi 
                               if not p in excluded            ]

        self.repo_paths_ai = [ p  for p in  self.repo_paths_ai 
                               if not p in excluded            ]

        self.repo_paths_ni = sorted(set(self.repo_paths_ni)) 
        self.repo_paths_pi = sorted(set(self.repo_paths_pi)) 
        self.repo_paths_ai = sorted(set(self.repo_paths_ai)) 
        
        
        if 0:
            print("PI")
            for p in self.repo_paths_pi:
                print(p)
            print("AI")    
            for p in self.repo_paths_ai:
                print(p)
            print("NI")
            for p in self.repo_paths_ni:
                print(p)
            quit()
        
  
    
    def define_repos(self):
        """Define how to deal with different repos."""
        
        pk = self.pathkeeper
        repo_paths = self.repo_paths_ni + self.repo_paths_pi + self.repo_paths_ai
        
        
        self.repos = []

        
        
        for key in repo_paths:
                
            repo            = Repository()
            repo.key        = key
            
            repo.ni         = repo.key in self.repo_paths_ni
            repo.pi         = repo.key in self.repo_paths_pi
            repo.ai         = repo.key in self.repo_paths_ai
            repo.outer      = repo.pi or repo.ai
            repo.tool_dir_rel = opj(*pk.dirs['peewit'])
            
            
            repo.no_commit  = repo.pi

            if repo.key == self.peewit_path:
                repo.path = self.pathkeeper.dir_abs_peewit()
            else:
                repo.path = self.pathkeeper.abs_path(key)
            
            if repo.ni:
                repo.run_dvs = GitPeewit()
                #repo.run_dvs = HgPeewit()
            else:
                repo.run_dvs = self.get_run_dvs(repo)
                

            if repo.pi:
                if repo.has_mdir():
                    repo.ppi = False
                else:
                    revision_file = opj( repo.path, self.revision_file)
                    assert os.path.isfile(revision_file), \
                        "%s cannot be passively integrated by archivist." % repo.key +\
                        "It has no dvs-managment-dir (like .hg) and no file " +\
                        ".peewit_revision either."
                    repo.ppi = True
                    repo.run_dvs = PseudoPI()

            if repo.ni:
                repo.run_dvs.excluded_dirs  = self.excluded_dirs[:]
                repo.run_dvs.excluded_dirs += [ '.svn', '.hg', '.git' , '.unison', 
                                                system_path(HgPeewit.tool_dir_name),
                                                system_path(GitPeewit.tool_dir_name),
                                              ]
                
                repo.run_dvs.excluded_suffices  = self.excluded_suffices[:] 
                repo.run_dvs.excluded_suffices +=\
                                [ 'pyc' ] + ([ 'pyd' ] if self.omit_pyd_files else [])

            if repo.key == self.project_path:
                assert repo.ni
                repo.run_dvs.excluded_dirs +=\
                                [ opj(*pk.dirs[kind]) + os.path.sep  
                                  for kind in pk.dirs.keys() if not kind in ['history'] ]
                
                ## be aware that they change for each run                 
                repo.always_changing = [ pk.history_path_rel_base('run_list'), 
                                         pk.history_path_rel_base('state_list'),]
                
                
            if repo.ni:         
                if not repo.has_mdir():
                    repo.run_dvs.exec_init_inyd(repo)    
                repo.run_dvs.exec_addremove(repo,  self.excluded_dirs,  self.excluded_suffices,)
                
       
            self.repos.append(repo)

        


   
            
    def store_run_iot(self):
        """ Stores run if on tip.
        Called before a run to make the current code state permanent
        such that can be can be reproduced at a later time.
        """ 
        
        
        
        if self.on_tip(): 
                
            run             = self.history_tuple()
            self.run_number = run[-1]
            self.write_to_history('run_list', run)

            for repo in self.repos:
                
                self.send("Visiting repo '%s'." % repo.key, lb=0)
                
                repo_version = repo.run_dvs.get_version(repo)
                    
                changes  = repo.run_dvs.get_changes(repo)
                
                #print("GGG")
                #print(changes)
                if all( [ system_path(repo.key,e[1]) in repo.always_changing
                          for e in changes ]):
                    comment  =  'Only always changing files have changed.' 
                else:
                    comment  = ''
                                
                if changes:
                    if repo.pi:
                        self.send("\nWarning: repo %s has uncommitted changes."  % repo.key, 
                                  "Run may not be reproducible.")
                    else:
                        repo.run_dvs.exec_clean(repo)
                        repo.run_dvs.exec_commit(repo, 
                                                 self.pathkeeper.dir_abs_project(),
                                                 self.pathkeeper.basename() )
                        repo_version = repo.run_dvs.get_version(repo)
                        self.send("Commit repo '%s' %s" % (repo.key, comment) )
                        
                else:
                    self.send("No changes.")
                    
                ## take hash as key
                state   = [ self.run_number, repo.key,
                            repo.run_dvs.__class__.__name__,repo_version ]
                
            
                self.write_to_history('state_list', state)
                        
            self.send('\nRun stored.\n')
        else:
            self.send('Repos are set to past revisions.'+\
                      'Omit storing current states.')


       

    def restore(self, run_number=None, treename=None, version=None, dont=False, 
                      group=None, host=None, user=None, ):
        """Update repos to versions that correpsond to the
        last experiment treename-version-*. 
        
        If no argument is given, it restores the tip.
        restore quits after update and self will not commit.
        """  
        pk = self.pathkeeper
        
        
        if dont:
            return

        self.define_repos()
        
        self.store_run_iot()
        
        if not run_number and not treename and not version:
            return self.restore_tip()
        
        if not group:
            group = pk.group
        
        if not treename and version:
            treename = pk.treename 
        
        ## get run-key-hash for last run having treename-version
        if run_number is None:
            run = self.get_fittest_run(treename, version, group, host, user)
        else:
            run = self.get_run_by_number(run_number)
        
        run_key_hash = str(run[-1])
        
        ## get repo-versions
        states  = self.read_history('state_list')
        
        matches = OrderedDict()
        
        for repo in self.repos:
            matches[repo.key] = []
            self.send("Look for demanded revision in repo %s" % repo.key)

            for entry in states:
                if entry[0] == run_key_hash and entry[1] == repo.key:
                    matches[repo.key] += [ (entry[-2],entry[-1]) ]
         
            if matches[repo.key]:
                self.send('Found %d matching states ' %len(matches[repo.key]) )
                self.send('Chose last one.')
                dvs_name, version = matches[repo.key][-1]
               

                repo.run_dvs = self.get_run_dvs_by_class_name(dvs_name)
                
                
                ## peewit restore can be dangereous while developing peewit
                if repo.key == self.peewit_path and self.omit_peewit_restore:
                    self.send('Omit peewit restore.')
                else:
                    self.send('Restore revision %s' % version )
                    repo.run_dvs.exec_update(repo, version)
                
            else:
                self.send('No matching state found. Use current state.')
                matches[repo.key] = '-1' 
    
        path = self.pathkeeper.history_path_abs('not_tip')
        if not os.path.isfile(path):
            save_file(run_key_hash, path)
    
        self.send('Past repo revisions restored. Quit now.')    
        quit() 

    def get_run_dvs_by_class_name(self,name):
        
        
        assert name in [ 'Hg', 'Git', 'HgPeewit', 'GitPeewit', 'PseudoPi' ]
        
        return eval(name + '()')
       
    def restore_tip(self):
        
        for repo in self.repos:
            if repo.key == self.peewit_path and self.omit_peewit_restore:
                self.send('Omit peewit restore.')
            else:
                repo.run_dvs.exec_update(repo,'tip')
                self.send('Restored tip of repo %s' % repo.key)
            
        path = self.pathkeeper.history_path_abs('not_tip')
        if os.path.isfile(path):
            os.remove(path)
            
        self.send('Back to tip. Quit now.')
        quit()
  

    def on_tip(self):
        """Whether repos are on tip. Not on tip means that past 
        revisions have restored are currently life in the file system."""
        ##TODO: Make real check wether repos are on tip instead?
        path = self.pathkeeper.history_path_abs('not_tip')
        return not os.path.isfile(path)
                  

    def export_project(self, 
                       follow_symlinks=True,
                       root_dir_suffix=None,
                       ):
        """Copies all repo-dirs into a folder and packs an tar-ball 
        from it. It does make exports on repos since we want to include
        the .hg-dirs to pack the whole history in.
        
        @param root_dir_suffix: suffix of dir in which all is copied,
            if None, it is set to dir-name of the project-dir
        @param follow_symlinks: by default symlinks are resolved, their targets
            are exported not the links themself
        """ 
        pk = self.pathkeeper
    
        self.define_repos()
        self.store_run_iot()
        export_name     = root_dir_suffix if root_dir_suffix else \
                          posixpath.split(pk.project_dir)[1]  
        export_dir_name = 'pw_base_dir_export__' +  export_name
        export_dir      = pk.abs_path( export_dir_name)
        
        shutil.rmtree(export_dir, ignore_errors=True)
        create_dir(export_dir)
        os.chdir(pk.abs_path(''))            
        
        for repo in self.repos:
            
            source           = system_path(repo.key)
            target_dir       = opj(export_dir, source)
            target_super_dir = os.path.split(target_dir)[0]
            os.chdir(pk.abs_path(''))            
        
            create_dir(target_super_dir)
            
            self.send('Clone repo %s to\n%s\n' % (repo.key,target_dir))
            ## dont distribute peewit this way
            if not repo.key == self.peewit_path:
                repo.run_dvs.exec_clone(repo, target_dir)
            
        tar_ball_name = export_dir_name + '.tar.gz'
        self.send('Pack tar-ball from project:\n', tar_ball_name)
        os.chdir(pk.abs_path(''))
        if os.path.isfile(tar_ball_name):
            os.remove(tar_ball_name)
        self.shell(['tar', 'czf', tar_ball_name, export_dir_name] )
        
        #target_super_dir = os.path.split(target_dir)[0]
        #os.chdir(target_super_dir)
   
        #command_3 =['tar', 'xcf', target_super_dir + '.tar.gz', target_dir ]
                
        #out_3 = self.shell(command_3)
        
        shutil.rmtree(export_dir, ignore_errors=True)
        
        

    def get_run_dvs(self, repo):
        """Returns agent-handler used for run-commit of repo."""
        
        hg_path  = opj(repo.path, '.hg')
        git_path = opj(repo.path, '.git')
        
        if   os.path.isdir(hg_path):
            return Hg()
        elif os.path.isdir(git_path):
            return Git()
        else:
            return None
              



    def history_tuple(self):
        """
        Builds tuple of values that represent the run. 
        The tuple has three parts:  a seven element _time-stamp_, 
        a _location_ part representing the user, a _path_ part, and finally,
        a part representing the tree. In addition there is hexadecimal hash 
        of length 10 at the end.
        """
        
        pk = self.pathkeeper
        
        ## build the time-stamp
        entry = []        
        year, month, day, hour, min, sec = time.localtime()[:6]
        year = '%04d' % year
        month, day, hour, min, sec = [ '%02d' % part for part in 
                                      (month, day, hour, min, sec) ]
        low   = str(time.time() % 1).split('.')[1][:9]
        
        ## and the rest 
        entry += [ year, month, day, hour, min, sec, low ]
        entry += [ pk.group, pk.host(), pk.user() ] 
        entry += [ pk.base_dir ,pk.project_dir ]
        entry += [ pk.treename,  str(pk.version) ]
        
        ## hash 
        entry      = [ str(part) for part in entry ]
        hash_value =  str(self.hash(tuple(entry)))
        entry      = tuple( entry + [hash_value] )
        
        return entry
    


    def hash(self, key):
        """Computes a hexadezimal hash value of length 10
        from hashable object key"""
        h = hashlib.md5()
        h.update(str(key).encode())
        out = h.hexdigest()[-10:]  ## shorter than hg-hashs
        return out

            

    def has_state_list_entry(self,repo):
        """unused.
        checks wether repository repo has an entry in the history table
        that maps a repo and run key to the revision key of repo."""
        path =  self.pathkeeper.history_path_abs('state_list')
        handle   = open(path, 'r')       
        reader     = csv.reader(handle)
        state_list = list(reader)
        handle.close()
        out = any([ e[1] == repo for e in state_list ] )
        return out
     
     
     
    def m_join(self, *parts):
        sep = '-'
        basename = ''
        for part in parts[:-1]:
            basename += str(part).lower() + sep
        basename += str(parts[-1]).lower()
        
        return basename

    def get_history_lists(self, table_name, group=None, host=None, user=None):
        """Reads all run-tables in the history directory."""
        
                
        pk = self.pathkeeper
        
        history_dir = pk.history_dir_abs()
        
        hist_paths = os.listdir(history_dir)
        
        if  host:
            start = self.m_join(group, host, user) 
        else:
            start = '.*'
            
        pattern = start + '.' + pk.suffices[table_name] + '$'
        
        run_list_files = [ p for p in hist_paths 
                           if re.match(pattern, p)]
    
        lists = {}
        
        for filename in run_list_files:
            
            path = opj(history_dir, filename)
            handle     = open(path, 'r')       
            reader     = csv.reader(handle)
            run_list   = list(reader)
            handle.close()
            
            sep = self.pathkeeper.np_sep
           
            group, host, user = filename.split('.')[0].split(sep)[0:3]
            
            lists[(group, host, user)] = run_list
        
        return lists
        
        
       
    def get_fittest_run(self,treename, version, group, host, user):
        """
        Selects a run from run-list-files, according to the 
        specification given by the args.
        
        This makes a very rough guess in case it cant find a fitting run
        produced on the local machine. 
        """
        pk = self.pathkeeper
        
        if group and host and user:
            prefered_location = (group,host,user)
        else:
            prefered_location = (pk.group, pk.host(), pk.user())
            
        if group and host and user:
            runs = self.get_history_lists('run_list', group, host, user)
        else:
            ## get all of them
            
            runs = self.get_history_lists('run_list')
        
        
        matches = {}
        for location in runs.keys():
            matches[location] = []
            for entry in runs[location]:
                if entry[-2] == version and entry[-3] == treename:
                    matches[location] += [ entry ]
        
        
                
        ##TODO: should take newest one, whether its from local host or not
        if prefered_location in matches and matches[prefered_location]:
           
            self.send('Found %d matching runs from location %s. Chose last one'\
                          % ( len(matches[prefered_location]), str(prefered_location) ) )
            run = matches[prefered_location][-1]
            return run
            
        elif matches:
            ##TODO: actually it is not about hosts, its about ..
            self.send('Could not find matching run from this location. '\
                      'Looking for one from other locations') 
            if prefered_location in matches:
                matches.pop(prefered_location)
            winner = None
            default_timestemp = [0,0,0,0,0,0,0]
            winner_timestemp  = default_timestemp
            for location in matches.keys():
                if matches[location]:
                    run      = matches[location][-1]
                    timestemp = [ int(x) for x in run[0:7] ]
                    if timestemp > winner_timestemp:
                        winner_timestemp = timestemp
                        winner = run
        
            self.send('Found matching runs from %d location. Chose newest one.' \
                      % len(matches))        
            return winner
         
        else:
            self.send('Could not find fitting run at all. Quit now.')
            quit()
                
    def get_run_by_number(self,run_number):
        
        run_number = run_number.strip()
        
        runs = self.get_history_lists('run_list')    
        for v in runs.values():
            for run in v:
                if run[-1] == run_number:
                    return run
        
        self.send('Could not find run %s.' % run_number)
        quit()
            
        

    def read_history(self,table_name):
        
        paths = self.pathkeeper.history_path_abs_all('state_list')
        #path       =  self.pathkeeper.history_path_abs(table_name)
        state_list = []
        for path in paths:
            handle      = open(path, 'r')       
            reader      = csv.reader(handle)
            state_list += list(reader)
            handle.close()
        
        return state_list
    
    def write_to_history(self, table_name, entry):
        """
        @param entry: a list of values
        @param table_name:identifier of a history-file
        
        Write one entry to history table on disk."""       
        path =  self.pathkeeper.history_path_abs(table_name)
        dir  =  os.path.dirname(path)
        
        mode = 'w' if not os.path.isfile(path) else 'a'
        
        if not os.path.isdir(dir):
            create_dir(dir)        
        file = open(path, mode)
        writer = csv.writer(file)
            
        writer.writerows([entry])
            
        file.close()

    @staticmethod
    def pack_peewit():
        """For devellopers. Packs a peewit distribution."""
        pk = PathKeeper(base_dir='code')
        
        key        = posix_path(pk.peewit_dir_rel_base())
        repo       = Repository()
        repo.key   = key
        repo.path  = pk.dir_abs_peewit()
        repo.run_dvs = dvs = Git()
        
        pw_base   = repo.path
        pw_stuff   = opj(pw_base,'stuff')
        
        dev_base  = os.path.split(pw_base)[0]
        
        dist_dir = opj(dev_base, 'distributions')
        dist_pw_base = opj(dist_dir, 'pw_base')
        
        version      = pk.peewit_version()
                
        first_run_script_path = opj(pw_stuff, 'peewit_first_run.sh')
        first_run_script      = read_file(first_run_script_path)
        first_run_script_     = re.sub(r'(DISTRO=peewit)(.*)',r'\1_%s' % version , first_run_script)
        save_file(first_run_script_, first_run_script_path)
        
        
        shutil.rmtree(dist_pw_base, ignore_errors=1)
        
        dvs.exec_clone(repo, dist_pw_base)
        
        revision_id = dvs.get_version(repo)
        
        shutil.rmtree(opj(dist_pw_base, '.git'), ignore_errors=True)
        
        save_file(revision_id , opj(dist_pw_base, Archivist.revision_file))
        
        os.chdir(dist_dir)
        
        tarball = 'peewit_%s.tar.gz' % version
        
        shell(['tar', 'czf', tarball, 'pw_base'])
        shell(['scp', tarball, 'pwtester@localhost:'])
        shell(['scp', first_run_script_path, 'pwtester@localhost:bin/'])
        
        
        
        
    
    @staticmethod
    def pull_peewit():
        """
        Pull peewit repo from peewit server into local working copy.
        
        @attention: this deletes all changes you made on the distributed
            directories.  
        """

        pk = PathKeeper()
        
        
        public_repo = "http://git.code.sf.net/p/peewit/repo"
        
        
        repo_path = pk.dir_abs_peewit()
        
        repo_parent_dir = os.path.split(repo_path)[0]
        
        os.chdir(repo_path)
        

        tool_dir     = opj( repo_path, Git.tool_dir)
        has_git_dir  = os.path.isdir(tool_dir)
        
        
            
        if has_git_dir:                        
            out = shell(['git', 'log', '16dd7823f505b730'])
            is_peewit_copy =  bool(out.count('opening'))
            if not is_peewit_copy:
                pk.bugger('Found .git/ in %s, but that is unrelated to public peewit repo.\n',
                            'Please remove that .git/ and try again') 
        else:
            is_peewit_copy = False
        
        
        if is_peewit_copy:
            command = ['git', 'pull', public_repo , 'master']  
            out     = shell(command)
        else:
            os.chdir(repo_parent_dir)
            pk.send('Clone from \n%s\ninto \n%s' % (public_repo, repo_path) )             
            shutil.rmtree(repo_path, ignore_errors=True)
            command = ['git', 'clone', public_repo, repo_path]
            out     = shell(command)
        
        
        pk.send('Pulled peewit from public repo und run update. Quit now.')
        quit()
             
    

class Repository(Named):
    """Keeps information on registered repositories."""
    
    def init(self,
        key       = None,
        run_dvs   = None,    ## a DVS instance
        reeps_addrem = None, ## reeps for add-remove call
        always_changing     = [],
        addremove_function  = None,
        outer       = False,
        path                = None
        ):
        """
        @param key: posix path to working copy relative to base dir 
        @param dvs: the version control system used by the repo
        @param reeps_addrem: regular expression exclude pattern for 
            addremove calls
        @param reeps: reeps for commit calls
        @always_changing: files that changes on any run and therefore might be
            treaten specifically
        @param addremove_function: individual addremove function called by
            Archivist.commit()
            
        See also Archivist.define_repos.
        """
        pass

    def has_mdir(self):
        if self.run_dvs:
            v = self.run_dvs.get_version(self)
            
            
            m = re.match(r'[\S]{10,12}', v)
            tool_dir      = opj( self.path, self.run_dvs.tool_dir)
            return (os.path.isdir(tool_dir) and m)
                 
        else:
            return False



class DVS(Named):
    """Driver for a distribute version control system."""
    
    def init(self,
             excluded_dirs=[],
             excluded_suffices=[]):
        pass

    def shell(self, cmd, **args):
        return shell(cmd, show_cmd_on_error=False, show='ex', **args)
    
    
class Git(DVS):
    
    ## fatal: bad default revision 'HEAD'
    ## retry or 
    ## -> git reset --hard 'HEAD'
    
    tool_dir_name = '.git'
    tool_dir = tool_dir_name
        
    def cmd_base(self, repo):
        return [ 'git' , '--git-dir=%s' % self.tool_dir, '--work-tree=%s' % repo.path ]    

    def get_version(self, repo):
        """Get the current revision key of repository repo."""
        
        os.chdir(repo.path )
        
        command_1 =  self.cmd_base(repo) + [ 'show', '--format=%H', '-s' ]
        
        output = self.shell(command_1)
        
        target = output.split('\n')[0].strip()
        
        return target

    def get_exclude_pattern(self, excluded_dirs,  excluded_suffices):
        """
        http://www.kernel.org/pub/software/scm/git/docs/gitignore.html
        """
        out = []
        for dir in self.excluded_dirs:
            out += [  posix_path(dir)  ]
        for ending in self.excluded_suffices:
            out += [  '*.' + ending ]

        return out
            

    def get_exclude_arg(self,pattern_list):

        if pattern_list:
            exclude_pattern_arg = "\n".join(pattern_list)
            
            return exclude_pattern_arg
        else:
            return None

    def exec_addremove(self, repo, excluded_dirs,  excluded_suffices, dry_run=False):
        
        pattern_list = self.get_exclude_pattern(excluded_dirs,  excluded_suffices)
        exclude_arg  = self.get_exclude_arg(pattern_list)
        
        info_dir     = opj(repo.path, self.tool_dir, 'info')
        
        create_dir(info_dir)
        
        exclude_path = opj(info_dir, 'exclude')
        
        save_file(exclude_arg, exclude_path)
        
        command = self.cmd_base(repo) + [ 'add', '-A']
        
        os.chdir(repo.path)
                    
        if dry_run:
            command += [ '--dry-run' ]
            
        out_2 = self.shell(command)
        
        return out_2
    
    
    def exec_init_inyd(self, repo):
        """
        Calls hg init on repo_path if not yet done.
        """
        os.chdir(repo.path )
        
        g_dir = opj(repo.path, self.tool_dir)
        create_dir(g_dir)
        dummy_path = opj(repo.path, repo.tool_dir_rel, '.git_dummy')
        
        save_file('', dummy_path)
        
        message = 'start archivist repo'
        
        command   = self.cmd_base(repo) + [ 'init']
        command_2 = self.cmd_base(repo) + [ 'add', dummy_path]
        command_5 = self.cmd_base(repo) + [ 'commit', 
                     '--author'  , 'peewit <pw@beach.fr>',  '-m' , message ]
        
             
        out    = ""
        out   += self.shell(command)
        out   += self.shell(command_2)
        out   += self.shell(command_5)
    
        return out


    def get_changes(self, repo):
        """
        @param repo: repository identifier, given as path relative to base
            directory in posix notation
        @param spec: a Repository object that specifies how to deal with a
            given repository.
            
        Looks for and return changes in repo and return list of pairs p. 
        The first element of p describes the change status of the file given 
        in the second elemnt of p. 
        
        If repo is not an repository, it return the singleton list containing
        an error message. Only works for hg by now.
        """
        
        os.chdir(repo.path )
        
        command_1 =  self.cmd_base(repo) + [ 'status', '--porcelain']
                
        output = self.shell(command_1)
        
        lines = [ line for line in output.split('\n') if line.strip() ]

        entries = [ [line[0], line[1], line[2], line[3:]] for line in lines ]
        
        #for e in entries:
        #    self.send(e)
        
        assert all([ e[2] == ' ' for e in entries]) 
               
        changed = [ (e[1],e[3]) for e in entries 
                    ##TODO: which one is it?
                    if (e[1] in ['M', 'D', 'A'] or e[0] in ['A','D','M'])  ]
            
        return changed


    def exec_commit(self, repo, project, experiment):
        """
        Calls hg commit as user 'peewit'.
        """
        os.chdir(repo.path )
        
        message  = 'peewit-run-commit \n' +\
                   '  project    : ' + project + '\n' +\
                   '  experiment : ' + experiment 
        
        command  = self.cmd_base(repo) + [ 'commit' , '-a']
                  
        command += [  '--author'  , 'peewit <pw@beach.fr>',  '-m' , message ]
             
        out = self.shell(command)
        
        return out


    def exec_clean(self, repo):
        """
        Calls hg commit as user 'peewit'.
        """
        os.chdir(repo.path )
        
        command  = self.cmd_base(repo) + [ 'gc' ]
                  
        
        out = self.shell(command)
        return out


    def exec_update(self, repo, state):
        """Calls hg update."""
        #http://www-cs-students.stanford.edu/~blynn/gitmagic/ch02.html
        os.chdir(repo.path )
        
        command  = self.cmd_base(repo) + [  'checkout' , '-f' ] 
    
        if state == 'tip':    
            command += [ 'master' ]
        else:
            command += [ state ]
        
        self.shell(command)
        

    def exec_clone(self, repo, target_dir):
        """special clone."""
       
        sep = os.path.sep
        os.chdir(repo.path )
        
        
        command  = self.cmd_base(repo) + ['ls-files' ] 
        
        out = self.shell(command).strip()
        
        files = out.split('\n')
        
        
        files_target = [ opj(target_dir, file ) for file in files ]
        
        dirs  = [ os.path.split(f)[0] for f in files_target
                  if f.count(sep) ]

        dirs = set(dirs)
        
        for dir in dirs:
            create_dir(dir)
            
            
        for file in files:
            #print "AAA", file, opj(target_dir, file )
            shutil.copy(file, opj(target_dir, file ))
        
         
        shutil.copytree( system_path( repo.run_dvs.tool_dir ), 
                                      opj(target_dir, repo.run_dvs.tool_dir ), 
                                      symlinks=False )
        
        command_2 = [ 'git' , '--git-dir=%s' % opj(target_dir,self.tool_dir), 
                              '--work-tree=%s' % opj(target_dir), 'gc']
        
        
        out_2 = self.shell(command_2)
        self.send(out_2)
                     

        

class GitPeewit(Git):
    
    tool_dir_name = '.git_peewit'
    tool_dir      = system_path(opj(*PathKeeper.dirs['peewit']), tool_dir_name)    
   
      
      
class Hg(DVS):
    
    tool_dir_name = '.hg'
    tool_dir = tool_dir_name
    
    def get_version(self, repo):
        """Get the current revision key of repository repo."""
        
        os.chdir(repo.path )
            
        hg_command_1 = [ 'hg' , 'id', '-i']
                
        output = self.shell(hg_command_1)
        
        try:
            line = output.split('\n')[0]
            target_match = re.match(r'^([0-9a-f]*).*',line)
            target      = target_match.group(1)
        except:
            if output.count('There is no Mercurial repository here'):
                return '-1'
            else:
                self.send(output)
                self.send(repo.path)
                self.bugger('Failed to get repo version.')
        
        return target

    def get_exclude_pattern(self, excluded_dirs,  excluded_suffices):
        """
        reep:regular expression exclude pattern    
        pattern that exclude from dvs in generall
        """
 
        out = []
        
        for dir in excluded_dirs:
            out += [  '.*' + esc(dir  + sep) + '.*' ]
        for ending in excluded_suffices:
            out += [  '.*\.' + ending ]
        
        return out
            

    def get_exclude_arg(self,pattern_list):
        """Maps a list of REXP pattern a file pattern string 
        that can be passed as value for the --exclude option of hg."""   
       
        if pattern_list:
            pattern_bracketed = [ '(' + p + ')' for p in pattern_list ]
            exclude_pattern_arg  = 're:' + "|".join(pattern_bracketed) 

            return exclude_pattern_arg
        else:
            return None


    def exec_addremove(self, repo, excluded_dirs,  excluded_suffices, dry_run=False):
        
        pattern_list = self.get_exclude_pattern(excluded_dirs,  excluded_suffices)
        exclude_arg  = self.get_exclude_arg(pattern_list)
    
        os.chdir(repo.path )
        
        hg_command_2 = [ 'hg' , 'addremove']
            
        if exclude_arg:
            hg_command_2 += [ '--exclude', exclude_arg ]
            
        if dry_run:
            hg_command_2 += [ '--dry-run' ]
            
        out_2 = self.shell(hg_command_2)
        
        return out_2

    def get_changes(self, repo):
        """
        @param repo: repository identifier, given as path relative to base
            directory in posix notation
        @param spec: a Repository object that specifies how to deal with a
            given repository.
            
        Looks for and return changes in repo and return list of pairs p. 
        The first element of p describes the change status of the file given 
        in the second elemnt of p. 
        
        If repo is not an repository, it return the singleton list containing
        an error message. Only works for hg by now.
        """
        
        os.chdir(repo.path )
        
        #print self.shell(['hg', 'manifest'])
        
        hg_command_1 = [ 'hg' , 'status']
                
        output = self.shell(hg_command_1)
        
        lines = [ line for line in output.split('\n') if line.strip() ]
        
        if lines == ['abort: there is no repo here']:
            return lines
        else:
            entries = [ [line[0], line[1], line[2:]] for line in lines ]
            
            assert all([ e[1] == ' ' for e in entries]) 
               
            changed = [ (e[0],e[2]) for e in entries if not e[0] in ['?', 'C', 'I'] ]
            
            return changed

    
    def exec_init_inyd(self, repo):
        """
        Calls hg init on repo_path if not yet done.
        """
        os.chdir(repo.path )
        
        if not os.path.isdir('.hg'):
            
            hg_command_1 = [ 'hg' , 'init']
                
            out_1 = self.shell(hg_command_1)
            
        else:
            out_1 = None
            
        return out_1


    def exec_pull(self, repo, source, exclude_history=True):
        """
        @param repo_path: absolute path to working copy,
        @param source: source to pull from
        @param exclude_history: wether to exclude history files from beinging
            updated, e.g. deleted
            
        For update of peewit distribution from peewit server. Forces to 
        overwrite with content from server, but history file are saved over.
        """

        os.chdir(repo.path)
        
        if not os.path.isdir(repo.tool_dir_rel):
            exclude_history = False
        
        message = '"peewit commit:  pull from server but save the history files"'
        
        hg_forget = [ 'hg', 'forget', repo.tool_dir_rel]
        hg_commit = [ 'hg', 'commit', '--user'  , 'peewit', '--message', message]
        
        
        hg_pull   = [ 'hg' , 'pull', '--force', source ]
        hg_update = [ 'hg' , 'update', '-C']
        
        hg_add    = [ 'hg', 'add', repo.tool_dir_rel]
        
        
        if exclude_history:
            self.shell(hg_forget)
            #shell(hg_commit)

        self.shell(hg_pull)
        self.shell(hg_update)

        if exclude_history:
            self.shell(hg_add)
            self.shell(hg_commit)
        
        




    def exec_commit(self, repo, project, experiment):
        """
        Calls hg commit as user 'peewit'.
        """
        os.chdir(repo.path )
        
        message  = 'peewit-run-commit \n' +\
                   '  project    : ' + project + '\n' +\
                   '  experiment : ' + experiment 
                  
        hg_command = [  'hg' , 'commit', 
                        '--user'  , 'peewit', 
                        '--message' , message,
                     ]
        
                             
        out = self.shell(hg_command)
        
        return out


    def exec_clean(self, repo):
        pass

    def exec_update(self, repo, state):
        """Calls hg update."""
        
        os.chdir(repo.path )
        
        hg_command  = [ 'hg' , 'update', '--clean' ]
        
        if not state is 'tip':
            hg_command += ['-r'  , state, ]
        
        self.shell(hg_command)
        

    def exec_clone(self, repo, target_dir):
        """Calls hg clone."""
       
        os.chdir(repo.path )
        
        hg_command  = [ 'hg' , 'clone', '--pull', repo.path, target_dir ]
                
        self.shell(hg_command)
     
          

          
        
class HgPeewit(DVS):
    
    tool_dir_name = '.hg_peewit'
    tool_dir   = system_path(opj(*PathKeeper.dirs['peewit']), tool_dir_name )    
    tmp_hg_dir = '.hg_as_it_was'
    
    
    def init(self, backend=Hg()):
        pass
  
    
    def mask_dvs(self, function_name, repo, *args,**kwargs):
        
        
        os.chdir(repo.path)
        
        if os.path.isdir('.hg'):
            shutil.move('.hg', self.tmp_hg_dir)
        
        assert not os.path.isdir('.hg')

        try:
            if os.path.isdir(self.tool_dir):
                shutil.move(self.tool_dir, '.hg')
            out = self.backend.__getattribute__(function_name)(repo, *args, **kwargs)
        except Exception as e:
            print(e)
        finally:
            shutil.move('.hg', self.tool_dir)
            assert not os.path.isdir('.hg')
            if os.path.isdir(self.tmp_hg_dir):
                shutil.move(self.tmp_hg_dir, '.hg')
            
        return out
            
    
    def get_version(self, repo):
        """Get the current revision key of repository repo."""        
        return self.mask_dvs('get_version', repo)
        

    def get_changes(self, repo):
        return self.mask_dvs('get_changes', repo)
        
    
    def exec_init_inyd(self, repo):
        return self.mask_dvs('exec_init_inyd', repo)
 

    def exec_addremove(self, repo,  excluded_dirs,  excluded_suffices, dry_run=False):
       
        return self.mask_dvs('exec_addremove', repo,  excluded_dirs,  excluded_suffices, dry_run=dry_run )


    def exec_commit(self, repo, project, experiment):
        return self.mask_dvs('exec_commit', repo, project, experiment )


    def exec_clean(self, repo):
        return self.mask_dvs('exec_clean', repo)


    def exec_update(self, repo, state):
        return self.mask_dvs('exec_update', repo, state)


    def exec_clone(self, repo, target_dir):
        return self.mask_dvs('exec_clone', repo, target_dir)
     



class PseudoPI(DVS):
    tool_dir_name = ''
    tool_dir = ''
    def get_version(self,repo):
        
        version_path    = opj(repo.path, Archivist.revision_file)
        version_id = get_content_lines(version_path)[0]            
                
        return version_id
     
    def exec_update(self,repo, revision_id):
        
        version_path            = opj(repo.path,  Archivist.revision_file)
        revision_id_from_file   = get_content_lines(version_path)[0]            
        if not revision_id_from_file == revision_id:   
            self.send('The revision-id in %s does not match the demanded revision for %s.'\
                      % ( Archivist.revision_file, repo.key))
            if repo.key is self.peewit_path:
                self.send('You can pull all previous revisions of peewit by archivist.pull_peewit()') 

    def exec_clone(self,repo, target_dir):
                
        shutil.copytree( system_path(repo.path), 
                                    target_dir, 
                                    symlinks=False, )
        
    
    def get_changes(self, repo):
        return []


if __name__ == "__main__":
    
    Archivist.pull_peewit()
    
    
    
