"""
This module implements the core experiment model, p-v-model, applied by peewit. 
This p-v-model fixed the structure of an experiment what lets the machine, 
to some extend understand the structure of the experiment, such  such that it 
can offer basic services to the user without having a clue what the experiment is about.
""" 

import optparse, os, inspect, ast, re
from copy import deepcopy
from time import  strftime
from cut.mixedtools import stems_from, get_args, value2string, tok, tokclock, shelloex, \
    timeout_call_2
from cut.container import indices
from cut.numpytools import is_int
from cut.baseclass  import Named
from collections import OrderedDict
from cut.fileaccess import save_file, create_dir
from pw_toolbox.matching import LTMatcher 
from numpy import empty

from peewit.vcubes import VCube
from peewit.paths import PathKeeper


class Experiment(Named):
    """Top-Level Class for running experiments. In order to define you either, use an complete
    e-tree specification or you start from a parent experiment that already been run and modify its 
    specification."""
    
    def init(self, 
             pathkeeper = None,
             launcher   = None,
             archivist  = None,
             verbose_descent =None,   # whether each descent should be announced
             version         = None,  # number or string naming the version of the e-tree
             partname        = None,  # key for different parts of the experiment.
                                      # call that launches sub-calls of module
                                      # gets key 'main'
             test_run        = False,
             remove_current_files = False,
             node_count_gc   = 0,
             
             input_vcube     = None,
             run_number      = None,
             
             draw_deps       = True,
             
             ):
        """
        @param remove_current_files: wether all files previously produced for given 
            tree-name and version-number should be removed before run.
        @param test_run: whether test_selections of e-nodes are applied for only branching 
            over subset of the descent values. The descent values of the 
            respective e-node are computed in any case."""
             
        self.enodes = []
        
        if test_run:
            self.version = str(self.version) + '__' + 'test'
        

        if pathkeeper:
            assert stems_from(pathkeeper, PathKeeper)
            self.pathkeeper = pathkeeper
        else:
            self.bugger()
        
        if not self.version is None:
            self.pathkeeper.version = str(self.version)
        
        ## set partname
        partname_prefix = str(partname) if partname else ''
            
        p = optparse.OptionParser()
        p.add_option('-p','--part',    action='store', type='string', default='plain')
        p.add_option('-v','--version', action='store', type='string', default='none')
        options = p.parse_args()[0]
        
        self.pathkeeper.partname = partname_prefix + str(options.part)
        

        ## alternate tree-name in case more than one tree is defined
        pass 
        self.pathkeeper.start_logging_again()
        
        if self.archivist:
            self.archivist.pathkeeper = deepcopy(self.pathkeeper)
            
        if self.launcher:
            self.launcher.experiment  = self
            self.launcher.pathkeeper  = self.pathkeeper
        
        
    
    def define(self, etree_raw=None):
        """
        @param etree_raw: specification of an e-tree, a tree of e-node objects;
                        
        Builds the e-tree from etree_raw. Is to be called before self.run
    
        etree_raw uses special tree notation: sub-tree are given as lists. The list elements 
        are all e-nodes. The last element can also be a list of sub-trees. If the last 
        element is also an e-node, the sub-tree is linear. If last elements is a list, 
        of sub-trees, there is a branching. 
        The motivation for this notation is that e-trees are rather linear, e.g. branchings 
        are rare.    
        """
        
        pk = self.pathkeeper 
        
        ## check whether other trees are define in same module
        ## if so, alter tree-name by appending suffix 
        ## such that "one tree one tree-name" holds
        
        if 'treenames' not in Experiment.__dict__:
            Experiment.treenames = set()
        
        treename_base = pk.treename.split('__')[0]
        
        if pk.treename in Experiment.treenames:
            all_numbers = [0]
            for tn in Experiment.treenames:
                parts = tn.split('__') 
                if len(parts) > 1 and is_int(parts[1]):
                    all_numbers += [ int(parts[1]) ]
            next_number = max(all_numbers) + 1
            pk.treename  = treename_base + '__' + str(next_number) 
            
        Experiment.treenames.add( pk.treename )
                
        self.handle_run_number()
        
        self.tree_runner = TreeRunner( self,
                                       etree_raw,
                                       pathkeeper   = pk,
                                       test_run     = self.test_run,
                                       draw_deps    = self.draw_deps,
                                       run_number   = self.run_number )

        if self.launcher:
            self.launcher.complete_axes()
            
    
    
    def handle_run_number(self):
        
        if self.archivist and self.launcher_is_used():
            self.launcher.archivist = self.archivist
        
        ## computing node should not archive, launching node should
        if self.archivist and self.pathkeeper.partname in ['main', 'plain']: 
            self.archivist.define_repos()
            self.archivist.store_run_iot()
            
        ## keep run-number
        if self.archivist and not self.launcher_is_used(): 
            run_number = self.archivist.run_number
            self.pathkeeper.store_run_number(run_number)         
            self.run_number = self.archivist.run_number

    def run(self):
        """
        Runs the experiment. 
        """
        
        pk = self.pathkeeper
                
        ## set partname
        if self.launcher_is_used() and pk.partname == 'plain':
            pk.partname = 'main'
        
        pk.create_project_dirs()
        
        ## logging
        if pk.partname in ['main', 'plain']: 
            pk.start_logging_again()
        else:
            ## launcher redirects output of partial run to file  
            if pk.has('recorder'): pk.stop_logging()
                
        
        self.send('\nRun', pk.basename())
        self.send("Started ", strftime("%m/%d %H:%M:%S"))
        self.send()
        
        
        
        ## remove previous trials if wished    
        if self.remove_current_files and pk.partname in ['main', 'plain']:
            if not self.launcher_is_used():
                pk.remove_current_files()
            else:
                if self.launcher.dismiss:
                    ## in this case we first want to dimiss an 
                    ## then remove the current files
                    ## (if not remove_current_files but launcher.dismiss
                    ##  then launcher will later call dismiss_launch() itself) 
                    self.launcher.dismiss_launch()
                    
                    pk.remove_current_files()
                    self.launcher.update_project_dir()
                    
                    self.send('\nLaunches dismissed. ' +\
                              'Quit experiment module after run().')
                    os._exit(0)
                else:
                    ## otherwise we first remove the current files
                    ## an then continue from scratch
                    pk.remove_current_files()
                    self.launcher.update_project_dir()
                

        if self.launcher_is_used():
            vc = self.run_launch()
        else:
            ## just do it and return to experiment module    
            vc = self.run_vanilla()
        
        
            
        return vc
        
        
    def run_launch(self):
            
        assert self.launcher_is_used() 
        
        pk       = self.pathkeeper
        partname = pk.partname 
                
        pk.create_launch_dirs()
        pk.save_pid_file()
    
        ## two cases: launch management or launch execution
        if  partname == 'main':
            ## no run, just launch management
            vc =  self.launcher.main()
            if vc is None:
                os._exit(0)
            return vc
        else:
            ## now in a subprocess, do the thing
            
            self.tree_runner.set_d_selection()  ## maps partname to d-selection
            
            vc = self.run_vanilla()
            
            self.launcher.finish_part(vc)
                
            self.send('Experiment was called by launcher. Quit main module after run().')
            os._exit(0)
            
        
        
    def run_vanilla(self):
        
        if self.launcher_is_used():
            save_file('running', self.pathkeeper.path_abs('state'))
        
        
        tokclock('tree-run')      
          
        vc = self.tree_runner.run()  
    
        tok(key='tree-run', message='\npv-tree build time:')        
        
        return vc
      
    def launcher_is_used(self):
        return self.launcher and\
               self.launcher.launch_specified and\
               not self.launcher.dont

    
    def get_node(self, spec):
        """Get first e-node with e-name spec or, if spec is list,
        first e-node named spec[-1] with e-nodes names spec[i] on the
        on the way from root to that target."""
        
        assert type(spec) in [ list, type('')]
        
        if type(spec) == type('jo'):
            path   = [spec]
        else:
            path   = spec
        
        
        enodes = self.tree_runner.enodes
        
        assert path 
        while path:
            enames = [ e.ename for e in enodes ]
            target_name = path.pop(0)
            assert target_name in enames, "No e-nodes named %s" % target_name
            index  = enames.index(target_name)
            target = enodes[index]
            enodes = enodes[index+1:]

        return target
        
    

class TreeRunner(Named):

    def init(self,
             experiment,
             etree_raw,
             run_index  = None, #
             pathkeeper = None,
             test_run   = False,
             draw_deps  = False,
             run_number = None,
             ):
        
        self.pathkeeper =\
        self.pk         = experiment.pathkeeper
        
        

        ## digest etree_raw        
        self.build_e_tree(etree_raw) 
        self.build_relations()
        
        ## set input_names, input_names_all
        self.collect_input_names()
        ## set input_node_map: ename -> enode
        self.build_input_node_maps()
        
        ##  set input_nodes_pre, input_nodes_post, input_nodes_pre
        self.settle_input_nodes()
        
        ## recursively dependencies_descent, dependencies_ascent
        self.build_dependencies()
        
        if self.draw_deps:
            self.draw_dependencies()
         
        self.init_vcubes()
        
        self.read_input_pvcs()
        
        
                        
        self.pk.target_ename = self.enodes[-1].ename \
                               if self.pk.target_ename is None \
                               else self.pk.target_ename
                               
        self.experiment.enodes = self.enodes
        self.experiment.eroot  = self.eroot
        self.experiment.tree_runner = self
        
                               
        ## apply selection of output values only in test runs
        if self.test_run:
            for node in self.enodes:
                node.descent_selection = node.test_selection
                                       

        for f in ENode().production_functions_all():  
            fname = f.__name__
            aname = 'verbose_' + fname
            if not self.experiment.has(aname):
                self.__setattr__(aname, None)
            else:
                verbose_nodes = self.experiment.__getattribute__(aname)
                enames = [ e.ename for e in self.enodes]
                self.__setattr__(aname, verbose_nodes)
            assert verbose_nodes is None or\
                   ( type(verbose_nodes) is list and \
                     all([ vn in enames for vn in verbose_nodes]) ),\
                        "Experiment.%s should be list of e-names not this: %s"\
                        % ( aname, verbose_nodes)
                    
  
    def set_d_selection(self):
        """
        partname is a string representing a slide of an experiment,
        the format is ename1__index1___ename2__index2_..
        where an index in as index of the respective descent values.
        This string is mapped to dict e-name -> list of indices of descent value
        which subsequently is used to set the d-selection that is the 
        selection of descent indices that are followed. 
        
        """
        
        partname = self.pathkeeper.partname
        
        assert not partname in ['main','plain']
        
        outer_parts = partname.split('___')
        
        part_dict = {}
        
        for outer in outer_parts:
            inner_parts = outer.split('__')
            ename       = inner_parts[0]
            index_raw   = inner_parts[1]
            
            part_dict[ename] = int(index_raw)

            assert str(part_dict[ename]) == index_raw
        
        ## set d-selections
        for ename, index in part_dict.items():
            for enode in self.enodes:
                if enode.ename == ename:
                    enode.descent_selection = [ index ] 
    

    ### building of the e-tree
        
    def build_e_tree(self, cube_tree_raw):
        """Recursive build of the e-tree. It becomes a bit complicated 
        as we want to allow v-cubes as element of the raw e-tree. They 
        have to expanded to corresponding e-nodes."""
        
        def build_et_rec(node, rest, index):
            
            ## vcube-case: inserts sequence of e-nodes obtained from vcube
            if type(node) is VCube:                
                if not node.has('eparent'):
                    node.eparent = None
                inserts = self.enodes_from_vcube(node)
                if inserts:
                    node, rest_ = inserts[0], inserts[1:] 
                    rest        = rest_ + rest
                else:
                    self.bugger("Inserted v-cube has no non-ancestor dim.")
            
            self.enodes    += [ node ]
            node.runner     = self
            node.index      = index
            index          += 1
            node.experiment = self.experiment
        
            ## recursive build of etree
            if rest:
                head, rest = rest[0], rest[1:]
                if   stems_from(head ,ENode) or stems_from(head ,VCube):
                    ## single subtree with head as root
                    subtrees = [ [ head  ] + rest ]
                elif type(head) is list:
                    subtrees = head
                else:
                    self.bugger("Invalid e-tree description. Element of raw e-tree " +
                                "either is e-node or list of sub-tree-descriptions. " +
                                str(type(head)) )
                    
                node.echildren  = []
                left_esibling    = None
                
                for subtree in subtrees:
                    
                    subroot            = subtree[0]
                    assert stems_from(subroot ,ENode) or stems_from(head ,VCube),\
                        "Subtree description demands e-node as head, not %s." % str(type(subroot))
                    
                    subroot.eparent     = node  ## this we set now since in case of vcube we have to
                                                ## to know ancestors
                    
                    subroot = build_et_rec(subroot, subtree[1:], index) 
                    
                    node.echildren     += [ subroot ] 
                    subroot.eparent     = node
                    subroot.left_esibling = left_esibling
                    if not left_esibling is None:
                        left_esibling.right_esibling = subroot
                    left_esibling         = subroot
            
            return node ## in case node was v-cube it now refers to an e-node
                    
        head, tail  = cube_tree_raw[0], cube_tree_raw[1:]
        self.eroot  = head
        self.enodes = []
        build_et_rec(head, tail,0)
        
        return self.eroot


    
    def enodes_from_vcube(self, vcube):
        ## TODO: allow renaming of cube dims themself, e.g. set enames of inserted nodes
        ## TODO: allow test/descent-selection 
        
        if vcube.eparent:
            ancestors = [ vcube.eparent ] + vcube.eparent.ancestors_t()
        else:
            ancestors = []
            
        ancestors_names = [ e.ename for e in ancestors ]
        
        
        ## apply input-refs
        if vcube.has('input_refs'):
            input_refs = OrderedDict()
            input_refs.update(vcube.input_refs)
            
            for vcube_ in vcube.vcubes.values():
            
                vcubes_ = OrderedDict()

                for ename, vcube__  in vcube_.vcubes.items():
                    
                    if ename in input_refs:
                        ename_ = input_refs[ename]
                        assert ename_ in ancestors_names
                        vcube__.target_ename = ename_
                        vcubes_[ename_] = vcube_.vcubes[ename] 
                    else:
            
                        vcubes_[ename]  = vcube_.vcubes[ename]
                
                vcube_.vcubes = vcubes_
            

        outer_dims = [ en for en in vcube.dim_names() if     en in ancestors_names ]
        inner_dims = [ en for en in vcube.dim_names() if not en in ancestors_names ]
        
        inserts = []
        
        for ename in outer_dims:
            enode = ancestors[ancestors_names.index(ename)]
            enode.pvc    = vcube.vcubes[ename]
            
        
        for ename in inner_dims:
            
            enode = VCNode( ename=ename,
                            #input_refs = vcube.input_refs,
                          )
            
            vc = vcube.vcubes[ename]
            
            
            enode.pvc    = vc  ## enode.dvc is build by deploy base on pvc
            ##                     ## enode does not make ascent
                                        
            inserts += [ enode ]
            ## might be set latter for better?
            
            for f in enode.production_functions_all():
                enode.input_names[f] = []
                
            enode.input_names[enode.descent]       =\
            enode.input_names_all      = vc.dim_names_non_target()
            
            
        
            assert set(enode.input_names_all).union(set([enode.ename])) ==\
                   set(enode.pvc.dim_names()),\
                ( enode.ename, enode.input_names_all, 
                 [enode.ename], enode.pvc.dim_names() )
            
        if inserts:
            inserts[0].eparent = vcube.eparent
        del vcube.eparent
        
        return inserts
       
       
        
    def read_input_pvcs(self):
        """Integrate pre-computed v-cubes given as pdvc or pavc."""
        
        def add_pvc(enode, pvc, mode):
        
            assert enode.ename == pvc.target_ename
                
            vc = enode.dvc if mode == 'descent' else enode.avc
            
            for ename_pvc in pvc.dim_names():
                assert ename_pvc in vc.dim_names()  or pvc.dim_sizes(ename_pvc) == 1, \
                "pre-cube of e-node %s has unknown non-singleton dimension %s" % \
                (enode.ename, ename_pvc)
            
            if mode == 'descent':
                enode.pdvc = pvc
            else:
                enode.pavc = pvc
            
            for ename_, pvc_ in pvc.vcubes.items():
                if not ename_ == pvc.target_ename:
                    enode_  = self.get_input_node(enode, ename_) 
                    mode_   = self.get_input_mode(enode, enode_)
                    add_pvc(enode_, pvc_, mode_)
        
                
        def traverse(enode):
            if enode.pdvc:
                add_pvc(enode, enode.pdvc, 'descent')
            for enode_ in enode.echildren:
                traverse(enode_)
            if enode.pavc:
                add_pvc(enode, enode.pavc, 'ascent')
        
        traverse(self.eroot)
        
        
     
        
    def build_relations(self):
        """build mapings enode_ -> relation type"""
         
        for a,enode in enumerate(self.enodes):
            
            enode.e_relations = e_relations = OrderedDict()
            
            for b,enode_ in enumerate(self.enodes):
                
                assert not enode is enode_ or b is a,\
                    "Do not use a node twice a tree." 
                
                if b < a:
                    if enode_ in enode.ancestors_t():
                        e_relations[enode_] = 'ancestor_t'
                    else:
                        e_relations[enode_] = 'ancestor_s'
                elif a == b:
                    e_relations[enode_]     = 'self'
                else:
                    if enode in enode_.ancestors_t():
                        e_relations[enode_] = 'successor_t'
                    else:
                        e_relations[enode_] = 'successor_r' 
                
            

    
    def collect_input_names(self):
        """For each production function collect input e-names and apply input-refs
        on them. 
        The fields 
            input_names,
            input_names_all, 
            input_defaults  
        are set for all (non-vc) e-nodes here."""
        
        for enode in self.enodes: 
            if stems_from(enode, VCNode):
                pass
            else:##
                input_mandatory      = OrderedDict() ## key: production-function
                input_optional       = OrderedDict()
                input_defaults       = OrderedDict()
                enode.input_names    = OrderedDict()
    
                input_names_all        = set()
                            
                for f in enode.production_functions_all():
                
                    input_mandatory[f],\
                    input_optional[f], \
                    input_defaults[f]  = list(get_args(f)[0]),\
                                         list(get_args(f)[1]),\
                                              get_args(f)[2]
                    ## input_defaults != enode.input_defaults
                    
                                              
                    ## refer to this one for keys
                    enode.input_names[f] = input_mandatory[f] +\
                                           input_optional[f]
                    
                    
                    input_names_all.update(set(enode.input_names[f]))
                    
                    
                    if f in [ enode.initial_siblings ]:
                        assert not enode.ename in enode.input_names[f],\
                            "%s.%s() cannot use sibling-values itself" %\
                            (enode.ename, f.__name__)
                                        
                    self.include_defaults( enode, input_defaults[f] )        
                
                del input_defaults ## to make clear that its not enode.input_defaults
                
                ## make it a list for applying input_refs
                enode.input_names_all = list(input_names_all)

                ## check orig. names in input_refs exist    
                for ename_orig,ename_new in enode.input_refs:
                    assert ename_orig in enode.input_names_all,\
                        ("%s.input_refs: '%s' is not used as an argument " + \
                            "of any production function of %s" )\
                            % (enode.ename, ename_orig, enode.ename)
                
                
                for f in enode.production_functions_all():
                
                    ## apply input-refs
                    for ename_orig, ename_new in enode.input_refs:
                        assert ename_new in [e.ename for e in self.enodes],\
                               "%s.input_refs: unknown input node '%s'" %\
                               (enode.ename, ename_new)
                               
                        for ename_list in [ input_mandatory[f],
                                            input_optional[f],
                                            enode.input_names[f],
                                            enode.input_names_all
                                             ]:
                            if ename_orig in ename_list: 
                                ename_list[ename_list.index(ename_orig)] = ename_new
                        
                        ename_dict = enode.input_defaults
                        if ename_orig in list(ename_dict.keys()): 
                            ename_dict[ename_new] = ename_dict.pop(ename_orig)
                
                    
                    ## check whether input exists
                    for ename in input_mandatory[f]:
                        assert ename in [ e.ename for e in self.enodes ],\
                            "%s.%s(): unknown input node '%s'" %\
                            (enode.ename, f.__name__, ename)
                    
                ## and back to set 
                ## referred in build_input_node_maps
                enode.input_names_all = set(enode.input_names_all)
    
 
                
        
    def include_defaults(self, enode, input_defaults):

        ## check consitency of default values         
        for ename,value in input_defaults.items():
            if ename in enode.input_defaults:
                try: 
                    if enode.input_defaults[ename] != value:
                        self.bugger("%s: please use same default value for '%s' " + 
                                    "for all production functions" %
                                    (self.ename, ename))
                except:
                    self.send("%s: cannot check equality of default values of '%s' " + 
                              "in different production functions. " %
                              (self.ename, ename))
            else:
                enode.input_defaults[ename] = value

   
    
             
    def build_input_node_maps(self):
        """Here we fix for all nodes to what node an e-name refers. 
        The resulting mapping enode.input_node_map does not depend 
        on the production-function. """
        
        for a,enode in enumerate(self.enodes):
            
            ancestors       = list(reversed( self.enodes[:a] ))
            ancestors_names = [ e.ename for e in ancestors ]
            
            successors_t    = list(reversed(enode.successors_t()))
            successor_names = [ e.ename for e in successors_t ]
            
            reachable_names = [ enode.ename ] + ancestors_names + successor_names

            enode.input_node_map  = OrderedDict()
            
            for ename in enode.input_names_all:
                if ename in reachable_names:
                    
                    ## TODO: ??
                    if ename in enode.input_defaults: 
                        enode.input_defaults.pop(ename)
                        
                    if ename == enode.ename:
                        enode_ = enode
                    elif ename in ancestors_names:
                        i = ancestors_names.index(ename)
                        enode_ = ancestors[i]
                    elif ename in successor_names:                        
                        i = successor_names.index(ename)
                        enode_ = successors_t[i]
                        
                    enode.input_node_map[ename] = enode_
            
            #if not enode.ename in enode.input_node_map:
            enode.input_node_map[enode.ename] = enode
            
            
            
    def settle_input_nodes(self):
        """Collect input-nodes and puts them into groups."""
         
        for enode in self.enodes: 
            
            #print("\nFFF " + enode.ename)
            for f in enode.production_functions_all():
                #print(f.__name__)
                input_nodes_f = set([ enode.input_node_map[en] 
                                      for en in enode.input_names[f]
                                      if not en in enode.input_defaults
                                      ])
                
                #print([e.ename for e in input_nodes_f])
                        
                if f in enode.production_functions_pre():         
                        enode.input_nodes_pre.update(input_nodes_f)
                if f in enode.production_functions_post():
                        enode.input_nodes_post.update(input_nodes_f)
                
                enode.input_nodes.update(input_nodes_f)
                
            enode.input_nodes_pre_direct  = enode.input_nodes_pre.copy() 
            enode.input_nodes_post_direct = enode.input_nodes_post.copy() 
            #print([e.ename for e in enode.input_nodes_pre])
                   

    def build_dependencies(self):
        """Collect dependencies for descent and ascent as sets of e-nodes."""
        
        def get_indirect_deps(enode, enode_):
            cube_type        = self.get_input_mode(enode, enode_)
            
            if   cube_type == 'descent' :
                indirect_deps  = [ e for e in enode_.dependencies_descent 
                                   if  e in enode.ancestors_t()  ]
            elif cube_type == 'ascent' :
                indirect_deps  = [ e for e in enode_.dependencies_ascent 
                                   if  e in enode.ancestors_t()  ]
            else:
                indirect_deps  = []
            
            return set(indirect_deps)
        
        
        def traverse(enode):
            
            ## direct dependencies
            enode.dependencies_descent = enode.input_nodes_pre.copy()
            enode.dependencies_ascent  = enode.input_nodes_post.copy()
            
            ## indirect dependencies descent            
            for enode_ in enode.input_nodes_pre:
                indirect_deps_d = get_indirect_deps(enode, enode_)
                #print("III")
                #print(enode.ename)
                #print(enode_.ename)
                #print([ e.ename for e in indirect_deps_d])
                if enode_ in enode.input_nodes_pre:
                    enode.dependencies_descent.update(indirect_deps_d)
                if enode_ in enode.input_nodes_post:
                    enode.dependencies_ascent.update(indirect_deps_d)
                
            for enode_ in enode.echildren:
                traverse(enode_)

            ## indirect dependencies ascent            
            for enode_ in enode.input_nodes_post:
                indirect_deps_d = get_indirect_deps(enode, enode_)
                if enode_ in enode.input_nodes_pre:
                    enode.dependencies_descent.update(indirect_deps_d)
                if enode_ in enode.input_nodes_post:
                    enode.dependencies_ascent.update(indirect_deps_d)
            
            #ancestors = enode.ancestors()
            #enode.dependencies_descent  = enode.dependencies_descent.intersection(ancestors)
            #enode.dependencies_ascent   = enode.dependencies_ascent.intersection(ancestors)
            
            enode.dependencies_descent  = self.sort_enodes(enode.dependencies_descent)        
            enode.dependencies_ascent   = self.sort_enodes(enode.dependencies_ascent)        
            
            #self.show_dependencies(enode)
            
        traverse(self.eroot)
        
        self.check_dependencies()
        
    
    def check_dependencies(self):
        
        ## check validity of dependecies
        for enode in self.enodes:
            
            ancestors_  = list(set(enode.ancestors()).union(set([enode])))
            ancestors__ = list(set(ancestors_).union(set(enode.successors_t())))
            
            assert set(enode.dependencies_descent).issubset(ancestors_), \
                        (enode.ename, enode.dependencies_descent, ancestors_  )
            
            assert set(enode.dependencies_ascent).issubset(ancestors__), \
                        (enode.ename, enode.dependencies_ascent, ancestors__  )
            
            
        ## check vc-nodes
        for enode in self.enodes:
            if stems_from(enode, VCNode):
                assert set([en.ename for en in enode.dependencies_descent]) ==\
                       set(enode.pvc.dim_names_non_target()),\
                    (enode.dependencies_descent, enode.pvc.dim_names())
        
        
        ## make sure that pre-productions do not demand successors
        for f in enode.production_functions_pre():
            for enode_ in self.enodes:
                assert not ( 
                    enode.e_relations[enode_].startswith('successor') and 
                    enode_ in enode.dependencies_descent ),\
                    "%s.%s(): cannot access '%s'"  %\
                    (enode.ename,  f.__name__,  enode_.ename) 
             
    
    def draw_dependencies(self):

        ##    size="3,2";
        ##         color="pink";
        ## ordering=out;
    
        
        header = """
digraph dependencies {
    rankdir=BT;
    node [shape=box]
    layout=dot;
    ranksep=1;
"""
        pk = self.pathkeeper
        title   = pk.treename + '__' + pk.version 
        
        
        if self.run_number: 
            title_  = '< <BR/><FONT POINT-SIZE="10">%s %s %s</FONT> >' % \
                        (title, " " * 5, self.run_number)
        else:
            title_  = '< <BR/><FONT POINT-SIZE="10">%s</FONT> >' % title
        header += '\n\tlabel=%s;' % title_
                
        nodes = ''
        N = len(self.enodes)
        
        rank_classes = OrderedDict()
        for enode in self.enodes:
            rank = N-enode.depth()
            if not rank in rank_classes:
                rank_classes[rank] = []
            rank_classes[rank] += [enode.ename]
        
        for enode in self.enodes:
            nodes += '\n\t%s [label="%s"];' %(enode.ename, enode.ename)
            #nodes += '\n\t{rank=%d %s}' % (enode.depth(), enode.ename)
            if not enode.echildren:
                nodes += '\n\t{rank=min %s}' % enode.ename
            else:
                nodes += '\n\t{rank=same %s}' % ' '.join([e.ename for e in enode.echildren])
            
        #for rank, enodes in rank_classes.items():
            #print(enodes)
            #nodes += '\n\t{rank=same %s}' % ' '.join(enodes)
        #    pass
        
        edges_descent = '\nedge [color="#20a020"];'
        for enode in self.enodes:
            for enode_ in enode.input_nodes_pre_direct:
                if enode_ in enode.ancestors_t():
                    edges_descent += '\n\t%s -> %s;' % (enode.ename, enode_.ename)                
                else:
                    #edges_descent += '\n\t%s -> %s [style="dashed"];' % \
                    edges_descent += '\n\t%s -> %s [color="#20a020:#20a020" style="solid"];' % \
                        (enode.ename, enode_.ename)
                
        
        edges_ascent = '\nedge [color="#a02020"];'
        for enode in self.enodes:
            for enode_ in enode.input_nodes_post_direct:
                if enode_ in enode.ancestors_t():
                    edges_ascent += '\n\t%s -> %s;' % (enode.ename, enode_.ename)
                else:
                    edges_ascent += '\n\t%s -> %s [color="#a02020:#a02020" style="solid"];' % \
                        (enode.ename, enode_.ename)
        
        footer = '\n}\n'

        ## write gv
        gv_content = header + nodes + edges_descent + edges_ascent + footer
        dir_gv   = self.pathkeeper.dir_abs_pdir('deps_gv')
        path_gv  = self.pathkeeper.path_abs('deps_gv', partname='e_nodes')
        
        ## plot 
        output_format = 'pdf'
        assert output_format in [ 'pdf', 'svg' ]
        if not os.path.isdir(dir_gv):
            create_dir(dir_gv)
        save_file(gv_content, path_gv)
        path_out = self.pathkeeper.path_abs('deps_' + output_format, partname='e_nodes')
        try:
            shelloex(['dot', '-T' + output_format, '-o' + path_out, path_gv],show='oe')
        except FileNotFoundError:
            self.send("graphviz not installed. Please install graphviz for plotting dependency"
                      + " graphs or add 'draw_deps=False' to Experiment-constructor.")
    
    def show_dependencies(self, enode):
        self.send()
        self.send( "e-name:       ", enode.ename)
        self.send( "ancestors:    ", [ e.ename for e in enode.ancestors_t()])
        
        self.send( "descent-args: ", [e.ename for e in enode.input_nodes_pre])
        self.send( "descent-deps: ", [e.ename for e in 
                                      self.sort_enodes(enode.dependencies_descent)])
        self.send( "ascent-args:  ", [e.ename for e in enode.input_nodes_post])
        self.send( "ascent-deps:  ", [e.ename for e in 
                                      self.sort_enodes(enode.dependencies_ascent)])
        self.send()
        
    
    def init_vcubes(self):
        """Initialized v-cube that are attached to e-nodes and that 
        hold the produced values. A ascent-cube is created only if the 
        corresponding node defines an ascent."""
        
        
            
        def traverse(enode):
            
            vcubes_descent = OrderedDict()
            side_vcubes_descent   = []
            for enode_ in enode.dependencies_descent:
                vcubes_descent[enode_.ename] = self.get_input_vcube(enode, enode_)         
                if not enode_ in enode.ancestors_t():
                    side_vcubes_descent += [ enode_.ename ]
              
            enode.dvc = VCube(target_ename  = enode.ename,
                              vcubes        = vcubes_descent,
                              side_vcubes   = side_vcubes_descent,
                              pathkeeper    = deepcopy(self.pathkeeper),
                              run_number    = self.run_number,
                              )
            
            for enode_ in enode.echildren:
                traverse(enode_)

            vcubes_ascent  = OrderedDict()
            side_vcubes_ascent    = []
            if enode.has_ascent():
                for enode_ in enode.dependencies_ascent:
                    vcubes_ascent[enode_.ename] = self.get_input_vcube(enode, enode_)
                    if not enode_ in enode.ancestors_t():
                        side_vcubes_ascent += [ enode_.ename ]         
                enode.avc = VCube(target_ename  = enode.ename,
                                  vcubes        = vcubes_ascent,
                                  side_vcubes   = side_vcubes_ascent,
                                  pathkeeper    = deepcopy(self.pathkeeper),  
                                  run_number    = self.run_number,                                
                                 )
            else:    
                enode.avc = None
                
        traverse(self.eroot)
            

    def get_input_node(self, enode, ename):
        return enode.input_node_map[ename]


    def get_input_mode(self, enode, enode_):
        """Returns 'descent' or 'ascent' indicating which vcube is
        of enode_ is accessed from enode."""
        
        if   enode.e_relations[enode_]  == 'self':
            mode = None
        elif enode.e_relations[enode_]  == 'ancestor_t':
            mode   = 'descent'
        else:
            if enode_.has_ascent():
                mode = 'ascent'
            else:
                mode = 'descent'

        return mode 

        
    def get_input_vcube(self, enode, en):
        
        if stems_from(en, ENode):
            enode_ = en
        elif type(en) is str:
            enode_ = self.get_input_node(enode,en)
        else:
            raise TypeError(en)
        
        mode = self.get_input_mode(enode, enode_)
        
        if   mode == 'descent':
            vcube = enode_.dvc
        elif mode == 'ascent':
            vcube = enode_.avc
        elif mode is None:
            vcube = None
        else:
            self.bugger()
        
        return vcube
        

    def get_inputs(self, enode, f, mid):
        
        inputs  = OrderedDict()
        v_names = OrderedDict()
        for ename in enode.input_names[f]:
            
            enode_  = self.get_input_node(enode, ename)
            
            if ename in list(enode.input_defaults.keys()):
                
                value = deepcopy(enode.input_defaults[ename])
                name  = value2string(value)
                
            elif enode_ is enode:
                
                value,\
                name  = self.get_sibling_inputs(enode, f, mid)

            else:    

                common_ancestor = enode.last_common_ancestor(enode_)
                relation        = enode.e_relations[enode_]
                vcube           = self.get_input_vcube(enode, enode_)
                assert type(vcube) is VCube, \
                        (enode.ename, ename, vcube, self.get_input_node(ename).ename)
                
                if relation == 'ancestor_t':
                                
                    value = vcube.get_value(mid)
                    name  = vcube.get_value_name(mid)
    
                elif relation == 'ancestor_s' or relation == 'successor_t':
                    
                    
                    m_index_   = self.prune_m_index(mid,common_ancestor.ename)
                    m_index__ = OrderedDict()
                    for k in m_index_.keys():
                        if k in vcube.dim_names():
                            m_index__[k] = m_index_[k]
                    
                    vcube_ = vcube.get_subcube(m_index__)
                    value  = vcube_
                    name   = 'vcube ' + relation
                else:
                    assert 0, (enode, enode_, relation)
                    
            inputs[ename]  = value    
            v_names[ename] = name
            
        return inputs,v_names

    
    def set_inputs(self, enode, f, mid, inputs):
        
        #inputs  = OrderedDict()
        #v_names = OrderedDict()
        
        for ename in enode.input_names[f]:
            
            enode_  = self.get_input_node(enode, ename)
            
            if ename in list(enode.input_defaults.keys()):
                ## nothing to do
                pass
                
            elif enode_ is enode:
                raise NotImplementedError()
                self.set_sibling_inputs(enode, f, mid, inputs[ename])

            else:    
            
                common_ancestor = enode.last_common_ancestor(enode_)
                relation        = enode.e_relations[enode_]
                vcube           = self.get_input_vcube(enode, enode_)
                assert type(vcube) is VCube, \
                        (enode.ename, ename, vcube, self.get_input_node(ename).ename)
                
                value = inputs[ename]
                
                if relation == 'ancestor_t':
                    
                    #vcube.show()
                    #print(mid)
                    #print(value) 
                     
                    vcube.set_value(mid, value )
                    
                elif relation == 'ancestor_s' or relation == 'successor_t':
                    
                    m_index_   = self.prune_m_index(mid,common_ancestor.ename)
                    m_index__ = OrderedDict()
                    for k in m_index_.keys():
                        if k in vcube.dim_names():
                            m_index__[k] = m_index_[k]
                    
                    vcube_ = value
                    vcube.set_subcube(m_index__, vcube_)
                
                    
                else:
                    assert 0, (enode, enode_, relation)
                    
            inputs[ename]  = value    
            #v_names[ename] = name
            
        
    
    
    def get_sibling_inputs(self, enode, f, mid):
        
        if   f is enode.ascent:
            vc = enode.avc 
        else: 
            vc = enode.dvc
        
        svalues_o = Named()
        svalues_d = OrderedDict()
        
        for ename_ in vc.dim_names_non_target():
            mid_ = deepcopy(mid)
            if mid_[ename_] > 0:
                mid_[ename_] -= 1
                vc_sub = vc.get_subcube(mid_)
                dvalues = list(vc_sub.cube.ravel())
                bubes   = list(vc_sub.bube.ravel())
                
                dvalues = [ (dvalues[i] if bubes[i] else None) for i in indices(bubes) ]
                
                assert len(dvalues) == vc_sub.dim_sizes()[enode.ename]
                
                svalues_d[ename_] = dvalues
            else:
                ## check once again because we risk endless loop here
                assert not enode.ename in enode.input_names[enode.initial_siblings]
                ## ok, now produce initial values which may depend on proper 
                ## ancestors
                svalues_raw = self.produce(enode, enode.initial_siblings)
                
                ## dont be picky on container type
                if not type(svalues_raw) is dict:
                    svalues_raw = svalues_raw.__dict__
                if ename_ in svalues_raw:
                    svalues_d[ename_] = svalues_raw[ename_]
                else:
                    svalues_d[ename_] = None
        
        for ename_, value_ in svalues_d.items():
            svalues_o.__setattr__(ename_, value_)
            
        return svalues_o, "sibling values"




    def prune_m_index(self, m_index, ename):
        
        m_index_ = OrderedDict()
        for k,v in m_index.items():
            m_index_[k] = v
            if k == ename:
                break
        return m_index_
            
            
    def run(self):
        
        self.run_index = OrderedDict()
        self.deploy(self.eroot)
        vc = self.enodes[-1].dvc
        
        return vc
    

    def deploy(self, enode):
        """Realizes enode by calling its production-functions based 
        on the current run-index. """

        
        if stems_from(enode, VCNode):
            self.deploy_vcube(enode, enode.pvc)
            return
        
    
        ## special case prunded production        
        if enode.prune or self.produce(enode, enode.prune_condition):
            self.deploy_pruned(enode)
            return
        
        first   = True    
        loop    = True
        
        while loop:
            
            d_values = self.produce(enode, enode.descent)
            if not d_values:
                self.deploy_pruned(enode)
                return
            assert d_values and getattr(d_values, '__iter__', False), \
                 "%s.descent should output non-empty iterable." % enode.ename 
            d_labels = self.make_labels(enode, d_values, enode.descent,
                                         enode.explicit_descent_labels) 
            
            if first:
                enode.dvc.set_values(d_values, self.run_index, d_labels)
                enode.d_values  = d_values
                size_old        = 0
            else:
                enode.dvc.add_values(d_values, enode.d_values, self.run_index, 
                                     d_labels)
                enode.d_values += d_values
                size_new        = len(enode.d_values)
                size_old        = size_new - len(d_values)
            
            for i in indices(d_values):
                self.run_index[enode.ename] = size_old + i
                for enode_ in enode.echildren:
                    self.deploy(enode_)
            self.run_index.pop(enode.ename)    
            
            
            if enode.has_ascent():
                a_values  = self.produce(enode, enode.ascent)
                a_labels  = self.make_labels(enode, a_values, enode.ascent,
                                              enode.explicit_ascent_labels)
                enode.avc.set_values(a_values, self.run_index, a_labels)
            
            first  = False        
            loop   = enode.has_loop_condition() and \
                     self.produce(enode, enode.loop_condition)
                    
        del enode.d_values 
    
    
    def deploy_pvc(self, enode):
        pass
        
    
         
    def deploy_pruned(self, enode):
        """no loops, no pruning, no ascent."""
        enode.dvc.set_values([], self.run_index,[])                            
        a_values = self.produce(enode, enode.ascent_pruned)  
        if enode.has_ascent_pruned():  
            labels   = self.make_labels(enode, a_values, enode.ascent_pruned,
                                              enode.explicit_ascent_labels)
            enode.avc.add_values(a_values, [], self.run_index, labels)
    
    
 
    def produce(self, enode, f):
        
        #assert  f == enode.descent or  f == enode.ascent, f.__name__
        
        if  f in [ enode.descent, enode.ascent ]:
            
            ## skip descent production if already done for relevant run-indexes  
            vc = enode.dvc if f == enode.descent else enode.avc
            values = self.get_former_values(vc, self.run_index)
            if not values is None:
                return list(values)
            
            ## use pvc-values if present
            pvc = enode.pdvc if f == enode.descent else enode.pavc
            if pvc:
                values = self.get_former_values(pvc, self.run_index)
                if not values is None:
                    return list(values)
            
           
        ancestor_names = [ e.ename for e in enode.ancestors_t() ]
        assert set(self.run_index.keys()) == set(ancestor_names),\
            (list(self.run_index.keys()), ancestor_names)
        
        #print("GGG")
        #print(enode.ename)
        #print(f.__name__)
        inputs, vnames = self.get_inputs(enode, f, self.run_index)
        
        verbose_list  = self.__getattribute__('verbose_' + f.__name__)
        if not verbose_list is None:
            if enode.ename in verbose_list or verbose_list == []:
                self.announce_production(f, enode, vnames )  
        
        #print()
        #print(enode.ename)
        input_values = list(inputs.values())
        
        def g(*args):
            return f(*args)
                       
        if enode.timeout and f in [ enode.descent]:
           
            inputs_values_, values, finished = \
                timeout_call_2(f, 
                               *input_values,
                               timeout=enode.timeout)
                
            if finished:
                
                inputs_ = OrderedDict(zip(inputs.keys(), inputs_values_))       
                self.set_inputs(enode, f, self.run_index, inputs_)
            
            else:
                values = []
                
        else:
            
            values =  f(*input_values)

        
        if f == enode.descent and enode.descent_selection:
               
            assert max(enode.descent_selection) < len(values),\
                       "%s has produced %d values only, but selection has index %d" %\
                       (enode.ename, len(values), max(enode.descent_selection) ) 
                       
            values = [values[i] for i in enode.descent_selection]
                
        return values


    def get_former_values(self, vc, mid):
        """Gets values from vc as specified by mid. Unknown dims in mid are
        ignored. If any index for a known key is too large None is returned.
        TODO: move to v-cube?
        """

        mid_        = deepcopy(mid)
        sizes       = vc.dim_sizes()
        target_name = vc.target_ename
        
        ## drop unknown keys
        for k in mid.keys():
            if not k in vc.dim_names():
                mid_.pop(k)

        ## check for too large indices
        for e in mid_.keys():
            if mid_[e] >= sizes[e]:
                return None
        
        ## only return non-empty sequence
        if sizes[target_name] > 0:
        
            mild              = vc.mild(None,mid_)
            mild[target_name] = list(range(sizes[target_name]))

            cube = vc.cube[list(mild.values())]
            bube = vc.bube[list(mild.values())]
            
            if bube.all():
                return cube.ravel()
        
        return None
    



    def deploy_vcube(self,enode, vcube):
        """Separate deploy-function for vc-nodes,
        no looping, no pruning, no ascent."""
        
        d_values, _, d_labels = self.produce_from_vcube(enode, vcube)
        
        ## TODO: what does all([ not b for b in b_vlaues]) mean?
        ## and should we pass  b_values too?
        enode.dvc.set_values(d_values, self.run_index, d_labels)
        
        for i in indices(d_values):
            self.run_index[enode.ename] =  i
            for enode_ in enode.echildren:
                self.deploy(enode_)
        
        self.run_index.pop(enode.ename)  

    def produce_from_vcube(self, enode, vcube):
        
        run_index_ = self.map_mid(enode, self.run_index)
        
        vc_sub = vcube.get_subcube(run_index_)
        
        d_values = list(vc_sub.cube.ravel())
        b_values = list(vc_sub.bube.ravel())
        
        v_labels  = vc_sub.value_labels
        
        assert len(d_values) == len(v_labels)
        if enode.descent_selection:
            d_values = [d_values[i] for i in enode.descent_selection]
            v_labels = [v_labels[i] for i in enode.descent_selection]
        
        return d_values, b_values, v_labels


    
    def make_labels(self, enode, values, production_f, explicit_f):
        
        d_labels_explicit     = self.produce(enode, explicit_f )        
        d_labels_varnames     = self.get_labels_from_varnames(production_f)
        
        d_labels_value_names  = [ VCube.value_name(v, i) for i,v in enumerate(values) ]
        d_labels_from_get_name = []
        
        try:
            for v in values:
                d_labels_from_get_name += [ v.get_name() ]
        except AttributeError:
            d_labels_from_get_name = None
            
    
        
        if d_labels_explicit and len(d_labels_explicit) == len(values):
            d_labels = d_labels_explicit
        else:
            if enode.labels_from_varnames and d_labels_varnames:
                d_labels = d_labels_varnames
            else:
                if d_labels_from_get_name and \
                   len(set( d_labels_from_get_name )) == len(values):
                    
                    d_labels = d_labels_value_names
                elif d_labels_varnames and len(d_labels_varnames) == len(values): 
                    d_labels = d_labels_varnames
                else:
                    d_labels = d_labels_value_names
        
        return d_labels
                 
        
    def map_mid(self, enode, mid):
        ## TODO: mapping based on labels and entries
        
        mid_out = OrderedDict()
        
        if not enode.has('matchings'):
            enode.matchings = OrderedDict()

        
        
        for ename in mid:                    
            if ename in enode.dvc.dim_names():
                if ename not in enode.matchings:
                    ## get head of mid
                    mid_ = OrderedDict()
                    for ename_ in mid:
                        if ename_ == ename:
                            break
                        else:
                            mid_[ename_] = mid[ename_]
    
                    svc_tree = enode.dvc.vcubes[ename].get_subcube(mid_)
                    svc_pre  = enode.pvc.vcubes[ename].get_subcube(mid_)
                    
                    assert all( [ s == 1 or e == ename 
                                  for e,s in svc_tree.dim_sizes().items() ] )
                    assert all( [ s == 1 or e == ename 
                                  for e,s in svc_pre.dim_sizes().items()  ] )
                    
                    values_t  = list(svc_tree.cube.ravel())
                    values_p  = list(svc_pre.cube.ravel())
                    
                    ## build similarities
                    Q = empty(( len(values_t),len(values_p)))                
                    for i in indices(values_t):
                        for j in indices(values_p):
                            name_t = value2string(values_t[i])
                            name_p = value2string(values_p[j])
                            Q[i,j] = LTMatcher.string_similarity(name_t, name_p)
                    
                    m =LTMatcher(Q)
                    enode.matchings[ename] = m.match()
                
                ## apply matching
                mid_out[ename] = enode.matchings[ename][mid[ename]]
            
            else:
                mid_out[ename] = mid[ename]
        
        
        return mid_out


    def sort_enodes(self, enodes):
        
        def in_enodes(enode):
            return any([ enode is e_ for e_ in enodes ])
        
        if type(enodes) in [ list, set ]:
            return [ e for e in self.enodes if in_enodes(e) ]
        elif type(enodes) == OrderedDict:
            out = OrderedDict()
            for e in self.enodes:
                if in_enodes(e):
                    out[e.ename] = e
            return out
        else:
            raise Exception()
        
        
        
        return [ ]

    
 
    def announce_production(self,f, enode, value_names):


        if   f in enode.production_functions_pre():
            deps =  enode.dependencies_descent
        elif f in enode.production_functions_post():
            deps =  enode.dependencies_ascent
        else:
            deps = []
                
        L_e = max([len(e.ename) for e in self.enodes]   + [0])
        
        self.send('\nproduce %s.%s' % (enode.ename, f.__name__)) 

        done = set()        
        for enode_ in reversed(enode.ancestors_t()):
            ename  = enode_.ename
                
            if enode_ in deps:
                vcube   = self.get_input_vcube(enode, enode_)
                v_name = vcube.get_value_name(self.run_index)
                self.send(ename+':',' '*(L_e-len(ename) + 1), v_name)
            
            done.add(ename)
                
        for ename,v_name in value_names.items():
            if not ename in done:
                self.send(ename+':',' '*(L_e-len(ename) + 1), v_name)
               
        self.send()
 
    @staticmethod
    def get_def_body_ast(f):
        """Returns AST rooted in the body of function definition 
        of function f"""
        
        source  = inspect.getsource(f).strip()
        source  = re.sub('(?s).*?(def\s+.*)',r'\1', source)
        asttree = ast.parse(source)
        body    = asttree.body[0]
        
        return body
        
 
    @staticmethod
    def guess_ouput_length( f):
        """Rough try to statically get length of returned expression,
        assuming the length is unique.
        
        TODO: deal with assignments before return: A = [ ..] .. return A"""
        
        body = TreeRunner.get_def_body_ast(f) 

        ## get first (breadth-first) return statement
        ## special cases are likely not to come first since
        ## they are often below if-statement
        first_return = None
        for sub in ast.walk(body):
            if type(sub) is ast.Return:
                first_return = sub
                break
        
        length = None
        
        if first_return:
            out = first_return.value
            if type(out) is ast.List:
                #elements_expression = out.body 
                elements = out.elts
                length = len(elements)                
        
        return length
    
    
    def get_labels_from_varnames(self,f):
        """Assuming that f reutnr list the elements of which all are references
        we can take there nemaes as labels."""
        
        body = TreeRunner.get_def_body_ast(f) 

        ## get first (breadth-first) return statement, see guess_ouput_length
        first_return = None
        for sub in ast.walk(body):
            if type(sub) is ast.Return:
                first_return = sub
                break
        
        labels = []
        
        if first_return:
            out = first_return.value
            if type(out) is ast.List:
                #elements_expression = out.body 
                elements = out.elts
                for entry in elements:
                    if type(entry) is ast.Name:                        
                        labels += [entry.id]
                    else:
                        labels = None
                        break
            else:
                labels = None
        else:
            ## this should not happen but thats not our business here
            labels = None
            
        return labels
    


class ENode(Named):
    """Defines a dimension in an experiment."""
    
    def init(self,

                ename          = None,
                experiment     = None,
                index          = None, 
                
                input_refs     = [],
            
                ## tree edges
                eparent        = None,
                echildren      = [],
                left_esibling  = None,
                right_esibling = None,
                
    
                input_names         = OrderedDict(), ## keys: production-function
                input_defaults      = OrderedDict(), ## default must not 
        
                input_names_all     = set(),         ## union over all pfs 
                                                      
                e_relations         = OrderedDict(), ## keys: e-nodes
                                                      
                input_node_map      = OrderedDict(), ## keys: e-names
                
                input_nodes         = set(), 
                input_nodes_pre     = set(),
                input_nodes_post    = set(),
                
                descent_selection   = [], ## subset of descent-value-indices
                test_selection      = [], ## use all descent values if no selection is made 
                ds                  = [], ## abrivation descent_selection
                ts                  = [], ## abrivation test_selection
                                        
                dvc            = None, ## v-cube for descent results 
                avc            = None, ## v-cube for ascent results
                pdvc           = None, ## pre-computed descent cube
                pavc           = None, ## pre-computed ascent cube
                
                
                prune          = False,  
                
                                
                labels_from_varnames = False, ## whether var-names  should be taken 
                                              ## as labels whenever descent return statement 
                                              ## gives a list of of variables
                timeout     = None, ## timeout for descent in seconds

            ): 
        
        if ename is None:
            self.ename = self.__class__.__name__
            
        if ts and not test_selection:
            self.test_selection = ts

        if ds and not descent_selection:
            self.descent_selection = ds
                        
        
    def descent(self, *inputs):
        """Yields list of objects that are bind to the pnode.dnodes 
        as dnode.value. 
        inputs are any predecessors, i.e. other enodes"""
        return [ None ]
   
    def explicit_descent_labels(self):
        """Explicit labels for descent values. These labels should 
        not depend on other e-nodes, e.g. this function has no arguments."""
        return None


    def ascent(self, *inputs):
        """Define ascent value."""
        return None
    
    def explicit_ascent_labels(self):
        """Explicit labels for ascent values. These labels should 
        not depend on other e-nodes, e.g. this function has no arguments."""
        return None

    def ascent_pruned(self, *inputs):
        """Return substitute for ascent output in case prune_condition holds.
        """
        return None
    
    def prune_condition(self, *inputs):
        """See ascent pruned.
        """
        return False  
    
    def loop_condition(self, *inputs):
        """Only evaluated if self.loop. If the returned value is true,
        the node is deploy again. The new descent values are appended to
        the old once while the ascent value is overwritten."""
        return False

    def initial_siblings(self, *inputs):
        """Return object that hold sibling values as members such that
        can be reference object.ename where ename is the name of the 
        sibling pivot."""
        return None
        
    def production_functions_all(self):
        return self.production_functions_pre() + self.production_functions_post()

    def production_functions_pre(self):
        """called before children are deployed, cannot access successors """
        return [ self.descent, 
                 self.explicit_descent_labels,
                 self.prune_condition,
                 self.initial_siblings
                ]

    def production_functions_post(self):        
        """called ater children are deployed, can access successors"""

        return [ self.ascent,
                 self.explicit_ascent_labels,
                 self.ascent_pruned,    ## though could also belong to pre
                                        ## since if called, children are not deployed
                 self.loop_condition, 
                 ]


    def ancestors(self):
                
        if self.left_esibling:
            return self.left_esibling.ancestors() + \
                   [ self.left_esibling ]          + \
                   self.left_esibling.successors_t()  
        else:
            if self.eparent:
                return [ self.eparent ] + self.eparent.ancestors()
            else:
                return [ ]
    
    def depth(self):
        if self.eparent:
            return self.eparent.depth() + 1
        else:
            return 0
              

    def ancestors_t(self):
        """Parent first."""
        if self.eparent:
            ancestor = self.eparent
        else:
            return [ ]
        return  [ ancestor ] + ancestor.ancestors_t()    
    
    
    def successors_t(self):
        """depth-right-first"""
        
        out = []
        if self.echildren:
            for enode in reversed(self.echildren):
                out += [ enode ] + enode.successors_t()
        return out
    
   

    def last_common_ancestor(self, enode_):
        ancestors_mhy = list(reversed(self.ancestors_t()))
        ancestors_thy = list(reversed(enode_.ancestors_t()))
 
        common_ancestor = None
        for d,node in enumerate(ancestors_mhy):
            
            if len(ancestors_thy) > d and node is ancestors_thy[d]:
                common_ancestor = node
            else:
                break
        
        return common_ancestor
        
    def has_ascent(self):
        return self.__class__.ascent != ENode.ascent
    
    def has_ascent_pruned(self):
        return self.__class__.ascent_pruned != ENode.ascent_pruned
    
    def has_loop_condition(self):
        return self.__class__.loop_condition != ENode.loop_condition
    
    

class VCNode(ENode):
    """ENodes that are created from vcubes. They are reduced:
    no ascent, no sibling-references, no pruning."""
    
    

if __name__ == '__main__':
    
    pass
    #tr = TreeRunner()
    
    #TreeRunner.guess_ouput_length(TreeRunner.hulla)





