"""A module that cares for parallelization of experimental runs. It defines 
parallelization back-ends that launch partial runs on the computing nodes.
The user has to configure the back-end in the project configuration module, by
setting remote user name and such. If the configuration is correctly set, the 
user can parallelize the computation by indicating the e-nodes along which 
concurrency can be be applied. 

"""

from cut.baseclass import Named
from cut.fileaccess import get_lines, posix_path, save_file,\
                                  save_in_shelve, load_from_shelve

from peewit.vcubes import VCube
from peewit.synchronization import Unison_Synchronizer, Dummy_Synchronizer
from cut.mixedtools import shell

from time import sleep
import os, re , posixpath, platform, random

from copy import deepcopy
from numpy.ma.core import ceil
from collections import OrderedDict
from os.path import join as opj
from peewit import archiving
from peewit.core import TreeRunner
from abc import abstractmethod,ABCMeta

class LauncherBase(Named, metaclass = ABCMeta):
    """
    Abstract class of for parallelization-backends. Should mention
    methods that are called by peewit.core.Experiment.
    Not up to date. And not really usefull either.
    """

    def init(self, 
             pathkeeper=None, 
             experiment=None,
             launch_specified = False,
             dont             = None,
             dimiss           = None,
             
             archivist        = None, ## access needed to store run_number
             
             
             ):
        pass
 
 
    def specify_launch(self,
                       target       = None, 
                       axes         = None,
                       dismiss      = False, 
                       dont         = False,
                       partial_finish = True,
                       shuffeled_launch = True,
                       rerun            = False,
                       ):
        """
        @param target: v-cube target e-name. If launcher is used, target has to be set 
            either here or as argument of self.run
        @param axes: a mapping (dict) from e-names that refer to the axis along which
            the parallelization takes place to the respective number of descent values,
            e.g. experimental parts (slides).
        @param dismiss: whether launcher should dismiss launches, instead of 
            committing them  
        @param partial_finish: whether finished part show be returned even though not
            all parts have finished
        """
        pass

        self.target  = self.pathkeeper.target_ename = target
        self.dismiss = dismiss 
        self.dont         = dont
        self.partial_finish = partial_finish
        self.shuffeled_launch = shuffeled_launch
        self.rerun  = rerun
        assert self.target 
        self.launch_specified = True

        self.axes    = axes 
        
        
        
    def complete_axes(self):
        
        axes_ = []
        for axis in self.axes:
            if type(axis) is tuple:
                assert len(axis) == 2 and\
                       type(axis[0]) is type('') and\
                       type(axis[1]) is int 
                axis_ = axis
            else:
                assert type(axis) is type('') 
                
                enode = self.experiment.get_node(axis)
                descent_size = TreeRunner.guess_ouput_length(enode.descent)
                if descent_size is None:
                    self.bugger("Cannot detect descent size for axis %s.\n" +
                                "Please specify number of descent values. Pass"+
                                "pair (ename,descent-size) instead of ename only." 
                                % axis)
                else:
                    axis_ = [axis, descent_size]
            
            
            axes_ += [axis_]
            
        self.axes = axes_
                
    
    @abstractmethod
    def set_paths(self, pathkeeper):
        """Called when all needed paths collect by self and pathkeeper"""
        self.pathkeeper = pathkeeper
        
             
       

    def dismiss_launch(self):
        """kill jobs on cluster"""
        assert self.dismiss        
        
        pk = self.pathkeeper
        
        if not os.path.isfile(pk.path_abs('state',
                                          target=pk.target_ename,
                                          partname = 'main')):
            
            self.send('Cannot dismiss launch belonging to %s.' % pk.basename(),
                      'Quit now.' )
            quit()
        
        partnames  = self.launch_axes_to_subrun_strings(
                         self.axes)
        
        self.dismiss_parts(partnames)
    
    @abstractmethod  
    def dismiss_parts(self, partnames):
        """kill jobs on cluster"""
        pass
        
    @abstractmethod
    def main(self):
        pass
 
    @abstractmethod
    def launch(self, partnames):
        pass
                
    @abstractmethod
    def finish_main(self, partnames):
        pass
    
    @abstractmethod
    def finish_part(self,vc):
        pass
    
    @abstractmethod
    def update_project_dir(self):
        pass
        


class Remote_Single_Head_Launcher(LauncherBase):
    """For launches on machines that are bound on one file system, such that 
    the local machine has to communicate to a single remote machine only.

    The remote head is assumed to run a posix-system. It is synchronized 
    with the local machine via unison.
        
    This class is abstract. It has three non-abstract descendants.
        * SGE_Laucher for Sun Grid Engine clusters
        * Remote_NFS_Launcher for nodes that share the filesystem among 
            each other but not with the local machine. The jobs are 
            distributed by direct ssh-access to the nodes
        * Local_NFS_Launcher, like Remote_NFS_Laucher but assuming that 
            the local machine is bound to the same file-system too.
            
    In any case you will need password less login (ssh-agent) on the
    remote head ( and nodes in case of NFS_Lauchers) because for each launch
    ssh-login is called multiple times.  
    """ 
    
    def init(self,
             remote_user = None,
             remote_head = None,
             remote_head_architecture = None,
             remote_home_dir   = None,
             remote_base_dir   = None,
             remote_nodes_spec = None,
             notification_mail_address = None,
             remote_peewit_path   = None,
             python               = 'python',
             synced_paths         = [],
             
             verbose_sync            = 0,
             
             remote_python_paths_os     = None,
             exclude_from_sync          = [],
             
             sync_executable_32         = None,
             sync_executable_64         = None,
             remote_sync_executable_32  = None,
             remote_sync_executable_64  = None,
             
             architecture               = None,
             always_exlude_from_sync    = [ '*.pyc', '.svn', '.unison', '.hg', '.git', 
                                            archiving.GitPeewit.tool_dir_name,
                                            archiving.HgPeewit.tool_dir_name,
                                            ],
             
             rerun                      = False,
             more_dirs                  = None, ## inherit from pathkeeper if None keep own
             full                       = False,
             ):
        
        self.synchronizer  = Unison_Synchronizer()
        
        #self.synchronizer.unison_dir_rel_project = '.unison'
       


    def set_paths(self, pathkeeper):
        """When object is created  most paths are not yet set. This should 
        be called when they are."""
        
        self.pathkeeper = pathkeeper
        
        pk = self.pathkeeper
        
        
        
        if self.remote_home_dir is None:
            self.remote_home_dir = pk.homedir()   
        
        if self.remote_base_dir is None:
            self.remote_base_dir = pk.base_dir
        
        if self.remote_python_paths_os is None:
            self.remote_python_paths_os = pk.remote_python_paths_os 



        ## remote peewit path
        if not self.remote_peewit_path:
            remote_peewir_path_rel_base =\
                posix_path(os.path.relpath(pk.dir_abs_peewit_packages(),
                                           pk.abs_path()))
        else:
            if not posixpath.isabs(self.remote_peewit_path): 
                remote_peewir_path_rel_base = self.remote_peewit_path
            else:
                remote_peewir_path_rel_base = None
        
        if remote_peewir_path_rel_base:
            self.remote_peewit_path = posixpath.join(self.remote_home_dir,
                                                     self.remote_base_dir,
                                                     remote_peewir_path_rel_base)
        
        
        
        self.exclude_from_sync += ['python_paths_os']
        
        self.exclude_from_sync  = [ posixpath.normpath(p) 
                                    for p in self.exclude_from_sync ]
        

            
        if self.more_dirs is None:
            self.more_dirs  = pk.more_dirs       

        for group in pk.path_groups_keys:   
            self.synced_paths += pk.path_groups[group]
        
        self.synced_paths += self.more_dirs
        
        self.synced_paths = [ p for p in self.synced_paths
                              if not p in self.exclude_from_sync ]
    

        ## synchronizer setting  
        self.synchronizer.local_home_dir  = pk.homedir()
        self.synchronizer.base_dir        = pk.base_dir
        
        self.synchronizer.remote_home_dir = self.remote_home_dir
        self.synchronizer.remote_base_dir = self.remote_base_dir
        self.synchronizer.remote_host     = self.remote_head
        self.synchronizer.remote_user     = self.remote_user
        
        self.synchronizer.unison_dir_parent = pk.dir_abs_pdir('peewit')
        
        self.synchronizer.verbose         = self.verbose_sync

        if not self.remote_sync_executable_32:
            self.remote_sync_executable_32 = self.sync_executable_32

        if not self.remote_sync_executable_64:
            self.remote_sync_executable_64 = self.sync_executable_64

        if self.architecture is None:
            self.architecture                = platform.architecture()[0][:2]
        
        sync_executable = None
        attribute = 'sync_executable_' + self.architecture
        if self.has(attribute):
            sync_executable = self.__getattribute__(attribute)
        
        if sync_executable and not posixpath.isabs(sync_executable): 
            sync_executable = posixpath.normpath(posixpath.join(pk.homedir(),
                                                                pk.base_dir,
                                                                sync_executable))
            
        self.synchronizer.sync_executable       = sync_executable
        
        
        if self.remote_sync_executable_32 is None:
            if pk.remote_sync_executable_32:
                self.remote_sync_executable_32 = pk.remote_sync_executable_32
            else:
                self.remote_sync_executable_32 = pk.sync_executable_32

        if self.remote_sync_executable_64 is None:
            if pk.remote_sync_executable_64:
                self.remote_sync_executable_64 = pk.remote_sync_executable_64
            else:
                self.remote_sync_executable_64 = pk.sync_executable_64

        
        
         
        if self.remote_head_architecture is None:
            if not pk.remote_head_architecture is None:
                self.remote_head_architecture = pk.remote_head_architecture
            else:
                self.remote_head_architecture = self.architecture 
        arch = self.remote_head_architecture[:2]
        
        self.remote_sync_executable = self.__getattribute__('remote_sync_executable_' + arch ) 
        
        if self.remote_sync_executable and\
            not posixpath.isabs(self.remote_sync_executable): 
            self.remote_sync_executable = posixpath.normpath(
                                        posixpath.join(self.remote_home_dir,
                                                       self.remote_base_dir,
                                                       self.remote_sync_executable))
        
        self.synchronizer.remote_sync_executable = self.remote_sync_executable

        
    def get_remote_peewit_paths(self):#self, architecture ):
        
        return [ posix_path(opj( self.remote_home_dir,
                                 self.remote_base_dir,
                                 p )) 
                 for p in self.get_remote_peewit_paths_rel()]#architecture) ]
        

    def get_remote_peewit_paths_rel(self):#, architecture ):
        
        pk = self.pathkeeper
        
        assert self.remote_peewit_path
        
        paths = [ self.remote_peewit_path  ]
        
        paths += pk.remote_python_paths_os  
        
        return paths
    
    


    def launch_axes_to_subrun_strings(self, launch_axes):
        """@param spec:  launch spec as list of (ename, descent_size) pairs."""
        
        def extend_subrun_strings(remaining_axes):
            
            if remaining_axes:
            
                axis       = remaining_axes.pop(0)
                ename      = axis[0]
                enode      = self.experiment.get_node(ename)
                dselection = enode.descent_selection
                
                if dselection:
                    n_parts    = min( int(axis[1]), len(dselection))
                    prefices   = [ ename + '__' + str(dselection[i]) 
                                  for i in range(n_parts)]
                else:
                    n_parts    = int(axis[1])
                    prefices   = [ ename + '__' + str(i) 
                                   for i in range(n_parts) ]
                    
                suffices   = extend_subrun_strings(remaining_axes)
                
                if suffices:
                    extended   = [ prefix + '___' + suffix for suffix in suffices 
                                                       for prefix in prefices ] 
                    return extended
                else:
                    return prefices
            
            else:
                
                return None
        
        
        launch_axes = list(deepcopy(launch_axes))

        return extend_subrun_strings(launch_axes)
    

         
 
    def merge_partial_cubes(self, subrun_strings):
        
        def merge_rec(subruns_suffix, subruns_prefix):
            
            if any(subruns_suffix):
            
                subruns_suffix  = deepcopy(subruns_suffix)  
                subruns_prefix  = deepcopy(subruns_prefix)
                
                first_axes = sorted(set([ subrun.pop(0) for subrun in subruns_suffix ]))
                
                ename_set   = set([ axis[0] for axis in first_axes ])
                assert len(ename_set) == 1
                
                ename   = ename_set.pop()
                indices = [ axis[1] for axis in first_axes ]
                
                subcubes = []
                
                for index in indices:
                    if any(subruns_prefix):
                        subruns_prefix_ = [ prefix + [ (ename,index) ]  
                                            for prefix in subruns_prefix ]
                    else:
                        subruns_prefix_ = [ [ (ename,index) ] ]
                        
                    next_vcube =  merge_rec( subruns_suffix, subruns_prefix_ )
                    
                    if not next_vcube.flag('read_failed'):
                        subcubes   += [ next_vcube ]
                    
                
                vc = subcubes[0]
                for vc_ in subcubes[1:]:
                    vc.merge_straight(vc_,ename)
                
            else:
                assert len(subruns_prefix) == 1
                subrun = subruns_prefix[0]
                
                subrun_string = "___".join([ "__".join(axis) for axis in subrun ])
                
                vc = VCube()
                vc.pathkeeper = deepcopy(self.pathkeeper)
                vc.pathkeeper.partname = subrun_string
                vc.read(self.pathkeeper.target_ename)
                            
            return vc
      
    
        #axes = self.axes 
        #axes = deepcopy(axes)
        
        subrun_strings = sorted(subrun_strings)
        
        subruns_suffix  = self.subrun_strings_to_subruns(subrun_strings)
        
        vc = merge_rec(subruns_suffix, [])
        vc.pathkeeper = deepcopy(self.pathkeeper)
        
        return vc
            

    def subrun_strings_to_subruns(self, subrun_strings):
        """Return launch spec as list of (ename, index) pairs."""
        
        subrun_strings = sorted(deepcopy(subrun_strings))   
        
        subruns = []
        
        for subrun_string in subrun_strings:
            
            subrun_string_parts = subrun_string.split('___')
            
            subrun_coordinates = []
            
            for subrun_string_part in subrun_string_parts:
                
                (axis,index) = subrun_string_part.split('__')
                
                subrun_coordinates += [(axis,index)]
            
            subruns += [subrun_coordinates]
        
        return subruns
         

    def main(self): 
        """Read states and continue accordinglgy."""
        
        partnames  = self.launch_axes_to_subrun_strings(self.axes)
        
        
        
        if self.dismiss:
            ## user wants to cancel previously committed launches 
            self.dismiss_parts(partnames)
            self.send('\nLaunches dismissed. Quit experiment module after run().')
            ## in case of dismiss remove_current_files 
            ## is done afterwards
            if self.experiment.remove_current_files:
                self.pathkeeper.remove_current_files()
                self.update_project_dir()
            quit()

        state = self.get_main_state(partnames)        

        if state == 'no state file':
            
            if self.archivist:
                run_number = self.archivist.run_number
                self.pathkeeper.store_run_number(run_number)

            self.launch(partnames)
            
            self.send('\nLaunches engaged. Quit experiment module after run().\n')
            return None
            
        elif state == 'running':
            
            self.pull_logs(partnames)
            
            self.send('\nLaunches running. Try again later. ', 
                      'Quit experiment module after run().\n')
            return None
            
        elif state == 'finished':

            
            vc             = self.finish_main(partnames)        
            run_number     = self.pathkeeper.read_run_number()    
            vc.run_number  = run_number
            
            self.send('\nLaunches have finished. '+
                       'Return merged v-cube and continue with experiment module.\n')
            return vc
        
        elif state == 'partially done':
            
            run_number = self.pathkeeper.read_run_number()
            vc  = self.finish_main(partnames)        
            vc.run_number  = run_number
            
                
            
            self.send('\nLaunches are partially done. '+
                       'Return merged v-cube and continue with experiment module.\n')
            
            if self.rerun:
                self.send("Launch rerun of unfinished ones.")
                
                states = self.get_part_states(partnames)
                    
                undone = [ pn for pn, st in states.items() if st == 'committed']
                
                if undone:
                    self.dismiss_parts(undone)
                    self.launch(undone)
            
                
            
            return vc
        
        elif state == 'complete':
            
            run_number = self.pathkeeper.read_run_number()    
    
            vc = self.merge_partial_cubes(partnames)
            vc.run_number  = run_number
            
            vc.pathkeeper.partname ='main'
            return vc      
            
        else:
            self.bugger(state)


    def get_main_state(self, partnames):
        """Called by self.main"""
        
        pk = self.pathkeeper 
        assert pk.partname == 'main', pk.partname

        ## check state file
        path_main_state = pk.path_abs('state')
        
        state = self.read_local_state_file(path_main_state)
        
        if state == "no state file":
            return state
        else:
            ## state_raw is for instance 'launches engaged', anyway
            ## we have to check whether launches are still running or not
        
            self.part_states = states = self.get_part_states(partnames)
            
            if    all([ states[pn] == 'complete' for pn in partnames ] ):
                return 'complete'
            
            elif  all([ states[pn] == 'finished' for pn in partnames ] ):
                return 'finished'
            
            elif  all([ states[pn] == 'no state file'  for pn in partnames ] ):
                return 'no state file'            
            
            elif  any([ states[pn] in ['finished','complete'] for pn in partnames ] )\
                  and self.partial_finish:
                return 'partially done'
              
            elif any([ states[pn] in ['committed', 'running'] for pn in partnames ] ):
                        
                return 'running'
            
            else:
                self.bugger([ states[partname] for partname in partnames])
            
    

               
    def launch(self, partnames):

        assert self.pathkeeper.partname == 'main' 
        
        self.update_all()
        
        self.write_state_files(partnames,'committed') 
        self.push_state_files(partnames)
        
        self.engage_nodes(partnames)
        sleep(0.5)
        ## write state for main-file
        save_file('launches engaged', self.pathkeeper.path_abs('state') )
   
   
    def finish_main(self, partnames):
        
        self.pull_cubes(partnames)
        self.pull_logs(partnames)
        
        if self.partial_finish:
            partnames_ = []
            for partname, state in self.part_states.items():
                if state == 'finished':
                    self.part_states[partname] = 'complete'
                if state in ['finished', 'complete']:
                    partnames_ += [ partname ]
                    
            self.write_state_files(partnames_, 'complete')
                
            partnames = partnames_
            
                    
        vc = self.merge_partial_cubes(partnames)
        
        vc.pathkeeper.partname ='main'
        
        return vc
      
    def finish_part(self,vc):
        vc.save(full=self.full)
        save_file('finished', self.pathkeeper.path_abs('state'))
      

 
    def update_all(self):
        """propagates the current state of base-dir to the nodes
        such that they are up to date in terms of code and data. 
        
        """
        pk = self.pathkeeper
        
        ## take out project dir
        np = posixpath.normpath
        project_dir = np(pk.extend_project_dir_rel_base_posix())
        synced_paths = [ p for p in self.synced_paths 
                         if not np(p) == project_dir ]
    
        ## since the repo can be large, we prune it a bit
        omit  = self.always_exlude_from_sync[:]
        
        for dir in synced_paths:
                
            self.send('Push directory %s....' % dir)
            self.synchronizer.push([dir], omit=omit )
            self.send('Push done.')
        
        self.update_project_dir()
        
        

    def update_project_dir(self):
        """propagates the current state project dir to remote head
        
        * run-files are not pushed
        * logs and v-cubes are only pushed if they do not belong 
            to pending runs
        * png file are not pushed
        """
        self.create_project_subdirs_on_remote_head()
        
        pk    = self.pathkeeper
        
        project_dir = pk.extend_project_dir_rel_base_posix()
        
        omit  = self.always_exlude_from_sync[:]
        
        omit += [ '*.png', posixpath.join(*pk.dirs['life']) ]
        
        ## take out logs and vcubes and treat them sparately
        log_dir = posixpath.join(*pk.dirs['log'])
        vc_dir  = posixpath.join(*pk.dirs['vcube'])
        
        omit += [ log_dir, vc_dir ]
        
        self.send('Push project dir without logs, vcube and run-files ....' )
        self.synchronizer.push([project_dir], omit=omit, )
        
        self.send('Push done.')

        self.update_logs_and_vcubes()
        
        
    def update_logs_and_vcubes(self):
        
        ## 1) get pending runs
        ## 2) build list of files to omit
        ## 3) push dir excluding those
        
        pk    = self.pathkeeper
        omit  = self.always_exlude_from_sync[:]
        
        pending = self.pending_parts()
        
        omit += [ p + '.' + pk.suffices['out'] for p in pending ]
        omit += [ p + '.' + pk.suffices['err'] for p in pending ]

        log_dir = pk.dir_rel_base_pdir_posix('log')
        
        self.send('Push log-dir ....' )
        self.synchronizer.push([log_dir], omit=omit )
        self.send('Push done.')

        omit  = self.always_exlude_from_sync[:]
        omit += [ p + '.' + pk.suffices['vcube'] for p in pending ]

        vc_dir  = pk.dir_rel_base_pdir_posix('vcube')       
        
        
        self.send('Push vcube-dir ....' )
        self.synchronizer.push([vc_dir], omit=omit )
        self.send('Push done.')

    
    def pending_parts(self):
        """lsit of basenames"""
        pk = self.pathkeeper
        
        state_dir = pk.dir_abs_pdir('state')
        state_files = os.listdir(pk.dir_abs_pdir('state'))
                   
        basenames = []
        for file_name in state_files:
            basename = '.'.join(file_name.split('.')[:-1])
            suffix   = file_name.split('.')[-1]
            assert suffix == pk.suffices['state']
            path = os.path.join(state_dir,file_name)
            state = self.read_local_state_file(path)
            
            if self.pending_state(state):
                basenames += [ basename ]
        
        
        basenames = sorted(basenames)
        return basenames
            
                   
    def pending_state(self, state):
        return state in ['committed', 'running', 'finished']
    


    def get_part_states(self,partnames):
        
        pre_pull_states, paths_rel_base = self.read_local_state_files(partnames)
    
        paths =   [pr for pn,pr in paths_rel_base.items()
                   if not pre_pull_states[pn] == 'complete'  ]
        
        if paths:
            self.send('Pull state files ....')
            self.synchronizer.pull(paths)
            self.send('Pull state files done', lb=2)        
        
        
        states, paths_rel_base  = self.read_local_state_files(partnames)
        
            
        self.send('\nStates partial runs: ')            
        for partname, state in states.items():
            self.send(partname, ': ', state)  
        self.send('\n') 
    
        return states
    
    
    def read_local_state_file(self, path):
        
        try:
            state = get_lines(path)[0].strip()
        except IOError:
            state = 'no state file'
                
        return state
        
      
    def read_local_state_files(self, partnames):
        
        pk = self.pathkeeper 
        paths_rel_base   = OrderedDict()
        paths_abs        = OrderedDict()
        states           = OrderedDict()
        
        
        for partname in partnames:
            paths_rel_base[partname]  = pk.path_rel_base('state', partname=partname)                                    
            paths_abs[partname]       = pk.path_abs('state', partname=partname)
            try:
                states[partname] = get_lines(paths_abs[partname])[0]
            except IOError:
                states[partname]     = 'no state file'
        
        return states, paths_rel_base
        
 
 
    def write_state_files(self,partnames, state_string):              
        
        pk = self.pathkeeper                         
        
        for partname in partnames:
            path_abs_state       = pk.path_abs('state', partname=partname)
            save_file(state_string, path_abs_state)
            
    
    def push_state_files(self, partnames):

        pk = self.pathkeeper                         
        paths_state = []
    
        for partname in partnames:
            path_rel_base  = pk.path_rel_base_posix('state',partname=partname)
            paths_state += [ path_rel_base ]
            
        self.send('Push state files....')
        self.synchronizer.push( paths_state )
        self.send('Push done.')
        
        


    def pull_cubes(self, partnames):    
        """           
        @param partnames: list of strings, each representing a part of the
            experiment. See peewit.core.Experiment.get_my_part_indices
        
        Makes v-cubes with from partial runs available at standard locations.
        """
        
        vcube_paths = [ self.pathkeeper.path_rel_base('vcube',partname=partname) 
                        for partname in partnames ]
        self.send('Pull v-cubes ....')
        self.synchronizer.pull( vcube_paths )
        self.send('Pull v-cubes done.',lb=2)
        
        
        
        
    def pull_logs(self, partnames):    
        """           
        @param partnames: list of strings, each representing a part of the
            experiment. See peewit.core.Experiment.get_my_part_indices
        
        Makes logs from remote partial runs available at standard log paths.
        """
        
        err_paths = []
        out_paths = []

        for partname in partnames:
            err_paths += [ self.pathkeeper.path_rel_base('err', partname=partname) ]
            out_paths += [ self.pathkeeper.path_rel_base('out', partname=partname) ]
        
        self.send('Pull logs ....')
        self.synchronizer.pull( err_paths + out_paths )
        self.send('Pull logs done.', lb=2)


    def python_call(self,partname):
        
        pk = self.pathkeeper

        error_path  = posixpath.join(self.remote_pdir_abs('log'),
                                    pk.filename('err',partname=partname))

        output_path = posixpath.join(self.remote_pdir_abs('log'),
                                    pk.filename('out',partname=partname))
                     
        module_path = posixpath.join(self.remote_base_dir,
                                     posix_path(pk.path_module_rel_base())) 

        
        command     = self.python + ' ' +\
                      module_path       +\
                      ' --version ' + str(pk.version) + ' --part ' + partname           
        
        
        return command, output_path, error_path 


    def create_project_subdirs_on_remote_head(self):
        
        dirs = []
        for dir in self.synced_paths:
            
            dirs += [ self.remote_dir_abs(dir) ]
            
        for dir_key in [ 'log', 'vcube', 'life',
                         'state', 'pid','jnr','qsub', 'phm']:
            
            
            dirs += [ self.remote_pdir_abs(dir_key) ]
            
        
        remote_command = "mkdir -p " + " ".join(dirs)
            
        command = ['ssh', self.remote_user + '@' + self.remote_head,
                       remote_command ]
            
        shell(command)
         
    def remote_dir_abs(self,dir_rel_base):
        return posixpath.join(self.remote_home_dir, 
                              self.remote_base_dir, 
                              dir_rel_base)

    def remote_pdir_abs(self,dir_key):
        return posixpath.join(self.remote_home_dir, 
                              self.remote_base_dir, 
                              self.pathkeeper.dir_rel_base_pdir_posix(dir_key))



class PBS_Launcher(Remote_Single_Head_Launcher):
    """For launches at machines that provide Portable Batch System (PBS) services.
    See for instance U{http://sc.tamu.edu/help/origins/batch.shtml}""" 
    def init(self,
             path_cluster_settings = None,
             queue                 = None,
             shuffeled_launch      = False,
             qdel_sep              = ','
             ):
        """@param path_cluster_settings: path to script on remote head 
            that sets env-variables needed in order to submit a job,
            for instance '/usr/local/sge-root/default/common/settings.sh'
        """
        pass
    
    def dismiss_parts(self, partnames):
        """kill jobs on cluster"""
        pk = self.pathkeeper
        
        
        job_numbers = []
        for partname in partnames:
            jnr_path     = pk.path_abs('jnr',partname=partname)
            job_number   = get_lines(jnr_path)[-1]
            job_numbers += [ job_number ]
        
        
        if self.path_cluster_settings:
            epart = 'source ' + self.path_cluster_settings + ' && ' 
        else:
            epart = ''
        epart += 'qdel ' + self.qdel_sep.join(job_numbers) + ''
        
        command = ['ssh', self.remote_user + '@' + self.remote_head, epart ]
        output = shell(command)
        self.send(output)
        self.send("Cluster dismiss done.")


    def engage_nodes(self, partnames):
        """
        @param partnames: list of strings, each representing a part of the
            experiment. See peewit.core.Experiment.get_my_part_indices 
        
        commits the partial runs to computing nodes and creates agent 
        that waits for the result.    
        """ 
        
        pk = self.pathkeeper

        if self.shuffeled_launch:
            partnames = deepcopy(partnames)
            random.shuffle(partnames)
        
        ## push qsub   
        paths_qsub  = []
        for partname in partnames:
            path_abs_qsub       = pk.path_abs('qsub', partname=partname)
            path_rel_home_qsub  = pk.path_rel_base('qsub', partname=partname)
            script = self.qsub_script(partname)
            save_file(script, path_abs_qsub)
            paths_qsub += [ path_rel_home_qsub ]

        self.send('Push qsub-scribts ....' )      
        
        self.synchronizer.push( paths_qsub )
        self.send('Push qsub-scribts done.')
        
        
        
        ## commit to PBS     
        for i,partname in enumerate(partnames):
            path =  self.remote_home_dir + '/' + self.remote_base_dir +\
                    '/' + paths_qsub[i]
            
            
            ## build command that will be called on remote host
            epart = 'export TERM=linux &&'
            
            if self.path_cluster_settings:
                epart = 'source ' + self.path_cluster_settings + ' && ' 
            
            epart += 'qsub ' + path 
            
            command = ['ssh', self.remote_user + '@' + self.remote_head, epart ]
            
            self.send("Cluster submit ....\n")            
            output = shell(command)
            
            self.send(output)
            self.send("Cluster submit done.\n")
            
            ## local save jnr             
            job_number   = self.extract_job_id(output)      
            jnr_path     = pk.path_abs('jnr',partname=partname)
            save_file(str(job_number) , jnr_path)
            
            sleep(0.5)
            
    

 

class SGE_Launcher(PBS_Launcher):
    """For launches at machines that provide Portable Batch System (PBS) services.
    See for instance U{http://sc.tamu.edu/help/origins/batch.shtml}""" 
    def init(self,
             path_cluster_settings = None,
             qdel_sep              = ','
             ):
        """@param path_cluster_settings: path to script on remote head 
            that sets env-variables needed in order to submit a job,
            for instance '/usr/local/sge-root/default/common/settings.sh'
        """
        pass

      
    def qsub_script(self, partname):
                  
        command, output_path, error_path  = self.python_call(partname)
    
        python_paths = self.get_remote_peewit_paths()
        
        
        script =  ('#$ -q %s \n' % self.queue if self.queue else '') +\
                   '#$ -cwd \n'  +\
                   '#$ -M %s \n' % self.notification_mail_address +\
                   '#$ -m bea \n' +\
                   '#$ -e %s \n' % error_path +\
                   '#$ -o %s \n' % output_path +\
                   '#$ -v PYTHONPATH=%s \n' % ':'.join(python_paths) +\
                   command + '\n'
        
        return script


    def extract_job_id(self, output):
        output       = output.strip()
        pattern      = r'Your\ job\ (\d{1,8})\ '
        findings     = re.findall( pattern, output )
        job_number   = findings[0]  
        return job_number
    
    
 

class OPBS_Launcher(PBS_Launcher):
    """For launches at machines that provide Portable Batch System (PBS) services.
    See for instance U{http://sc.tamu.edu/help/origins/batch.shtml}""" 
    def init(self,
             path_cluster_settings = None,
             qdel_sep              = ' ',
             time_bound            = '06:00:00',
             mem_bound             = '2gb'
             ):
       
        pass

        
    def qsub_script(self, partname):
                  
        command, output_path, error_path  = self.python_call(partname)
    
        python_paths = self.get_remote_peewit_paths()#self.architecture)
        
        
        
        script =  ('#PBS -q %s \n' % self.queue if self.queue else '') +\
                   '#PBS -M %s \n' % self.notification_mail_address +\
                   '#PBS -m bea \n' +\
                   '#PBS -e %s \n' % error_path +\
                   '#PBS -o %s \n' % output_path +\
                   '#PBS -l walltime=%s,mem=%s \n' % ( self.time_bound, self.mem_bound) +\
                   '#PBS -v PYTHONPATH=%s \n' % ':'.join(python_paths) +\
                   command + '\n'  
        return script

    def extract_job_id(self, output):
        
        job_number    = output.strip()
        return job_number


    
 

        
class Remote_NFS_Launcher(Remote_Single_Head_Launcher):
    """
    Distribute jobs by ssh on machines that are connected to one file server 
    to which the local machine is not connected.
    It is almost the same as PBS_Launcher. Only committing and dismissing is 
    different.
    The role of cluster-head is taken by an arbitrary node connected to the NFS.
    The user has to specify which node can take how many jobs. 
    """ 
            
    def dismiss_parts(self, partnames):
        """Kill jobs on computing nodes."""
        pk = self.pathkeeper
        
        phm = load_from_shelve(pk.path_abs('phm'))
        
        self.send('Pull pids ....')
        pid_paths =  [ pk.path_rel_base('pid', partname=partname) 
                       for partname in partnames ]
        
        self.synchronizer.pull(pid_paths)
        self.send('Pull pids done.', lb=2)

        pids = []
        for partname in partnames:
            pid_path     = pk.path_abs('pid', partname=partname)
            pid   = get_lines(pid_path)[0]
            pids += [ pid ]
            
            epart = 'kill ' + pid 
            command = ['ssh', self.remote_user + '@' + phm[partname], epart ]
            output = shell(command)
            
            self.send(output)
        
        self.send("Cluster dismiss done.")


    def engage_nodes(self, partnames):
        """
        @param partnames: list of strings, each representing a part of the
            experiment. See peewit.core.Experiment.get_my_part_indices 
        
        commits the partial runs to computing nodes and creates agent 
        that waits for the result.    
        """ 
        
        pk = self.pathkeeper

        ## how distribute jobs to host
        phm = self.get_phm(partnames, verbose=True)

        #pam = self.get_pam(partnames, verbose=False)
        
        save_in_shelve(phm, pk.path_abs('phm'))
                
        ## commit to PBS     
        
        ## TODO: store pids
        for partname in partnames:

            python_paths = self.get_remote_peewit_paths()#pam[partname])

            command,  output_path, error_path  = self.python_call(partname)

            
            command_ = \
                'export PYTHONPATH=$PYTHONPATH:' + ':'.join(python_paths) +' ; ' +\
                'nohup nice ' + command +\
                ' 1> ' + output_path + ' 2> ' + error_path + ' &'
            
            ssh_command = ['ssh', '-f', self.remote_user + '@' + phm[partname], command_ ]

            sleep(0.5)

            shell(ssh_command)
            
            
            
        self.send("Remote submit done.\n")
            
              
    def get_phm(self, partnames, verbose=False):
        "partname-host mapping"

        ## list copy of remote_node_spec
        rms = []
        for  node in self.remote_nodes_spec:
            rms += [ [node[0], node[1]] ]

        ## want pop it
        partnames = partnames[:]

        
        n_nodes = sum([ node[1] for node in rms  ])
        n_jobs  = len(partnames)
        
        for node in rms:
            ## how many job should node take at most
            node[1] = ceil( node[1] * (1. * n_jobs/n_nodes) )
                
        ## build the mapping
        phm = {}
        for node in rms:
            while node[1] > 0 and partnames:
                partname = partnames.pop(0)        
                phm[partname] = node[0]
                node[1] = node[1] - 1 

        if verbose:
            self.send('\nDistribution of jobs:\n')
            for part in sorted(phm):
                self.send(part, ': ', phm[part])  
            self.send('\n')
            
        return phm
        

    def get_pam(self, partnames, verbose=False):
        "partname-archictecture mapping"

        phm = self.get_phm(partnames, verbose=False)

        pam = {}

        for partname in partnames:
            for node in self.remote_nodes_spec:
                if phm[partname] == node[0]:
                    try:
                        assert node[2] in ['32bit', '64bit', '32', '64']
                        pam[partname] = node[2]
                    except:
                        self.send('Cannot read archicture of node %s, set to 64bit' \
                                  % node[0])
                        pam[partname] = '64bit'
                    break
            
                   
         
        if verbose:
            self.send('Architecture for jobs:\n') 
            for part in sorted(pam):
                self.send(part, ': ', pam[part])  
            self.send('\n')
          
        return pam


class Local_NFS_Launcher(Remote_NFS_Launcher):
    """
    Distribute jobs by ssh on machines that are connected to one file server 
    to which the local machine is also connected.
    It inherits all from Remote_NFS_Launcher, only the the synchronizer is 
    replaces by a dummy-synchronizer. The latter does nothing, since 
    anything takes place on the same file-system.
    """ 
    
    def init(self):
        
        self.synchronizer                 = Dummy_Synchronizer()
        
        
           
        
def Launcher(key):
    """
    Factory yielding launcher defined in this module.
    @todo: this cant be the wright way to do this. In particular
        we would like Laucher(key) to yield the specified lauchner 
        _and_ Launcher to be a type itself.
    """
            
    klm = {'SGE': SGE_Launcher(),
           'RFS': Remote_NFS_Launcher(),
           'LFS': Local_NFS_Launcher()
           }
        
    if key in klm:
        return klm[key]
    else:
        raise NotImplementedError('unknown launcher back-end:' + key)




if __name__ == '__main__':
    
    lc = Launcher('LFS')
    print(lc.__dict__)
    
    
    
