"""Whenever it is about paths handling, peewit relies on this modules."""

import os, re, platform, socket, shutil, logging, sys, inspect, posixpath, getpass, glob

from os.path import join as opj

        

import peewit.__init__

from cut.baseclass  import Named
from cut.fileaccess import system_path, create_dir, save_file,\
     get_lines, posix_path, get_content_lines

from time import sleep
from collections import OrderedDict


class PathKeeper(Named):
    """Keeps the the path information of an experiment or an an v-cube.
       Provides methods for querying the path. 
       """

    def peewit_version(self):
        """Upcoming release number
        Testet with python 3.3.2, numpy 1.7.1"""
        return '0.10'
    
    def init(self,

             base_dir     = None,
             project_dir  = None,
             treename     = None,
             version      = None,
             partname     = None,
             target_ename = None,

             python_paths       = [],
             python_paths_32    = [],
             python_paths_64    = [], 
             python_paths_os    = [], 
             more_dirs          = [],
             remote_python_paths_os    = [],
             remote_head_architecture  = None,
             sync_executable_32        = None,
             sync_executable_64        = None,
             remote_sync_executable_32 = None,
             remote_sync_executable_64 = None,
             
             
             ):

        for paths in self.path_groups_keys:
            self.__setattr__(paths, [])
    
    
    use_pylab = True
    
    path_groups_keys_a  = [ 
                            'project_path',
                            'python_paths',
                            'python_paths_32',
                            'python_paths_64',
                            'python_paths_os',
                           ]

    path_groups_keys_r  = [ 'peewit_path',
                            'python_paths_os',
                          ]
    
    
    path_groups_keys    = path_groups_keys_a + path_groups_keys_r


    suffices = {}

    suffices_p = {'vcube'   : 'vc',
                  'out'     : 'out',
                  'err'     : 'err',
                  'pid'     : 'pid',   # pid of run
                  'jnr'     : 'jnr',   # sge job number 
                  'phm'     : 'phm',   # part host mapping
                  'state'   : 'state', 
                  'qsub'    : 'pbs',   # PBS script
                  'run_number' : 'rn', 
                  'deps_gv' : 'gv', 
                  'deps_pdf': 'pdf', 
                  'deps_svg': 'svg', 
                  
                  
                  
                 
                  } 
    suffices_h = { 
                  'run_list'   : 'runs',
                  'state_list' : 'versions',
                  'not_tip'    : 'no_at_tip'
                  }
    
    suffices.update(suffices_p,)
    suffices.update(suffices_h,) 
    
    
    
    dirs     = { 'vcube': ['vcubes'],
                 'log'     : ['logs'],
                 'out'     : ['logs'],
                 'err'     : ['logs'],
                 'life'    : ['.peewit','run_files'],
                 'pid'     : ['.peewit','run_files','pids'],
                 'jnr'     : ['.peewit','run_files','jnrs'],
                 'phm'     : ['.peewit','run_files','phms'],
                 'state'   : ['.peewit','run_files','states'],
                 'qsub'    : ['.peewit','run_files','qsubs'],
                 'run_number': ['.peewit', 'run_files','rns'],
                 
                 'plot'    : ['plots' ],
                 'deps_gv' : ['plots', 'e_nodes' ],
                 'deps_pdf': ['plots', 'e_nodes' ],
                 'deps_svg': ['plots', 'e_nodes' ],
                 
                 'history' : ['.peewit'],
                 'peewit'  : ['.peewit'],
                 'snippets': ['tex' ],
                 'kept'    : ['kept'],
                }
    
    np_sep = '--'
    

    def digest_paths(self, archivist=None, launcher=None):
        """
        @type archivist: peewit.Archivist
        @type launcher: peewit.Launcher
        """  
    
        ## set treename to name of calling module
        ## often sys.argv[0]) by must not
        
        
        stack_short     = [ s for s in inspect.stack() if os.path.isfile(s[1]) ]
                
        #for i,s in enumerate(stack_short):
        #    print("SSA " + str(i), s[0], s[1], s[2], s[3], type(s[3]))
        maybe_path      = stack_short[2][1] ## 0: paths.py, 1: project_cont.py
        path            = maybe_path
        base            = os.path.basename(path)
        treename        = os.path.splitext(base)[0]
        self.treename   = treename
        

        #path            = sys.argv[0]
        #dir             = sys.path[0] 
        module_abs      = os.path.abspath(path)
        self.module_abs = module_abs
        
        #print "EEE", self.module_abs
        ## read user paths definitions 
        
        self.set_base_dir()
        
        ## store orignal desciptors
        self.path_groups_orig = OrderedDict()
        for group in self.path_groups_keys:
            if self.has(group):
                self.path_groups_orig[group] = self.__getattribute__(group)
    
        peewit  = os.path.relpath(self.dir_abs_peewit(), self.abs_path(''))
        self.path_groups_orig['peewit_path']  = [ peewit ] 

        project = self.project_dir
        self.path_groups_orig['project_path'] = [ project ] 
        
    
        ## normalize
        def normpath(p):
            return posixpath.normpath(posix_path(p))

        self.base_dir    = normpath(self.base_dir)        
        self.project_dir = normpath(self.project_dir)
        
        self.path_groups_norm = OrderedDict()
        for group in self.path_groups_keys:
            self.path_groups_norm[group] = [ normpath(p) 
                                             for p in self.path_groups_orig[group] ]
            

        ## uniquize
        self.path_groups_uniq = OrderedDict()
        all_paths = set()
        for group in self.path_groups_keys:
            self.path_groups_uniq[group] = []
            for path in self.path_groups_norm[group]:
                if not path in all_paths:
                    self.path_groups_uniq[group] += [ path ]
        
        self.path_groups = self.path_groups_uniq
        
        
        ## register python paths
        self.extend_python_path()
        
        if archivist:
            archivist.set_paths(self)
        
        if launcher:
            launcher.set_paths(self)
            
        
    def set_base_dir(self):
        """In case of project export the user who downloads the project
        overwrites the base-dir in the file project-dir/.peewit/base_dir .
        Otherwise the user would have to place the base-dir at the same place 
        rel to HOMEDIR as the author of the project.
        In order to find that base_dir-file peewit assumes that the first 
        dir that has a .peewit-dir going up from the main module is the project dir.
        """ 
        
        #module_dir = os.path.dirname(self.module_abs)
        
        #self.send('Set base-dir', self.module_abs)
        
        derived_project_dir = None
        head, tail  = os.path.split(self.module_abs)
        while tail: 
            if os.path.isdir(opj(head,  *self.dirs['peewit'])):
                derived_project_dir = head
                break
            head, tail = os.path.split(head)
            
        if derived_project_dir:
            pw_dir  = opj( derived_project_dir,
                           *self.dirs['peewit'])
                      
            path    = opj(pw_dir, 'base_dir')
            
            if os.path.isfile(path) and get_content_lines(path):
                base_dir_overwrite = get_content_lines(path)[0]
                self.base_dir = base_dir_overwrite
                self.send("Overwritten base-dir rel HOMEDIR: ", self.base_dir)
        
        assert self.base_dir is not None
        
    def extend_python_path(self):
        """Extend python path with paths defined by user in project configuration"""

        paths = []
        
        ## plain user space paths
        paths += [  opj(self.dir_abs_base(),system_path(path) ) 
                    for path in self.path_groups['python_paths'] ]

        
        ## arch.-dependent paths
        architecture = platform.architecture()[0]
        assert architecture in ['32bit', '64bit' ]

        if   architecture == '32bit':
            paths += [  opj(self.dir_abs_base(),system_path(path) )
                        for path in self.path_groups['python_paths_32'] ]
        elif architecture == '64bit':
            paths += [  opj(self.dir_abs_base(),system_path(path) ) 
                        for path in self.path_groups['python_paths_64'] ]
        
        ## outer space paths
        for path in self.path_groups['python_paths_os']:
            if posixpath.isabs(path):
                paths += [ system_path(path) ]
            else:
                paths += [ opj(self.dir_abs_base(),system_path(path) )  ]
        
        ## do it
        os.sys.path = os.sys.path[:1] + paths + os.sys.path[1:]
        
        #for p in os.sys.path:
        #   self.send(p)
        
        
    def store_run_number(self, run_number):
        create_dir(self.dir_abs_pdir('run_number'))
        save_file( str( run_number ), self.path_abs('run_number'))
           

    def read_run_number(self):
        path = self.path_abs('run_number')
        try:
            run_number = get_lines(path)[0].strip()
        except IOError:
            run_number = None
              
        return run_number
        
 
    def save_pid_file(self):
        create_dir(self.dir_abs_pdir('pid'))
        save_file( str( os.getpid()), self.path_abs('pid'))
        

    def create_project_dirs(self):
            
        create_dir(self.dir_abs_pdir('log'))
        create_dir(self.dir_abs_pdir('vcube'))
        ## TODO: presentation  paths-integration
        create_dir(self.dir_abs_pdir('plot'))
        create_dir(self.dir_abs_pdir('history'))
        save_file('hg wants files.',
                        opj(self.dir_abs_pdir('history'), '.hg_dummy'))
        
    def create_launch_dirs(self):
        create_dir(self.dir_abs_pdir('life'))
        create_dir(self.dir_abs_pdir('state'))
        create_dir(self.dir_abs_pdir('pid'))
        create_dir(self.dir_abs_pdir('jnr'))
        create_dir(self.dir_abs_pdir('qsub'))
        create_dir(self.dir_abs_pdir('phm'))



    def remove_current_files(self):
        """
        Removes files written by peewit for the current version of the current 
        e-tree. It is called if the Experiment.remove_current_files flag is set and 
        after dimiss of a launch. 
        """
        pattern_target   = r'.[^-]*'
        pattern_partname = r'.[^-]*'
        
        paths_to_be_removed = []
        kinds  = [ kind for kind in list(self.suffices_p.keys()) ]
        #if not kind in ['err','out'] ]
        for kind in kinds:
            
            dir          = self.dir_abs_pdir(kind)
            path_pattern = self.filename(kind, 
                                         target   = pattern_target,  
                                         partname = pattern_partname,)
            
            if os.path.isdir(dir):
            
                files_all       = list(os.walk(dir))[0][2]
            
                files_match = [ x for x in files_all if re.match(path_pattern, x) ]
            
                paths_to_be_removed += [ opj(dir,p) for p in files_match ]

        if paths_to_be_removed:
            self.send('Clean-up: remove files produced for this version.')
        else:
            self.send('Clean-up: nothing to remove.')
                    
        for path in paths_to_be_removed:
            if os.path.isfile(path):
                os.remove(path)
            
        sleep(0.5)
        

 


    def mm_join(self, *parts):
        sep = self.np_sep
        basename = ''
        for part in parts[:-1]:
            basename += str(part).lower() + sep
        basename += str(parts[-1]).lower()
        
        return basename


    def basename(self, target=None, partname=None):
        target   = self.target_ename if target is None else target
        partname = self.partname if partname is None else partname

        return  self.mm_join( self.treename, self.version, target, partname)     

    def read_basename(self,basename):
        
        nameparts = basename.split(self.np_sep)
        
        self.treename      = nameparts[0]
        self.version       = nameparts[1]
        self.target_ename  = nameparts[2]
        self.partname      = nameparts[3]
    


    def host(self):
        return socket.gethostname()

    def user(self):
        #return os.getlogin()
        ## more reliable?
        return getpass.getuser()

    def homedir(self):
        return os.path.expanduser('~') 


    def extend_project_dir(self, *extensions):
               
        return opj(self.dir_abs_base(), 
                   self.extend_project_dir_rel_base(*extensions))
        
    def extend_project_dir_rel_base(self, *extensions):
        
        return system_path(self.project_dir, *extensions)   


    def extend_project_dir_rel_base_posix(self, *extensions):
        
        return posix_path(self.extend_project_dir_rel_base(*extensions))   



    def dir_abs_life(self):
        return self.extend_project_dir('life')

  
    def dir_rel_base_pdir(self, kind):
        return self.extend_project_dir_rel_base(*self.dirs[kind])


    def dir_rel_base_pdir_posix(self, kind):
        return self.extend_project_dir_rel_base_posix(*self.dirs[kind])


    
    ######
    
    def abs_path(self, extension=None):
        """Return abs to extension that is given in posix-notation 
           relative to the peewit base directory"""
        
        base = self.dir_abs_base()
        
        if extension:
            return opj( base, system_path(extension) )
        else:
            return base 

    
    def dir_abs_base(self):
        
        if os.path.isabs(self.base_dir):
            return system_path(self.base_dir)
        else:
            return opj(self.homedir(), 
                       system_path(self.base_dir))
        
    
    def dir_abs_pdir(self,kind):
        return opj(self.dir_abs_base(), self.dir_rel_base_pdir(kind) )
        

    def dir_abs_peewit(self):
        
        path_abs_core  = peewit.__init__.__file__
        dir_abs_peewit = opj(path_abs_core, os.pardir, os.pardir, os.pardir)
        out = os.path.normpath(dir_abs_peewit)
        return out
        
    
    def dir_abs_peewit_packages(self):
        
        path_abs_core  = peewit.__init__.__file__
        dir_abs_peewit = opj(path_abs_core, os.pardir, os.pardir)

        return os.path.normpath(dir_abs_peewit)        

    def dir_abs_project(self):
        return opj( self.dir_abs_base(), 
                    system_path(self.project_dir) )
    
    ######

    def path_abs(self, kind, partname=None, target=None):    
        return opj(self.dir_abs_pdir(kind),
                            self.filename(kind, partname, target)
                            )
                          
    def path_rel_base(self, kind, partname=None, target=None):    
        return opj(self.dir_rel_base_pdir(kind),
                            self.filename(kind, partname, target)
                            )

    def path_rel_base_posix(self, kind, partname=None, target=None):    
        return opj(self.dir_rel_base_pdir_posix(kind),
                            self.filename(kind, partname, target)
                            )
 
    def filename(self, kind, partname=None, target=None):
        return  self.basename(target=target, partname=partname) + '.' + self.suffices[kind]
   

    def filename_life(self,target=None, partname=None):
        return  self.basename(target=target, partname=partname) + '.life'    
    
    def path_abs_life(self,target=None,partname=None):    
        return opj( self.dir_abs_life(),
                    self.filename_life(target, partname) )
                                

    def path_module_rel_base(self):
        module_abs = self.module_abs
        return module_abs[len(self.dir_abs_base()) + 1 :] 

    def peewit_dir_rel_base(self):
        return os.path.relpath(self.dir_abs_peewit(), self.abs_path(''))

 
    def history_dir_rel_base(self):
        return system_path(self.project_dir, *self.dirs['history'] )

    def history_dir_abs(self):
        return system_path(self.dir_abs_base(), 
                           self.history_dir_rel_base())

    def history_path_rel_base(self,table_name):
        filename = self.mm_join(self.group, self.host(), self.user()) +\
                   '.' + self.suffices[table_name]                    
        return opj( self.history_dir_rel_base(), filename) 
                   
    def history_path_abs(self, table_name):
        return opj(self.dir_abs_base(),
                   self.history_path_rel_base(table_name))

    
    def history_path_abs_all(self, table_name):
        history_dir = self.history_dir_abs()
        files       = glob.glob1(history_dir, "*." +  self.suffices[table_name] )
        return [ opj(history_dir, file) for file in files] 
         

    def start_logging_again(self):
        
        if self.has('recorder'):
            rc = self.recorder 
            
            old_handler = rc.handler
            
            rc.handler.close()
            
            
            
            new_out_path = os.path.join(self.dir_abs_pdir('log'),
                                        self.filename('out') )
            
            dir = os.path.split(new_out_path)[0] 
             
            create_dir(dir)
            
            if os.path.isfile(rc.out_path):
                shutil.move(rc.out_path, new_out_path )
            
            rc.out_path = new_out_path 
            
            
            rc.handler = logging.FileHandler(rc.out_path,'a')
    
            rc.handler.setFormatter(rc.formatter) 
                    
            rc.logger.addHandler(rc.handler)
            rc.logger.removeHandler(old_handler)
        
        
    
    def stop_logging(self):
        if self.has('recorder'):
        
            rc = self.recorder 
            rc.logger.close()
            
            file = open(rc.out_path)
            log_so_far = file.read().strip()
            file.close()
            if not log_so_far:
                os.remove(rc.out_path)
            
            
            
            sys.sdtout = self.stdout
            sys.sdterr = self.stderr
            
            ## not needed because shell greps stdout,stderr directly
            ## when called by launcher
            #sys.stdout.write(log_sor_far)
            #os.remove(rc.output)
        
        
        

class StreamLogger(object):
    """
    This is a modified snippet provided by 
    Denis Barmenkov  on stackoverflow.com.
    """
    
    def __init__(self, stream, logger, prefix=''):
        self.stream = stream
        self.logger = logger
        self.prefix = prefix
        self.data = ''

    def write(self, data):
        self.stream.write(data)
        

        self.data += data
        tmp = str(self.data)
        if '\x0a' in tmp or '\x0d' in tmp:
            tmp = tmp.rstrip('\x0a\x0d')
            self.logger.debug('%s%s' % (self.prefix, tmp))

            self.data = ''

    def flush(self):
        self.stream.flush()



class Recorder(Named): 
    
    def start_logging(self, path):
                
        self.logger  = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)
        
        self.out_path = path
        
        self.handler = logging.FileHandler(self.out_path,'w')

        self.logger.addHandler(self.handler)

        self.formatter = logging.Formatter('%(message)s')
        
        self.handler.setFormatter(self.formatter) 
        
        ## in case of single and main runs,
        ## duplicate stderr and stdout to same log-file
        
        self.sdtout = sys.stdout
        self.sdterr = sys.stderr
        
        sys.stdout = StreamLogger(sys.stdout, self.logger )
        sys.stderr = StreamLogger(sys.stderr, self.logger )
          




if 'recorder' not in PathKeeper.__dict__ and not '--part' in sys.argv:
    
    
    pk_dummy = PathKeeper()
    path          = inspect.stack()[0][1] 
    parent_dir    = os.path.split(path)[0]
    
    dir_orig      = os.path.abspath(os.path.curdir)
    if parent_dir:
        os.chdir(parent_dir)
        
    log_dir_rel   = opj(*pk_dummy.dirs['log'])
    
    dir_guess_1 = opj(os.path.pardir, log_dir_rel)
    dir_guess_2 = opj(log_dir_rel)
    dir_guess_3 = opj(os.path.curdir)
    
    
    if os.path.isdir(dir_guess_1):
        dir = dir_guess_1
    elif os.path.isdir(dir_guess_2):
        dir = dir_guess_2
    else:
        dir = dir_guess_3
    
    path = os.path.join( dir, 'unnamed.out')
    
    PathKeeper.recorder = Recorder()
    PathKeeper.recorder.start_logging(path)

    os.chdir(dir_orig)
    

if 'wrote_version' not in PathKeeper.__dict__:
    ### write version file
    pk = PathKeeper()
    
    version_path = opj(pk.dir_abs_peewit(), 'stuff', 'version.txt')
    content_new  = pk.peewit_version() 
        
    if os.path.isfile(version_path):
        content_old = get_lines(version_path)[0]
        if not content_new == content_old:
            save_file(content_new, version_path)
    else:
        save_file(content_new, version_path)
        
    PathKeeper.wrote_version = True
    

if __name__ == '__main__':

    pass







