"""
File synchronization between local machine to remote host. Used by peewit.paralellization. 
"""

import os
from os.path import join as opj
from time import sleep

import ntpath

from cut.fileaccess import system_path, create_dir, save_file, posix_path
from cut.baseclass import Named
from cut.mixedtools import shell
from peewit.paths import PathKeeper
import posixpath

class Dummy_Synchronizer(Named):
    """Does not do anything. Is used for launches where computing hosts
    are bound to smae file system as local machine."""
    
    
    def pull(self, paths, omit=[]):
        sleep(0.5)
        pass
    
    def push(self, paths, omit=[]):
        sleep(0.5)
        pass
    
    def sync(self, paths, action='sync', omit=[]):
        sleep(0.5)
        pass
    

class Unison_Synchronizer(Named):
    
    def init(self,
             base_dir        = None,
             local_home_dir  = None,
             remote_home_dir = None,
             remote_base_dir = None,
             remote_user     = None,
             remote_host     = None,
             verbose         = 0,
             unison_dir_name = '.unison',
             unison_dir_parent = None,
             sync_executable         = None,
             remote_sync_executable  = None,):
        pass

    def pull(self, paths, omit=[]):    
        return self.sync(paths, action='pull', omit=omit)
    
    
    def push(self, paths, omit=[]):
        return self.sync(paths, action='push', omit=omit)
    
    
    def sync(self, paths, action='sync', omit=[]):
        """@param paths: paths relative to base-dir""" 
        
        assert action in ['push', 'pull', 'sync']
        
        ## root dir of sync
        root_rel_base    = self.define_root(paths)
        root_rel_home    = posix_path( opj(self.base_dir, root_rel_base))

        root_local       = posix_path(opj(self.local_home_dir, root_rel_home))

        base_dir_remote  = opj( self.remote_home_dir, self.remote_base_dir) 
        root_remote_abs  = posix_path( opj( base_dir_remote, root_rel_base ) )
        
        
        root_remote      = 'ssh://%s@%s/%s' % (self.remote_user, 
                                               self.remote_host, 
                                               root_remote_abs)
        
        prf ='root   = %s   \n' % root_local  +\
             'root   = %s   \n' % root_remote +\
             'prefer = %s   \n' % root_local  +\
             'batch  = true \n'               +\
             'confirmbigdeletes = false\n'             
        
        
        if self.remote_sync_executable:
            prf += 'servercmd=' + self.remote_sync_executable + '\n'
         
        if not action == 'sync':
            forced = root_local if action == 'push'  else root_remote
            prf += 'force  = %s \n' % forced 

        for path in paths:
            
            path = posixpath.normpath(path)
            l_prefix = len(root_rel_base)
            path_descriptor = posix_path(path[l_prefix:]) 
            if path_descriptor.startswith('/'):
                path_descriptor = path_descriptor[1:]
            if path_descriptor == '':
                path_descriptor = './'
            
            prf += 'path   = %s   \n' % path_descriptor 

        for pattern in omit:
            prf += 'ignore = Name  %s \n' % pattern
        
        if self.verbose:
            self.send('\nApplied unison conf:\n' + prf)

        

        unison_dir       = opj( self.unison_dir_parent, self.unison_dir_name)
             
        if self.verbose:
            self.send("local unison dir:", unison_dir)
                               
        create_dir(unison_dir)
        
        unison_conf_base        = self.remote_host + '_' + action
        unison_conf_base_suffix = unison_conf_base + '.prf'
        
        unison_conf_path = system_path( unison_dir, unison_conf_base_suffix)
        
        save_file(prf,unison_conf_path)
                
        env           = os.environ
        env['UNISON'] = unison_dir
        
        executable = self.sync_executable if self.sync_executable else 'unison'
        
        command       = [executable, '-ui', 'text', unison_conf_base]
        
        output = shell(command)
        
        if self.verbose:
            self.send('\n' + output +  '\n')
        else:
            self.send(lb=0)
            dummy = output[:]
        

    def define_root(self, paths):
        """
        @param paths: 
        @return: the parent dir of the greatest common path of elements in paths

        """


        if   len(paths) == 1:
            root, dummy = os.path.split(paths[0])
            if dummy == '':
                root, dummy = os.path.split(root)
        elif len(paths) >  1:
            root, dummy = os.path.split(os.path.commonprefix(paths))
        else:
            self.bugger('no paths given')
        
        return root
        


    
if __name__ == '__main__':
    
    paths = ['ad/af/asdfg.dd', 'ad/af/asdrr.dd']

    root_rel_home, dummy = os.path.split(os.path.commonprefix(paths))
    
    

    if 0:
    
        pk = PathKeeper()
        
        us = Unison_Synchronizer(remote_host='cluster',
                                remote_user='lw',
                                remote_homed_dir='/nis/home/lw',
                                local_home_dir=pk.homedir())
        
        path0 = 'c:\\asd\\asdd\\ada/asdad'
        path1 = 'eee/asda/'
        path2 = 'eee/ggfg/rer'
        
        s = posix_path(path0, os_path=ntpath)
        
        #print(path)
        #print(s) 
        
        
        #quit()
        
        paths = [ path1,
                  path2 ]
        
        paths = ['bin/ed', 'bin/ee']
        
        
        us.push(paths)
        
    
    
    
    
      
    
