"""A simple way to explain the p-v-model for experiment used by peewit is this.
A p-v-model conform experiment is an experiment that yields a multidimensional 
result cube. The number of values for each of each experimental dimension does 
not depend on the values of other dimension, hence result is a cube. In the 
peewit framework such a cube is called v-cube and this is the corresponding module.
""" 


from numpy import array, shape, concatenate, nanmax, nanmin, nanargmax, nanargmin,\
                  reshape, empty, median, size, nan, zeros, sum

from cut.numpytools import nan_mean, nan_std, number_like, subarray, m_indices,\
    check_m_index, first
from cut.baseclass  import Named
from cut.mixedtools import value2string

from cut.container import multi_range, has_unique_value, non_equal_elements, indices 
from collections import OrderedDict 
from cut.fileaccess import save_in_shelve, create_dir, load_from_shelve

from copy import deepcopy


import os
from peewit.paths import PathKeeper

## TODO: in case no nan-variant is provided peewit should not aggregate over
##       unset values
afunctions   = {'mean':nan_mean, 'std':nan_std, 'min':nanmin, 'max':nanmax, 
                'median':median, 'first':first  }

afunctions_a = {'argmin':nanargmin, 'argmax':nanargmax }


class VCube(Named):
    """A v-cube is a multi-dimensional array with arbitrary entries along 
    with meta-data that describe the dimensions."""
    
    ## for future migrations only
    __version__ = (0,)
    
    pathkeeper = None
    
    
    def init(self, 
             
             cube = array(None,dtype=object),  ## actual values 
             bube = array(None,dtype=bool),    ## whether values is set
             sube = array(None,dtype=object),  ## shadow values
             
             has_shadow_values  = False,
             value_labels       = [],   ## names for values along last dimension
             
             vcubes        = OrderedDict(),  ## ename to parent vcube                          
             target_ename  = None,
             side_vcubes   = [],  ##
             
             basename      = None,
             pathkeeper    = None,
             run_number    = None,
             read_failed   = False,
            ):
        
        
        if not pathkeeper:
            if VCube.pathkeeper:
                self.pathkeeper = VCube.pathkeeper
            else:
                self.pathkeeper = PathKeeper()
        
        if target_ename:
            self.pathkeeper.target_ename = target_ename
        
        if basename:
            assert self.pathkeeper
            self.pathkeeper.read_basename(basename)
            self.read()
            return
        elif pathkeeper:
            #self.read()
            pass
        
    
        self.vcubes[self.target_ename] = self
        
        n_dims     = len(self.vcubes)            
        cube_shape = tuple( [1] * (n_dims - 1) + [0] )
        
        self.cube = array([],ndmin=n_dims, dtype=object)
        self.cube.reshape( cube_shape )
        
        self.bube = array([],ndmin=n_dims, dtype=bool)
        self.bube.reshape( cube_shape )
        
        self.sube = array([],ndmin=n_dims, dtype=object)
        self.sube.reshape( cube_shape )

        self.check_basic()
        
        
    def __call__(self,
                 ename             = None,
                 input_refs        = [],
                 test_selection    = [],
                 descent_selection = [],
                 ):
        """For using vcube as enode in etree-definition.
        """
        
        ## if ename rename self
        pass
        ##
        self.input_refs = input_refs
        self.test_selection = test_selection
        self.descent_selection = descent_selection

        return self
    
    
    def equals(self, vc):
        """Equality. We want __eq__ to remain identity.
        """
        
        main = ( self.target_ename == vc.target_ename   and
                 self.value_labels == vc.value_labels   and
                 self.dim_names()  == vc.dim_names()    and
                 self.dim_sizes()  == vc.dim_sizes()    and
                 self.side_vcubes  == vc.side_vcubes    and
                 ( self.bube == vc.bube ).all()         and
                 ( self.cube == vc.cube ).all()         
               )
        
        if not main:
            return False
        
        for ename in self.dim_names():
            if not ename == self.target_ename:
                if not self.vcubes[ename].equals(vc.vcubes[ename]):
                    return False
        
        return True
        
        
        
    
    def show(self, **args):
        
        from . import presentation
        
        vi = presentation.Viewer(self, **args)
        vi.print_vcube()

    
    def save(self, full=None):
        """
        @param full: list of non-target dims for which we save values  
            not shadow-values, None means all
        """

        d = self.pathkeeper.dir_abs_pdir('vcube')
        create_dir(d)
        path = os.path.join( d, self.pathkeeper.path_abs('vcube'))  
        self.send('\nSave vcube\n%s\n' % path) 
        vcube_copy = self.deepcopy()
        save_in_shelve( vcube_copy, path )
        

    def read(self, target=None):
        self.send('Read saved v-cube')
        if self.pathkeeper.partname is None:
            guess = 'main'
            self.send('Guess part-name: ', guess)
            self.pathkeeper.partname = guess
        
        
        filename  = self.pathkeeper.filename('vcube',target=target)
        path      = os.path.join( self.pathkeeper.dir_abs_pdir('vcube'),
                                  filename
                                 )
        path_kept = os.path.join( self.pathkeeper.dir_abs_pdir('kept'),
                                  filename
                                 )
        self.send('%s' % path)
        if not ( os.path.isfile(path) or os.path.isfile(path_kept) ):
            self.send("cannot find vc-file \n%s" % filename )
            self.read_failed = True
        else:             
            #o = load_from_shelve(path)
            try:
                o = load_from_shelve(path)
                self.set_attributes(o.__dict__)
            except Exception as e:
                try:
                    print(e)
                    o = load_from_shelve(path_kept)
                    self.set_attributes(o.__dict__)
                    
                except:
                    self.send('Reading v-cube failed.')
                    self.read_failed = True
        
        
        if not self.read_failed:
            
            ## For reasons we have not understood 
            ## refs to vcubes that should be identical are not 
            ## after storing and reading again
            ## Therefore we reset the refs in .vcubes
            
            self.vcubes[self.target_ename] = self
            #print(self.vcubes.keys())
            #print(self.side_vcubes)
            
            for ename in self.vcubes.keys():
                vcube_ = self.vcubes[ename]
                #print("\n" + ename)
                #print(vcube_.vcubes.keys())
                for ename_ in vcube_.vcubes.keys():
                    if ename_ in self.vcubes:  
                        vcube_.vcubes[ename_] = self.vcubes[ename_]
           
            ##      
            self.check()
        
        self.send()


    def thin_copy(self):
        """Deepcopy of meta-data, non-recursive.."""
        out_cube = VCube()
        
        out_cube.target_ename       = self.target_ename
        out_cube.pathkeeper         = deepcopy(self.pathkeeper)
        out_cube.run_number         = self.run_number
        out_cube.value_labels       = deepcopy(self.value_labels)
        out_cube.vcubes             = OrderedDict( [(e,None) for e in self.vcubes.keys() ])
        out_cube.side_vcubes        = self.side_vcubes[:]#OrderedDict( [(e,None) for e in self.side_vcubes.keys() ])
        
        out_cube.vcubes[out_cube.target_ename]   = out_cube
        out_cube.basename           = self.basename           
        out_cube.read_failed        = self.read_failed
        out_cube.has_shadow_values  = self.has_shadow_values
        #
        return out_cube

    def copy(self):
        """Deepcopy beside for self.cube if an object-cube"""
        
        return self.copy_g( shallow_copy=True )
    
    def deepcopy(self):
        """Likewise copy but also tries to deepcopy cube. If this fails,
        values are replaced by shadow_values."""
        return self.copy_g()
    
    def copy_g(self, 
               recursive    = True, 
               shallow_copy = False,
               shadow_copy  = False,
               cdone        = None ):
        """Generic copy, deep or shallow on cubes. Engine for copy 
        and deepcopy.If values cannot be deepcopied, shadow-values are copied instead
        @param shadow_copy: copy shadow-values instead of values
        """
        
        if cdone is None:
            cdone={}
        #print("VVV")
        #print(cdone)
        #print(self.value_labels)
        
        vcube = self.thin_copy()
        #print(vcube.value_labels)
        if   shallow_copy:
            vcube.cube = self.cube.copy()
        elif shadow_copy:
            self.shadow_copy(vcube) 
        else:
            try:
                vcube.cube = deepcopy(self.cube)
            except:
                self.send("Could not serialize cube %s. Write shadow values."
                          % vcube.target_ename)
                self.shadow_copy(vcube) 
        
        vcube.bube = deepcopy(self.bube)
        vcube.sube = deepcopy(self.sube)
        
        cdone[id(self)] = vcube
        
        if recursive:
            
            for ename in self.dim_names():
                id_original = id(self.vcubes[ename])
                
                if id_original in cdone:
                    vcube.vcubes[ename] = cdone[id_original]
                else:
                    
                    cdone[id_original]   = \
                    vcube.vcubes[ename] = self.vcubes[ename].copy_g( recursive=recursive,
                                                                     shallow_copy=shallow_copy,
                                                                     shadow_copy=shadow_copy,
                                                                     cdone=cdone
                                                                     )
                     
        return vcube 



    def shadow_copy(self, vcube):
    
        sizes = list(self.dim_sizes().values())
            
        sube = empty(sizes, dtype=object)
            
        for m_index in multi_range(sizes):
            if self.bube[m_index]:
                value  = self.cube[m_index]
                sube[m_index] = self.shadow_value(value)
            else:
                sube[m_index] = None
            
        vcube.has_shadow_values = True
        vcube.cube = sube
        
        

    def shadow_value(self, value):
        """smaller, derived value that is to some extend sufficient to indentify value"""
        return str(value)


    def drop(self, ename):
        """Remove vcube with target ename from self.vcubes. Also remove all dependent 
        vcubes."""
        
        Manipulation.drop_singleton_parents(self, ename)
        
        return self
        


    def check_basic(self):
        """Basic assertions on v-cubes that can be check even though
        the v-cubes is not fully build"""
        
        assert non_equal_elements(self.dim_names())
        
        assert shape(self.cube) == shape(self.bube)
        assert not self.sube or shape(self.sube) == shape(self.cube)
        
        assert self.target_ename in self.vcubes
        
                

    def check(self, done=None):
        """Assertion on grown-up vcubes, ones that are not on their ways
        being build. 
        """
        
        if done is None:
            done = []
        
        self.check_basic()
                
        assert len(self.vcubes) == len(shape(self.cube)), \
            (self.target_ename, len(self.vcubes), len(shape(self.cube)) )

        assert not self.dim_names() or self.cube.size > 0
        
        assert self.vcubes[self.target_ename] is self,\
            ( self.target_ename, self, self.vcubes[self.target_ename]) 
        
        for ename, vcube_ in self.vcubes.items():
                
            assert vcube_, (self.target_ename, ename)
            
            ## target name constraint
            assert ename == vcube_.target_ename
            
            if ename in self.side_vcubes:
                ## singleton side cube constraint
                assert self.dim_sizes()[ename] == 1, \
                    (self.target_ename,ename, self.dim_sizes())
            else:
                ## unique size constraint                
                assert len(vcube_.value_labels) == self.dim_sizes()[ename],\
                (self.target_ename, ename, vcube_.value_labels, self.dim_sizes())
                ## subset constraint
                non_side_cubes_ = [ vc__ for en__, vc__ in vcube_.vcubes.items() 
                                    if not en__ in vcube_.side_vcubes ]
                assert set(non_side_cubes_).issubset(self.vcubes.values()), \
                    (self.target_ename, ename)
                
        
        ## recurse 
        done += [ self ]
        for vcube_ in self.vcubes.values():
            if not vcube_ in done:
                vcube_.check(done=done)
        
        
        
    def dim_names(self):
        return list(self.vcubes.keys())   

    
    def dim_names_non_target(self):
        out = self.dim_names()
        if self.target_ename in out:
            out.remove(self.target_ename)
        return out
    
    def target_index(self):
        return self.dim_names().index(self.target_ename)     
    
    def dim_sizes(self):
        assert shape(self.cube) == shape(self.bube)
        assert len(self.dim_names()) == len(shape(self.cube)),\
             ( self.target_ename, self.dim_names(), shape(self.cube)) 
        return OrderedDict(list(zip(self.dim_names(), shape(self.cube))))   

    
    def effective_dim_sizes(self):
        out = OrderedDict()
        for i,ename in enumerate(self.dim_names()):
            bube = deepcopy(self.bube)
            for j in indices(self.dim_names()):
                if i != j:
                    bube = bube.max(axis=j)
            buber = list(bube.ravel())
            first = buber.index(True)
            last  = len(buber) - list(reversed(buber)).index(True) - 1
            size  = last - first + 1
            out[ename] = size
        return out
    
    def get_value(self, mid={}):
        """Assuming that all dims not mention in mid are singletons
        the unique value of the cube is returned.
        
        For specifying entries and sub-cubes we deal with dicts that
        map e-names either to indices, list of indices, or more general to
        specification of index-lists. To this end we use the following convention
        for variable names:
        
        mid:   multi-index-dict, maps e-names to indices
        mild:  multi-index-list-dict, maps e-names to list of indices.
        misd:  multi-index-specification-dict: maps e-names to indices,
               list of indices or slices. 
               
        milds are convenient as we can index arrays by them. We therefore 
        provide the function self.mild(..) that ouputs a mild based on 
        a misd (what includes mids)  
        """
        
        mild = self.mild(self.dim_names(), mid, singles=True)
        value  = subarray(self.cube, list(mild.values()))
        assert value.size == 1, \
            "There are more than one values to fit your query. " +\
            "Use get_subcube() instead"
        
        return value.ravel()[0]
    
    
    def get_values(self):
        squeezed =  self.cube.squeeze()
        try:   
            len(squeezed)
            return squeezed.tolist()
        except TypeError:
            return [squeezed]
        
 
    def set_value(self, mid, value):
        """Set a single entry of the cube. For explantion of mid see
        get_value."""
        for ename, size in self.dim_sizes().items():
            assert ename in list(mid.keys()) or size == 1
            
        mild = self.mild(misd=mid, first=True)
        
        #print("GGGGGGGGGGGGGGG")
        #self.show()
        #print(mild)
        #print(value)
        #print(self.cube.shape)
        #indexes = [ m[0] for m in mild.values() ]
        indexes = list(mild.values())
        #print(indexes)
        #print(self.cube.dtype)
        ## trick needed to pass iterables as entry? 
        value_a = array([None], dtype=object)
        value_a[0] = value
        self.cube[indexes] = value_a
        self.bube[indexes] = True
        #self.show()


    def get_subcube(self, mid):
        """Return sub-cube of self as specified by mid. The sub-cube
        has only those dims that are not fixed in mid. """
        
        ## type-check mid
        assert all( type(spec) is int for spec in mid.values())
    
        vcube = self.copy()
        for ename, index in mid.items():
            selection = [index]
            Manipulation.selection_parents(vcube, ename, selection, by_names=False)
            Manipulation.drop_singleton_parents(vcube, ename)
            
        return vcube
    
    def set_subcube(self, mid, vcube):
        
        mild              = self.mild(None,mid)
        
        mil = list(mild.values())
            
        cube_ = vcube.cube
        bube_ = vcube.bube
        
        self.cube[mil] = cube_
        self.bube[mil] = bube_
        if self.sube:
            sube_ = subarray(vcube.bube,mil)
            self.sube[mil] = sube_
        
            
              
    def mild(self,all_dims=None, misd={}, singles=True, first=False):
        """Multi-Index-List-Dict. Returns dict enames to list of indices 
        accordings to misd. misd is a Multi-Index-Specification-Dict that 
        maps enames to a index-spec. A spec is either an single index, a list 
        of indices or a slice, see code.
        This function is mainly used to complete partial index-specification.
        """
        
        if all_dims is None:
            all_dims = self.dim_names()
        
        ## filter sizes over all_dims        
        sizes = {}
        for e,s in self.dim_sizes().items():
            if e in all_dims:
                sizes[e] = s

        mild = OrderedDict( list(zip(all_dims, [ None ] *  len(all_dims))) )
            
        for ename in all_dims:
            all_indices = list(range(sizes[ename]))
            if ename in misd and not misd[ename] is None:
                spec = misd[ename]
                if   type(spec) is int:
                    if singles:
                        mild[ename] = [ spec ]
                    else:
                        mild[ename] = list(range(spec))
                elif type(spec) is list:
                    mild[ename] = spec
                elif type(spec) is slice:
                    mild[ename] = list(spec)
                else:
                    self.bugger(ename, spec)
            else:
                if first:
                    mild[ename] = [0]
                else:
                    mild[ename] = all_indices
                          
        return mild


    def get_value_name(self, mid):
        assert self.target_ename in mid
        
        value = self.get_value(mid)
        index = mid[self.target_ename]
        name  = VCube.value_name(value, index)
        assert type(name) == type('')
        
        return name

    
    
    @staticmethod
    def value_name(value, index):

        name = value2string(value)
              
        if len(name) > 31:
                name = '#' + str(index)
        
        return  name

    def set_labels(self, labels):
        
        if not self.value_labels:
            self.value_labels = labels
        else:
            m = min( len(self.value_labels), len(labels))
            for j in range(m):
                if self.value_labels[j] != labels[j]:
                    self.value_labels[j] =  '@' + str(j)
            if len(self.value_labels) < len(labels):
                self.value_labels += labels[len(self.value_labels):]
                    
    
    def set_values(self,values, mid, labels):
        """ Sets values on target-dim with other dims selected by mid.
        """
        
        mid = deepcopy(mid)
        
        assert labels is None or len(values) <= len(labels),\
               (len(values), len(labels), values, labels) 
        
        mid_ = OrderedDict()
        for ename in self.vcubes:
            if  ename in mid.keys():
                mid_[ename] = mid[ename]
            else:
                mid_[ename] = 0
            
         
        self.extend_cubes(values, mid_)
        
        target     = self.target_ename
        
        mid__ = OrderedDict()
        for k,index in mid_.items():
            if k in self.dim_names():
                mid__[k] = index
        
        for j,v in enumerate(values):
            mid__[target] = j
            
            m_index  = tuple(mid__.values())
            
            self.cube[m_index] = v
            self.bube[m_index] = True
            
        self.set_labels(labels)
    
        
    def add_values(self,values_new, values_old, mid, labels_new):
        """
        Adds values to target-dim with other dims selected by mid."""
        
        mid = deepcopy(mid)
        
        size_old    = len(values_old) if values_old else 0
        values_all  = values_old + values_new
        labels_all  = self.value_labels[:len(values_old)] + labels_new 
        
        assert len(values_all) == len(labels_all)
        
        self.extend_cubes(values_all, mid)
        
        target  = self.target_ename
        
        mid_    = OrderedDict()
        for k,index in mid.items():
            if k in self.dim_names():
                mid_[k] = index

        
        for j,v in enumerate(values_new):
            mid_[target] = size_old + j
            
            m_index  = tuple(mid_.values())
            self.cube[m_index] = v 
            self.bube[m_index] = True
        
        self.set_labels(labels_all)
             
             
    def extend_cubes(self, values, m_v_index):
        """
        Increases shape where m_v_index is too small. 
        Subroutine for self.add_values.
        """
        
        
        
        target = self.target_ename
        sizes  = self.dim_sizes()
        
        m_v_index = deepcopy(m_v_index)
        
        ## in this case: maximal index such that
        ## target is tread like other dims concerning cube-expansion
        m_v_index[target] = len(values) - 1
        
        ## get dtype
        
        cube_dtype = self.new_dtype(self.cube, values)
        
        self.cube  = self.cube.astype(cube_dtype)
        
        
        
        for i,ename in enumerate(self.dim_names()):
            size_d = m_v_index[ename] + 1 - sizes[ename]
            if size_d > 0:
                sizes_        = self.dim_sizes()
                sizes_[ename] = size_d
                sizes_        = list(sizes_.values())
                
                new_part_v    =         empty(sizes_, dtype=cube_dtype)
                new_part_b    = False * empty(sizes_, dtype=bool)
                
                self.cube = concatenate((self.cube,new_part_v), axis=i)
                self.bube = concatenate((self.bube,new_part_b), axis=i)


    def new_dtype(self,cube, values):
        
        ## if values are sequence we are in trouble
        ## because array(values) would try to make 
        ## multi-dim-array
        
        ## type from cube
        if self.cube.size > 0:
            type_pre = self.cube.dtype
            first_pre = self.cube.ravel()[0]
            
            if number_like(first_pre):
                dummy_pre  = array([first_pre])
                type_new   = dummy_pre.dtype
            else:
                type_new = object            
        else:
            type_pre = None
            
        ## type from values
        if not values:
            t = object
        else:
        
            assert values
            first    = values[0]
            if number_like(first):
                dummy_new  = array([first])
                type_new   = dummy_new.dtype
            else:
                type_new = object
            
            if  type_pre == None:
                t = type_new
            elif type_pre == type_new:
                t =  type_pre
            elif type_pre == object or type_new == object:
                t = object
            else:
                t = concatenate((dummy_pre,dummy_new)).dtype    
            
        return t
        
        
        
    def merge_straight(self, minor, join_dim, main=None):
        
        assert join_dim in self.dim_names()        
        assert self.dim_names()   == minor.dim_names()
        
        for ename_ in self.dim_names():            
            assert ename_ == join_dim or self.dim_sizes()[ename_] == minor.dim_sizes()[ename_],\
            (ename_,  self.dim_sizes()[ename_],minor.dim_sizes()[ename_] )
            
        
        target_index = self.dim_names().index(join_dim)
        
        self.cube = concatenate((self.cube, minor.cube), axis=target_index )
        self.bube = concatenate((self.bube, minor.bube), axis=target_index )
        
        if self.target_ename == join_dim:          
            self.value_labels += minor.value_labels[:] 
        
        if main is None :
            for ename_, vc  in self.vcubes.items():
                if ename_ != self.target_ename and join_dim in vc.dim_names(): 
                    vc_ = minor.vcubes[ename_]
                    assert join_dim in vc_.dim_names()
                    vc.merge_straight(vc_, join_dim, main=self )            
            
        

    def objectize_cube(self, **args):
        """Build array of type CubeElement where each element hold 
        the corresponding element of self.cube."""

        o_cube = empty( self.cube.shape, dtype='O')
        for mindex in multi_range(list(self.dim_sizes().values())):
            value = self.cube[mindex]     
            o_cube[mindex] = CubeElement(value=value, **args)

        return o_cube


    def aggs(self,aggregations):
        """aggs takes a list of aggregations that are computed in inverse order on
        a copy of self and returns the resulting v-cube.
        An element of the list a tuple of strings each representing an 
        aggregation given in polish notation. You can apply the these
        operators and operands: 
        sel ename value_name__1 value_name__2.. : select from dimension ename the
            values with names value_name__i  
        max  ename__1  ename__2 .. : 
        min  ename__1  ename__2 .. : 
        mean ename__1  ename__2 .. :
        std  ename__1  ename__2 .. : apply corresponding aggregation to the
            dimensions ename__i
        argmax ename__1  ename__2 .. : 
        argmin ename__1  ename__2 .. : find the joint optimizers in 
            dimensions ename__i for any combination of values of the remaining 
            dimensions. These aggregation can only be done last, e.g. be the 
            first the list, since they return an v-cube.    
        """
        
        assert not self.read_failed
        assert not self.cube is None
             
        vcube = self.copy()
        for ag in reversed(aggregations):
            ag_name = ag[0]
            ag_args = ag[1:]
            
            ## sel
            if ag_name in ['sel','del','seli']:
                
                ename        = ag_args[0]
                values_names = list(ag_args[1:]) 
                by_names     = False if ag_name in ['seli'] else True    
                Manipulation.selection_parents(vcube, 
                                                       ename,
                                                       values_names, 
                                                       by_names=by_names,
                                                       mode=ag_name)
            ## max
            elif ag_name in list(afunctions.keys()):
                enames = ag_args
                ag_function = afunctions[ag_name]
                Manipulation.ag(vcube, enames, ag_function)
                
                
            ## argmax
            elif ag_name in [ name       for name in afunctions_a.keys()]:
                arg_ag_function = afunctions_a[ag_name]
                enames          = list(ag_args)
                return Manipulation.arg_ag_objects(vcube,
                                                   enames, 
                                                   arg_ag_function)
                     
            ## argmaxi,...
            elif ag_name in [ name + 'i' for name in afunctions_a.keys()]:
                ag_name         = ag_name[:-1]
                arg_ag_function = afunctions_a[ag_name]                
                ## not an v-cube but an i-cube
                ## such that this can only be last aggregation
                return Manipulation.arg_ag_indices_r(vcube, 
                                                     ag_args, 
                                                     arg_ag_function)
            else:
                self.bugger("Unknown aggregation '%s'." % ag_name)
                                    
        return vcube
        
        
    def arg_ag(self, arg_aggregation):
          
        arg_ag_function = afunctions_a[arg_aggregation[0]]
        enames          = arg_aggregation[1:]


        vcubes = Manipulation.arg_ag_objects(self, enames, arg_ag_function )
        
        out  = []
        if all( [ vc.cube.size == 1 for vc in list(vcubes.values()) ]):
            for vc in vcubes.values():
                out += [ vc.cube.ravel()[0] ]
            return out
        elif len(vcubes) == 1:
            return vcubes.popitem()[1]
        else:
            return vcubes
                
   
   


class ICube(Named):
    """A cube (array) holding list of indices. Plus some meta-data.
    """

    def init(self, 
             vcube,         ## vcube to which the indices refer to
             targets = [],  ## list of dims to which the indices refer to
             ):
        
        dims_all = vcube.dim_names()
        
        assert( all( [en in dims_all for en in targets]))
        
        cube_shape = []
        for en in dims_all:
            if not en in targets:
                cube_shape.append( vcube.dim_sizes()[en] )
        cube_shape = tuple(cube_shape)
        
        self.iube  = self.empty_iube(cube_shape)
        
        self.enames = vcube.dim_names()
    
        self.dims_other = [ en for en in self.enames if not en in self.targets]
    
    def empty_iube(self, cube_shape):
        iube  = empty(cube_shape, dtype=object)
        
        for m_index in multi_range(cube_shape):
            iube[m_index] = []
        
        return iube
            
    def check(self):
        
        assert set(self.targets).issubset(set(self.vcube.dim_names()))
        assert self.iube.dtype == object
        assert len(self.dims_other) == len(self.iube.shape)
        assert len(self.targets) + len(self.iube.shape) == len(self.vcube.dim_names()),\
               (len(self.targets), len(self.iube.shape),len(self.vcube.dim_names()))
        
        return True
    
    
class Manipulation(Named):


    @staticmethod
    def selection_parents(vcube, ename, selection, 
                    by_names=True, inline=True, mode='sel' ):
        
        for vcube_ in reversed(list(vcube.vcubes.values())):
            if ename in vcube_.dim_names():
                Manipulation.selection(vcube_, ename, selection, 
                                       by_names=by_names, mode=mode )
        
        
    @staticmethod
    def selection(vcube, ename, selection, 
                  by_names=True, mode='sel' ):
        """    
        @param selection: list of of indices or value-names
        If selection for a given ename is empty (==[]) then no selection is applied.
        """ 
        
        
        assert ename in vcube.dim_names()

        if not type(selection) is list:            
            selection = [ selection ]
        
        itype = str if by_names else int
        if by_names:
            selection = [ str(s) for s in selection ]
            
        assert all([ type(s) is itype for s in selection ] ),\
            selection   
        
        labels   = vcube.vcubes[ename].value_labels
        n_values = len(labels)
        
        if by_names:
            assert all( [ vl in labels for vl in selection ] ),\
                (selection, labels) 
            selection = [ labels.index(vl) for vl in selection ] 
        else:
            assert all([ s < n_values for s in selection ]), (selection, labels) 
        
    
        if mode == 'del':
            all_indices = list(range(n_values))
            selection = [ i for i in all_indices if not i in selection ]
                    
        if selection == []:
                pass
        else:
              
            misd = vcube.mild(vcube.dim_names(), {ename:selection} )
            
            mis = list(misd.values())
            vcube.cube  = subarray(vcube.cube,mis)
            vcube.bube  = subarray(vcube.bube,mis)
            
            if ename == vcube.target_ename:
                vcube.value_labels = [ labels[i] for i in selection ] 
        


  
    @staticmethod
    def drop_all_singletons(vcube):
        
        for ename in vcube.dim_names():            
            if vcube.dim_sizes()[ename] <= 1 and ename != vcube.target_ename:
                Manipulation.drop_singleton_parents(vcube, ename)
  
  
        
  
    @staticmethod
    def drop_singleton_parents(vcube, ename):
         
        for vcube_ in list(reversed(list(vcube.vcubes.values()))):
            
            if ename in vcube_.dim_names() and\
               ename != vcube_.target_ename and\
               vcube_.dim_sizes()[ename] == 1:
                    
                    Manipulation.drop_singleton(vcube_, ename)
              
        
    
    @staticmethod
    def drop_singleton(vcube, ename):
        """Drops dim ename, assuming its singleton and not target."""
        assert vcube.dim_sizes()[ename] == 1,\
             (vcube.target_ename, ename,  vcube.dim_sizes()[ename]) 
        assert not vcube.target_ename == ename
        
        cube_shape = list(vcube.dim_sizes().values())
        
        eindex = vcube.dim_names().index(ename)
        
        new_shape  = cube_shape[:eindex] + cube_shape[eindex+1:] 
        
        vcube.cube = reshape(vcube.cube, new_shape)
        vcube.bube = reshape(vcube.bube, new_shape)
        if vcube.sube:
            vcube.sube = reshape(vcube.sube, new_shape)
        
        vcube.vcubes.pop(ename)


    @staticmethod
    def make_singleton_parents(vcube, ename):
        
        
        for vcube_ in vcube.vcubes.values():
        
            if ename in vcube_.dim_names():
                
                Manipulation.make_singleton(vcube_, ename)
            

    @staticmethod
    def make_singleton(vcube_,ename, label='shrink'):
        
        
        
        e_index      = vcube_.dim_names().index(ename)
        
        vcube_.cube  = first(vcube_.cube, axis=e_index)                    
        
        if vcube_.cube.dtype != object:
            vcube_.cube  = vcube_.cube.astype(object)
        
        for mi in m_indices(vcube_.cube):
            vcube_.cube[mi] = None ##TODO: use proper none-type instead?
            if vcube_.sube:
                vcube_.sube[mi] = None
            
        vcube_.bube  = Manipulation.abknd(vcube_.bube, sum, 
                                          axis=e_index, dtype=bool)
        
        
        if vcube_.target_ename == ename:
            vcube_.value_labels = [str(label)]
        
   
    @staticmethod
    def ag(vcube, enames, ag_function):
        """An aggregation is a function that summarizes the value along the
        dimension ename to a single value. 
        Parent cube that also depends on ename also have to be aggregated along
        that dimension, because the values in self are not associated with concrete 
        values of the parent cubes anymore. As it is unclear,  
        what the aggregated value in case of the parent would be, the concerned cube
        are noned, e.g. all value are set to None.
        TODO: does this mean false in bube? By now: no.
        """ 
        
        
        for ename in reversed(enames):
            
            assert ename in vcube.dim_names()
            
            for vcube_ in vcube.vcubes.values():
                
                if ename in vcube_.dim_names():
                    
                    e_index      = vcube_.dim_names().index(ename)
                    
                    if  vcube_ is vcube: # and ag_function is not first:
                        if ag_function in [nan_mean, nan_std]:
                            try:
                                vcube.cube = vcube.cube.astype('float64')
                            except:
                                pass
                                
                        vcube.cube  = Manipulation.abknd(vcube.cube, ag_function, 
                                                         axis=e_index, dtype=vcube.cube.dtype)
                        vcube.bube  = Manipulation.abknd(vcube.bube, sum, 
                                                         axis=e_index, dtype=bool)
                        
                    else: ## parent cube
                        if vcube_.dim_sizes()[ename] == 1:
                            label = vcube_.value_labels[0]
                        else:
                            label=ag_function.__name__ 
                        
                        Manipulation.make_singleton(vcube_, ename, label=label)
                        
                        if vcube_.cube.dtype != object:
                            vcube_.cube  = vcube_.cube.astype(object)
                        for mi in m_indices(vcube_.cube):
                            vcube_.cube[mi] = None ##TODO: use proper none-type instead?
                            
                        vcube_.bube  = Manipulation.abknd(vcube_.bube, sum, 
                                                          axis=e_index, dtype=bool)
                    
            
            vcube.check()
        


    @staticmethod
    def arg_ag_objects(vcube, enames, arg_ag_function):
        
    
        icube       = Manipulation.arg_ag_indices_r(vcube, enames, arg_ag_function)
        
        icube.check()
        
        dims_other   = icube.dims_other
        
        
        assert enames == icube.targets, (enames, icube.targets)
        
        out = OrderedDict()
        
        for index_iube, ename in enumerate(icube.targets):
            
            vcube_source  = vcube.vcubes[ename]
            dims_source   = vcube_source.dim_names()
            
            dims_out      = [ e for e in vcube.dim_names() if e in dims_other]
            targets       = [ e for e in vcube.dim_names() if e in icube.targets]
            ## create new vcube with target ename
            vcube_        = vcube_source.thin_copy()
            vcube_.vcubes = OrderedDict()
            cube_shape_   = OrderedDict()
            for e in dims_out:
                cube_shape_[e]    = vcube.dim_sizes()[e]
                vcube_.vcubes[e]  = vcube.vcubes[e]
 
            #cube_shape_[ename]   = 1           
            #vcube_.vcubes[ename] = vcube_

            cube_  = empty(list(cube_shape_.values()), dtype=vcube_source.cube.dtype)
            bube_  = empty(list(cube_shape_.values()), dtype=bool)
            
                     
            for m_index_out_d in m_indices(cube_, list(cube_shape_.keys())):
                
                mit_out = tuple(m_index_out_d.values())
                
                mid_source = OrderedDict()
                
                for e in dims_source:
                    if   e in dims_out:
                        mid_source[e] = m_index_out_d[e]
                    elif e in targets:
                        mid_source[e] = icube.iube[mit_out][index_iube]
                    else:
                        assert 0, (dims_source, dims_out, targets)
                
                value = vcube_source.cube[tuple(mid_source.values())]
                isset = vcube_source.bube[tuple(mid_source.values())]
                
                cube_[mit_out] = value
                bube_[mit_out] = isset
                              
                
                
                
            vcube_.cube = cube_
            vcube_.bube = bube_
            
            out[ename] = vcube_
        
       
        if len(out) == 1:
            out = out.popitem()[1]
            
        return out
        
    @staticmethod
    def arg_ag_indices_r(vcube, enames, arg_ag_function):
        icube = ICube(vcube)
        
        for ename in reversed(enames):
            icube = Manipulation.arg_ag_indices(vcube, 
                                                ename, 
                                                arg_ag_function, 
                                                icube_pre = icube )
        

        return icube

    @staticmethod
    def arg_ag_indices(vcube, ename, arg_ag_function, icube_pre = None):
        """
        Returns index-v-cube that holds indices on dim ename but
        still has same dims as vcube. The indices refer to the values of
        dim ename given by the arg_ag_function applied to vcube. 
        """
        pass
        if not icube_pre:
            icube_pre = ICube(vcube)
        
        icube_pre.check()
        
        #e_index_all   = vcube.dim_names().index(ename)
        e_index_other = icube_pre.dims_other.index(ename)
        
        cube_ag_pre   = Manipulation.subcube_by_icube(vcube, icube_pre)
        iube_next     = arg_ag_function(cube_ag_pre, axis=e_index_other)
        
        icube_out     = ICube(vcube, targets = [ename] + icube_pre.targets )
        iube_out      = icube_out.iube
        
        ## merge iube_pre and iube_add into iube_out
        for m_index in m_indices(iube_out):
            
            index_next        = int(iube_next[m_index])
            
            m_index_pre       = m_index[:e_index_other] + (index_next,) +\
                                m_index[e_index_other:]
            
            check_m_index(m_index_pre, icube_pre.iube )
            
        
            iube_out[m_index] = [ index_next ] + icube_pre.iube[m_index_pre]
            
            
        return icube_out
        
        
    @staticmethod
    def subcube_by_icube(vcube, icube):
        """
        Return sub-cube of vcube.cube according to indices provided by icube."""
        
        icube.check()
        
        dims        = vcube.dim_names()
        dims_target = icube.targets
        dims_other  = icube.dims_other
        
        cube_out = empty(icube.iube.shape,dtype=vcube.cube.dtype)
        
        for m_index in m_indices(cube_out):

            m_index_ = [ None ] * len(dims)
            for i,ename in enumerate(dims):
               
                if ename in dims_target:
                    m_index_[i] = icube.iube[m_index][dims_target.index(ename)]
                else:
                    m_index_[i] = m_index[dims_other.index(ename)]
            m_index_ = tuple(m_index_)
            
            
            cube_out[m_index] = vcube.cube[m_index_]  
        
        
        assert icube.iube.shape == cube_out.shape
        return cube_out
            
        
 
        
    @staticmethod 
    def abknd(A, agg_func, axis, dtype):
        """aggregate but keep number of dimension"""
        
        old_shape  = shape(A) 
        
        A          = A.astype(type(0.))
        A          = agg_func(A, axis=axis)
        new_shape  = tuple( list(old_shape[:axis])+[1]+list(old_shape[axis+1:]))
        A          = reshape(A , new_shape)
        A          = A.astype(dtype)
        
        return A
 
    @staticmethod 
    def partially_permute(vcube, enames, rigid=True, main=None, vlabels=None):
        """Reorders the dimension. If dimension are missing in enames,
        then enames specifies the end of the dimension list."""
        

        assert set(enames).issubset(set(vcube.dim_names()))
        
        not_mentioned   = [ ename for ename in vcube.dim_names()   
                            if not ename in enames ]
        
        new_order       = not_mentioned  + enames
        
        
        def permute(vcube_, new_order):
            
            
            permutation  = [ vcube_.dim_names().index(ename)   
                             for ename in new_order ]
        
            vcube_.cube       = vcube_.cube.transpose(permutation)
            vcube_.bube       = vcube_.bube.transpose(permutation)
            if vcube_.sube:
                vcube_.sube   = vcube_.sube.transpose(permutation)
                           
            vcubes_ = OrderedDict()
            for ename in new_order:
                vcubes_[ename] = vcubes[ename]
                
            vcube_.vcubes = vcubes_
            
            
        
        vcubes  = OrderedDict(vcube.vcubes.items())
                    
        for ename in new_order:
            vc = vcubes[ename]
            new_order_    = [ en for en in new_order 
                              if en in vc.dim_names() ]
            
            permute(vc, new_order_)
            
        vcube.check()
    
    
    @staticmethod
    def rename_dim(main, ename, new_ename):
        
        for vcube in main.vcubes.values():
            
            vcubes_ = OrderedDict()

            for ename_, vcube_ in vcube.vcubes.items():
                
                if ename == ename_:
                    vcubes_[new_ename] = vcube_
                else:
                    vcubes_[ename_]    = vcube_
                
            
            vcube.vcubes = vcubes_
        
            
    



class MergingOOS(Named):        

    def init(self):
        self.bugger("Class out of service.")


    @staticmethod
    def merge_vcubes(sub_vcubes, ename):
        
        if sub_vcubes:
            vc = None
            for sub_vcube in sub_vcubes:
                
                if sub_vcube.has('pnode'):
                    del sub_vcube.pnode
                
                vc = VCube.merge_two_vcubes(vc, sub_vcube, ename=ename)\
                     if vc else sub_vcube
            return vc
        else:
            return VCube()#value_names=OrderedDict([ (ename,[]) ]))


    @staticmethod
    def merge_two_vcubes(mhy, thy, ename=None):
        
        if not mhy.dim_names():
            return thy
        if not thy.dim_names():
            return mhy
        
        assert mhy.dim_names() 
        assert thy.dim_names() 
        
        mhy_dims = mhy.dim_names()
        thy_dims = thy.dim_names()
        
        if ename:
            assert ename in mhy_dims and ename in thy_dims, ename
                    
        ## do we fit?
        assert set(mhy_dims) == set(thy_dims) 
        if not mhy_dims == thy_dims:
            ## turn a blind eye to different order
            thy.partially_permute(mhy_dims)    
        our_dims = thy_dims = mhy_dims
        
        #ename_join = ename = our_dims[0] if ename is None else ename
        ename_join = ename

        for _,ename in enumerate(our_dims):
            
            thy_vnames = thy.value_names[ename]
            mhy_vnames = mhy.value_names[ename]
            
            if  ename == ename_join:
                ## ok, called by Experiment build_vcube
                pass
                #assert set(thy_vnames).isdisjoint(set(mhy_vnames))
            
            elif len(thy_vnames) == len(mhy_vnames):
                ## todo: build new vnames                
                for i,vn in enumerate(thy_vnames):
                    if not mhy_vnames[i] == vn:
                        mhy_vnames[i] = thy_vnames[i] = '@' + str(i)
                                               
            else:
                ## complicated case
                if  len(thy_vnames) < len(mhy_vnames):
                    thy = VCube.trim_dim_size(thy,mhy, ename)
                else:
                    mhy = VCube.trim_dim_size(mhy,thy, ename)

        ush = VCube.concat_vcubes([mhy,thy], ename_join)
        
        return ush
                    
    @staticmethod  
    def trim_dim_size(minor, major, ename):
        """Makes vcubes minor, major to have same dim-size at
        dim ename. Asserts that major is not smaller at ename and takes its
        value names as basis"""
        
        
        minor_vnames = minor.value_names[ename]
        major_vnames = major.value_names[ename]
    
        
        free = OrderedDict(list(zip(list(range(len(major.dim_names() ))),
                               major.dim_names(), 
                               )) )
        positions = OrderedDict()
        for i,ename_ in enumerate(minor_vnames):
            
            if ename_ in list(free.values()):
                j_ = list(free.values()).index(ename_)
                j  = list(free.keys())[j_]
                positions[i] = j
                free.remove(j)
                
        for i,ename_ in enumerate(minor_vnames):
            
            if not i in list(positions.keys()):
                j = list(free.keys())[0]
                free.pop(j) 
                positions[i] = j
                
        
        dummy = minor.selections({ename:[0]}, by_names=False, inline=False )
        
        axis = minor.dim_names().index(ename)
        minor_part_shape = shape(dummy.cube)
        minor_shape      = list(minor_part_shape[:])
        minor_shape[axis]=len(major_vnames)
                
        minor_cube = nan * empty(minor_shape)
        minor_bube = zeros(minor_shape, dtype=bool)

        slice_template = [ slice(size) for size in shape(minor.cube) ]

        for i,j in positions.items():
            slices_old = slice_template[:]
            slices_new = slice_template[:]
            
            slices_old[axis] = slice(i,i+1)
            slices_new[axis] = slice(j,j+1)
            
            minor_cube[slices_new] = minor.cube[slices_old]
            minor_bube[slices_new] = minor.bube[slices_old]
        
        new_minor = minor.thin_copy()
        #new_minor.pathkeeper   = minor.pathkeeper
        #new_minor.target_ename = minor.target_ename
        new_minor.cube = minor_cube
        new_minor.bube = minor_bube
        new_minor.value_names[ename] = major.value_names[ename]
        
        return new_minor    
                
                    
    @staticmethod
    def concat_vcubes(vcubes, ename):
        """Concatenates cubes in vcubes to self along axis of ename.
        Assumes that vcubes fit together well and challenge this."""
        
        ## check assumptions
        #all_target_enames   = [ vcube.value_names[ename]   for vcube in vcubes ]
        #assert non_equal_elements(all_target_enames), all_target_enames
        
        all_dim_sizes   = [ vcube.dim_sizes() for vcube in vcubes ] 
        for i in indices(vcubes):
            all_dim_sizes[i].pop(ename) 
        assert has_unique_value(all_dim_sizes)
        
        out_cube = vcubes[0].thin_copy()
        

        axis = out_cube.dim_names().index(ename)
        ## including type conversion
        out_cube.cube = concatenate( [vc.cube for vc in vcubes ], axis= axis)
        out_cube.bube = concatenate( [vc.bube for vc in vcubes ], axis= axis)
        
        out_cube.value_names[ename] = []        
        for vc in vcubes:
            out_cube.value_names[ename] += vc.value_names[ename]
        
        return out_cube                         
  

            
    @staticmethod
    def bold_merge(major, minor, ename=None):
        
        merge_dim = ename 
        
        new_minor=major.align_dims(minor)
        
        if merge_dim is None:
            for ename, _ in major.value_names.items():
                dim_values   = major.value_names[ename]
                dim_values_  = new_minor.value_names[ename]
                            
                if set(dim_values).isdisjoint( set(dim_values_) ):
                    merge_dim = ename 
                    break
        
        assert merge_dim
        
        out = VCube.merge_two_vcubes(major, new_minor, merge_dim)
        
        return out
    
        
    def align_dims(self, minor):
        """Aligns minors dims two the dims of self. That is each dim of minor
        is either mapped  to a existing dim of self or a corresponding dim is 
        added to minor.
        
        Returns vcube having same dims as self such that it can be merge with self."""

        mhy = self
        thy = minor
        
            
        for ename in thy.dim_names():
            if not ename in mhy.dim_names():
                if thy.dim_sizes()[ename] == 1:
                    thy.drop_singleton(ename) 
                    
        assert set(mhy.dim_names()).issuperset(set(thy.dim_names()))
        
        for ename in mhy.dim_names():
            if not ename in thy.dim_names():
                value_name = mhy.value_names[ename][0]
                thy.add_singleton(ename, value_name, position=-1)
        
        assert set(mhy.dim_names()) == set(thy.dim_names())
                
        thy.partially_permute(mhy.dim_names())
        
        return thy
        
        
        
        
        
  
    def add_singleton(self, ename, value_name):
        
        assert ename not in self.dim_names()
    
        position = len(self.dim_names()) - 1 
            
        self.value_names.insert(position, ename, [value_name])        
        
        cube_shape = list(shape(self.cube))
        
        new_shape  = cube_shape[:position] + [1] + cube_shape[position:] 
        
        self.cube = reshape(self.cube, new_shape)
        self.bube = reshape(self.bube, new_shape)
        


    
class ManipulationOOS(Named):
    """Manipulation function that are out of service. """
    
    def init(self):
        self.bugger("Class out of service.")

  
    def shrink(self, enames=None, keep_one=True):
        """Removes dimensions. If e-names, it removes all dims but those in e-names. 
        Otherwise, all dims that have size 1 are removed. 
        #If all dims have size 1 and 
        #keep_one, the first dim is kept alive though.
        """
        
        if enames:
            dim_names_new = enames 
        else:
            dim_names_new = [ ename for ename in self.dim_names() 
                              if self.dim_sizes()[ename] > 1 ]
        
        
        value_names_new = {}
        slices          = []

        for ename, s in self.dim_sizes().items():
            if ename in dim_names_new:
                slices  += [ slice(s) ]
                value_names_new[ename]            = self.value_names[ename]
            else:
                assert s == 1
                slices  += [ 0 ] 
       
        
        self.cube = self.cube[slices]
        self.bube = self.bube[slices]
        
        #self.dim_names              = dim_names_new
        self.value_names            = value_names_new
    

    def extract_member(self,member_name):
        
        vcube = self.copy()
        
        assert member_name in self.cube.ravel()[0].__dict__

        for mindex in multi_range(list(self.dim_sizes().values())):
            vcube.cube[mindex] = self.cube[mindex].__getattribute__(member_name)
        
        return vcube

    def map(self, minor, apply):
        """minor is a vcube with same dims as self. map applies 
        binary function apply to corresponding entries of minor and 
        self and returns resulting vcube.
        """
        
        out_cube  = self.thin_copy() 
        out_cube.cube = empty(shape(self.cube))
        
        for mindex in multi_range(list(self.dim_sizes().values())):
            if self.bube[mindex]:
                out_cube.cube[mindex] = self.cube[mindex](*minor.cube[tuple(mindex)])

        out_cube.bube = deepcopy(self.bube)

        return out_cube




    def repeat_singleton(self, ename, inline=False):
        
        vcube = self if inline else self.copy()
        
        e_index       = vcube.dim_names().index(ename)
        
        cube_shape    = list(vcube.dim_sizes().values())
        
        assert tuple(cube_shape) == shape(vcube.cube)
        
        s0  = cube_shape[e_index]
        
        reduced_shape = cube_shape[:e_index] + [1] + cube_shape[e_index+1:]
        

        for mindex in multi_range(reduced_shape):
            values = []
            balues = []
            mindex = list(mindex)
            for i in range(s0):
                mindex[e_index] = i
                value = vcube.cube[tuple(mindex)]
                
                balue = vcube.bube[tuple(mindex)]
                
                if balue:
                    values += [value]
                    balues += [balue]
                
            if len(balues) == 1:
                for i in range(s0):
                    mindex[e_index] = i    
                    vcube.cube[tuple(mindex)] = values[0] 
                    vcube.bube[tuple(mindex)] = balues[0] 
                        
        return vcube
  

        
    def rename_dim(self, ename,new_ename):
        
        eindex = self.dim_names().index(ename)
        
        new_value_names = self.value_names[ename]
        
        self.value_names.pop(ename)
        self.value_names.insert(eindex,new_ename, new_value_names)
        
    def rename_values(self,ename, mapping):
        
        value_names = self.value_names[ename]
        
        for old_name,new_name in mapping.items():
            vindex = value_names.index(old_name)
            value_names.pop(vindex)
            value_names.insert(vindex,new_name)
        
        


        
    ##TODO: rewrite, though it is not used by core modules
    def partially_permute_values(self, ename, value_names):
        """Permutes values of given dim. In contrast to partially_permute,
        take given value_names as beginning of the permutation."""
        
        eindex = self.dim_names().index(ename)
        
        value_names = deepcopy(value_names)
        
        not_mentioned = [ vn for vn in self.value_names[ename] 
                          if not vn in value_names ]
        
        new_order     = value_names + not_mentioned 
        
        new_cube = deepcopy(self.cube)
        new_bube = deepcopy(self.bube)

        slice_template = [ slice(size) for size in shape(self.cube) ]

        
        for new_index,vn in enumerate(new_order):

            old_index = self.value_names[ename].index(vn)
             
            slices_old = slice_template[:]
            slices_new = slice_template[:]
            
            slices_old[eindex] = slice(old_index,old_index+1)
            slices_new[eindex] = slice(new_index,new_index+1)
            
            new_cube[slices_new] = self.cube[slices_old]
            new_bube[slices_new] = self.bube[slices_old]
            
        self.cube = new_cube
        self.bube = new_bube
        self.value_names[ename] = new_order
        

           




 

   
class CubeElement(Named):
    """Wrapper class for elements of v-cubes. Needed to customize v-cube printing."""
    
    def init(self, 
             value=None,
             marker=None,
             marker_default = ' ',
             print_length_max   = None,
             ):
        
        pass
    
    def set_value(self,v):
        self.value = v
        
        
    def __repr__(self):
        return self.__str__()
        
    
    def __str__(self):
        
        marker = self.marker_default if self.marker is None else self.marker
        value_string = str(self.value)
        
        print_length_max = self.print_length_max
        
        
            
        
        if print_length_max is None:
            print_length_max = 14
            print_length = len(value_string)
        else:
            print_length = min([print_length_max, len(value_string)])

        
        
        value_string = value_string[:print_length]
        
        
        n_spaces = print_length_max - len(value_string) + 1
        
        spaces   = "".join([' '] * n_spaces)
        out      = ' ' + marker + value_string + spaces 

          
        return  out



if __name__ == '__main__':
    
    
    pass
    
    
    
    

    